@@ -1,29 +1,31 @@
 # Copyright 2002-2003 Nick Mathewson.  See LICENSE for licensing information.
-# $Id: BuildMessage.py,v 1.33 2003/01/05 13:19:53 nickm Exp $
+# $Id: BuildMessage.py,v 1.47 2003/06/11 19:05:35 nickm Exp $
 
 """mixminion.BuildMessage
 
    Code to construct messages and reply blocks, and to decode received
    message payloads."""
 
-import sys
-import zlib
 import operator
+import sys
+import types
+
 import mixminion.Crypto as Crypto
 from mixminion.Packet import *
-from mixminion.Common import MixError, MixFatalError, LOG
+from mixminion.Common import MixError, MixFatalError, LOG, UIError
 
 if sys.version_info[:3] < (2,2,0):
     import mixminion._zlibutil as zlibutil
 
 __all__ = ['buildForwardMessage', 'buildEncryptedMessage', 'buildReplyMessage',
-           'buildReplyBlock', 'decodePayload' ]
+           'buildReplyBlock', 'checkPathLength', 'decodePayload' ]
 
 def buildForwardMessage(payload, exitType, exitInfo, path1, path2,
                         paddingPRNG=None):
     """Construct a forward message.
             payload: The payload to deliver.  Must compress to under 28K-22b.
-                  If it does not, MixError is raised.
+                  If it does not, MixError is raised.  If the payload is
+                  None, 28K of random data is sent.
             exitType: The routing type for the final node. (2 bytes, >=0x100)
             exitInfo: The routing info for the final node, not including tag.
             path1: Sequence of ServerInfo objects for the first leg of the path
@@ -40,19 +42,30 @@ def buildForwardMessage(payload, exitType, exitInfo, path1, path2,
     if not path2:
         raise MixError("Second leg of path is empty")
 
-    LOG.debug("Encoding forward message for %s-byte payload",len(payload))
-    LOG.debug("  Using path %s/%s",
-                   [s.getNickname() for s in path1],
-                   [s.getNickname() for s in path2])
-    LOG.debug("  Delivering to %04x:%r", exitType, exitInfo)
+    suppressTag = 0
+    if exitType == DROP_TYPE:
+        suppressTag = 1
+        payload = None
 
     # Compress, pad, and checksum the payload.
-    payload = _encodePayload(payload, 0, paddingPRNG)
+    if payload is not None:
+        payload = _encodePayload(payload, 0, paddingPRNG)
+        LOG.debug("Encoding forward message for %s-byte payload",len(payload))
+    else:
+        payload = paddingPRNG.getBytes(PAYLOAD_LEN)
+        LOG.debug("Generating DROP message with %s bytes", PAYLOAD_LEN)
+
+    LOG.debug("  Using path %s:%s",
+                   ",".join([s.getNickname() for s in path1]),
+                   ",".join([s.getNickname() for s in path2]))
+    LOG.debug("  Delivering to %04x:%r", exitType, exitInfo)
 
     # Choose a random decoding tag.
-    tag = _getRandomTag(paddingPRNG)
-    exitInfo = tag + exitInfo
-    return _buildMessage(payload, exitType, exitInfo, path1, path2,paddingPRNG)
+    if not suppressTag:
+        tag = _getRandomTag(paddingPRNG)
+        exitInfo = tag + exitInfo
+    return _buildMessage(payload, exitType, exitInfo, path1, path2,
+                         paddingPRNG)
 
 def buildEncryptedForwardMessage(payload, exitType, exitInfo, path1, path2,
                                  key, paddingPRNG=None, secretRNG=None):
@@ -160,6 +173,11 @@ def _buildReplyBlockImpl(path, exitType, exitInfo, expiryTime=0,
        """
     if secretPRNG is None:
         secretPRNG = Crypto.getCommonPRNG()
+    if expiryTime is None:
+        # XXXX This is dangerous, and should go away; the user should
+        # XXXX *always* specify an expiry time.
+        LOG.warn("Inferring expiry time for reply block")
+        expiryTime = min([s.getValidUntil() for s in path])
 
     LOG.debug("Building reply block for path %s",
                    [s.getNickname() for s in path])
@@ -211,10 +229,6 @@ def buildReplyBlock(path, exitType, exitInfo, userKey,
     # message with 99.6% probability.  (Otherwise, we'd need to repeatedly
     # lioness-decrypt the payload in order to see whether the message was
     # a reply.)
-
-    # XXXX D'oh!  This enables an offline password guessing attack for
-    # XXXX anybody who sees multiple tags.  We need to make sure that userKey
-    # XXXX is stored on disk, and isn't a password.  This needs more thought.
     while 1:
         seed = _getRandomTag(secretRNG)
         if Crypto.sha1(seed+userKey+"Validate")[-1] == '\x00':
@@ -225,24 +239,53 @@ def buildReplyBlock(path, exitType, exitInfo, userKey,
     return _buildReplyBlockImpl(path, exitType, exitInfo, expiryTime, prng,
                                 seed)[0]
 
+def checkPathLength(path1, path2, exitType, exitInfo, explicitSwap=0):
+    """Given two path legs, an exit type and an exitInfo, raise an error
+       if we can't build a hop with the provided legs.
+
+       The leg "path1" may be null."""
+    err = 0 # 0: no error. 1: 1st leg too big. 2: 1st leg okay, 2nd too big.
+    if path1 is not None:
+        try:
+            _getRouting(path1, SWAP_FWD_TYPE, path2[0].getRoutingInfo().pack())
+        except MixError:
+            err = 1
+    # Add tag as needed to last exitinfo.
+    if exitType != DROP_TYPE and exitInfo is not None:
+        exitInfo += "X"*20
+    else:
+        exitInfo = ""
+    if err == 0:
+        try:
+            _getRouting(path2, exitType, exitInfo)
+        except MixError:
+            err = 2
+    if err and not explicitSwap:
+        raise UIError("Address and path will not fit in one header")
+    elif err:
+        raise UIError("Address and %s leg of path will not fit in one header",
+                      ["first", "second"][err-1])
+    
 #----------------------------------------------------------------------
 # MESSAGE DECODING
 
 def decodePayload(payload, tag, key=None,
-                  #storedKeys=None, # 'Stateful' reply blocks are disabled.
-                  userKey=None):
+                  userKeys=None):
     """Given a 28K payload and a 20-byte decoding tag, attempt to decode and
        decompress the original message.
 
            key: an RSA key to decode encrypted forward messages, or None
-           userKey: our encryption key for reply blocks, or None.
+           userKeys: a map from identity names to keys for reply blocks,
+                or None.
 
        If we can successfully decrypt the payload, we return it.  If we
        might be able to decrypt the payload given more/different keys,
        we return None.  If the payload is corrupt, we raise MixError.
     """
-    # FFFF Take a list of keys?
-    # FFFF Allow callbacks?
+    if userKeys is None:
+        userKeys = {}
+    elif type(userKeys) is types.StringType:
+        userKeys = { "" : userKeys }
 
     if len(payload) != PAYLOAD_LEN or len(tag) != TAG_LEN:
         raise MixError("Wrong payload or tag length")
@@ -252,26 +295,19 @@ def decodePayload(payload, tag, key=None,
     if _checkPayload(payload):
         return _decodeForwardPayload(payload)
 
-    # ('Stateful' reply blocks are disabled.)
-
-##    # If we have a list of keys associated with the tag, it's a reply message
-##    # using those keys.
-
-##     if storedKeys is not None:
-##      secrets = storedKeys.get(tag)
-##      if secrets is not None:
-##          del storedKeys[tag]
-##          return _decodeReplyPayload(payload, secrets)
-
     # If H(tag|userKey|"Validate") ends with 0, then the message _might_
     # be a reply message using H(tag|userKey|"Generate") as the seed for
     # its master secrets.  (There's a 1-in-256 chance that it isn't.)
-    if userKey is not None:
-        if Crypto.sha1(tag+userKey+"Validate")[-1] == '\x00':
-            try:
-                return _decodeStatelessReplyPayload(payload, tag, userKey)
-            except MixError:
-                pass
+    if userKeys:
+        for name,userKey in userKeys.items():
+            if Crypto.sha1(tag+userKey+"Validate")[-1] == '\x00':
+                try:
+                    p = _decodeStatelessReplyPayload(payload, tag, userKey)
+                    if name:
+                        LOG.info("Decoded reply message to identity %r", name)
+                    return p
+                except MixError:
+                    pass
 
     # If we have an RSA key, and none of the above steps get us a good
     # payload, then we may as well try to decrypt the start of tag+key with
@@ -375,7 +411,7 @@ def _buildMessage(payload, exitType, exitInfo,
         reply = path2
         path2 = None
     else:
-        if len(exitInfo) < TAG_LEN:
+        if len(exitInfo) < TAG_LEN and exitType != DROP_TYPE:
             raise MixError("Implausibly short exit info: %r"%exitInfo)
         if exitType < MIN_EXIT_TYPE and exitType != DROP_TYPE:
             raise MixError("Invalid exit type: %4x"%exitType)
@@ -388,7 +424,7 @@ def _buildMessage(payload, exitType, exitInfo,
     if paranoia:
         nHops = len(path1)
         if path2: nHops += len(path2)
-        secretRNG = Crypto.TrueRNG(SECRET_LEN*len(nHops))
+        secretRNG = Crypto.getTrueRNG()
     else:
         secretRNG = paddingPRNG
 
@@ -429,26 +465,18 @@ def _buildHeader(path,secrets,exitType,exitInfo,paddingPRNG):
            paddingPRNG: A pseudo-random number generator to generate padding
     """
     assert len(path) == len(secrets)
-    if len(path) * ENC_SUBHEADER_LEN > HEADER_LEN:
-        raise MixError("Too many nodes in path")
 
-    # Construct a list 'routing' of exitType, exitInfo.
-    routing = [ (FWD_TYPE, node.getRoutingInfo().pack()) for
-                node in path[1:] ]
-    routing.append((exitType, exitInfo))
-
-    # sizes[i] is size, in blocks, of subheaders for i.
-    sizes =[ getTotalBlocksForRoutingInfoLen(len(ri)) for _, ri in routing]
-
-    # totalSize is the total number of blocks.
-    totalSize = reduce(operator.add, sizes)
-    if totalSize * ENC_SUBHEADER_LEN > HEADER_LEN:
-        raise MixError("Routing info won't fit in header")
+    routing, sizes, totalSize = _getRouting(path, exitType, exitInfo)
+    if totalSize > HEADER_LEN:
+        raise MixError("Path cannot fit in header")
 
     # headerKey[i]==the AES key object node i will use to decrypt the header
     headerKeys = [ Crypto.Keyset(secret).get(Crypto.HEADER_SECRET_MODE)
                        for secret in secrets ]
 
+    # Length of padding needed for the header
+    paddingLen = HEADER_LEN - totalSize
+
     # Calculate junk.
     #   junkSeen[i]==the junk that node i will see, before it does any
     #                encryption.   Note that junkSeen[0]=="", because node 0
@@ -459,25 +487,28 @@ def _buildHeader(path,secrets,exitType,exitInfo,paddingPRNG):
         #
         # Node i+1 sees the junk that node i saw, plus the junk that i appends,
         # all encrypted by i.
+
         prngKey = Crypto.Keyset(secret).get(Crypto.RANDOM_JUNK_MODE)
-        # NewJunk is the junk that node i will append. (It's as long as
-        #   the subheaders that i removes.)
-        newJunk = Crypto.prng(prngKey,size*128)
+
+        # newJunk is the junk that node i will append. (It's as long as
+        #   the data that i removes.)
+        newJunk = Crypto.prng(prngKey,size)
         lastJunk = junkSeen[-1]
         nextJunk = lastJunk + newJunk
-        # Node i encrypts starting with its first extended subheader.  By
-        #   the time it reaches the junk, it's traversed:
-        #          All of its extended subheaders    [(size-1)*128]
-        #          Non-junk parts of the header      [HEADER_LEN-len(nextJunk)]
-        #
-        # Simplifying, we find that the PRNG index for the junk is
-        #    HEADER_LEN-len(lastJunk)-128.
-        startIdx = HEADER_LEN-len(lastJunk)-128
+
+        # Before we encrypt the junk, we'll encrypt all the data, and
+        # all the initial padding, but not the RSA-encrypted part.
+        #    This is equal to - 256
+        #                     + sum(size[current]....size[last])
+        #                     + paddingLen
+        #    This simplifies to:
+        #startIdx = paddingLen - 256 + totalSize - len(lastJunk)
+        startIdx = HEADER_LEN - ENC_SUBHEADER_LEN - len(lastJunk)
         nextJunk = Crypto.ctr_crypt(nextJunk, headerKey, startIdx)
         junkSeen.append(nextJunk)
 
     # We start with the padding.
-    header = paddingPRNG.getBytes(HEADER_LEN - totalSize*128)
+    header = paddingPRNG.getBytes(paddingLen)
 
     # Now, we build the subheaders, iterating through the nodes backwards.
     for i in range(len(path)-1, -1, -1):
@@ -490,12 +521,34 @@ def _buildHeader(path,secrets,exitType,exitInfo,paddingPRNG):
                             None, #placeholder for as-yet-uncalculated digest
                             rt, ri)
 
-        extHeaders = "".join(subhead.getExtraBlocks())
-        rest = Crypto.ctr_crypt(extHeaders+header, headerKeys[i])
-        subhead.digest = Crypto.sha1(rest+junkSeen[i])
+        # Do we need to include some of the remaining header in the
+        # RSA-encrypted portion?
+        underflowLength = subhead.getUnderflowLength()
+        if underflowLength > 0:
+            underflow = header[:underflowLength]
+            header = header[underflowLength:]
+        else:
+            underflow = ""
+
+        # Do we need to spill some of the routing info out from the
+        # RSA-encrypted portion?  If so, prepend it.
+        header = subhead.getOverflow() + header
+
+        # Encrypt the symmetrically encrypted part of the header
+        header = Crypto.ctr_crypt(header, headerKeys[i])
+
+        # What digest will the next server see?
+        subhead.digest = Crypto.sha1(header+junkSeen[i])
+
+        # Encrypt the subheader, plus whatever portion of the previous header
+        # underflows, into 'esh'.
         pubkey = path[i].getPacketKey()
-        esh = Crypto.pk_encrypt(subhead.pack(), pubkey)
-        header = esh + rest
+        rsaPart = subhead.pack() + underflow
+        esh = Crypto.pk_encrypt(rsaPart, pubkey)
+
+        # Concatentate the asymmetric and symmetric parts, to get the next
+        # header.
+        header = esh + header
 
     return header
 
@@ -596,121 +649,44 @@ def _decodePayloadImpl(payload):
 
     # Uncompress the body.
     contents = payload.getContents()
-    # FFFF - We should make this rule configurable.
+    # If the payload would expand to be more than 20K long, and the
+    # compression factor is greater than 20, we warn of a possible zlib
+    # bomb.
     maxLen = max(20*1024, 20*len(contents))
 
-    return uncompressData(payload.getContents(), maxLength=maxLen)
+    return uncompressData(contents, maxLength=maxLen)
 
 def _checkPayload(payload):
     'Return true iff the hash on the given payload seems valid'
     return payload[2:22] == Crypto.sha1(payload[22:])
 
-#----------------------------------------------------------------------
-# COMPRESSION FOR PAYLOADS
+def _getRouting(path, exitType, exitInfo):
+    """Given a list of ServerInfo, and a final exitType and exitInfo,
+       return a 3-tuple of:
+           1) A list of routingtype/routinginfo tuples for the header
+           2) The size (in bytes) added to the header in order to
+              route to each of the nodes
+           3) Minimum size (in bytes) needed for the header.
 
-# Global: contains 0 if we haven't validated zlib; 1 if we have, and 0.5
-#    if we're in the middle of validation.
-_ZLIB_LIBRARY_OK = 0
+       Raises MixError if the routing info is too big to fit into a single
+       header. """
+    # Construct a list 'routing' of exitType, exitInfo.
+    routing = [ (FWD_TYPE, node.getRoutingInfo().pack()) for
+                node in path[1:] ]
+    routing.append((exitType, exitInfo))
 
-def compressData(payload):
-    """Given a string 'payload', compress it with the 'deflate' method
-       as specified in the remailer spec and in RFC1951."""
-    if not _ZLIB_LIBRARY_OK:
-        _validateZlib()
+    # sizes[i] is number of bytes added to header for subheader i.
+    sizes = [ len(ri)+OAEP_OVERHEAD+MIN_SUBHEADER_LEN for _, ri in routing]
 
-    # Don't change any of these options; if different Mixminion clients
-    # compress their data differently, an adversary could distinguish
-    # messages generated by them.
-    zobj = zlib.compressobj(zlib.Z_BEST_COMPRESSION, zlib.DEFLATED,
-                            zlib.MAX_WBITS, zlib.DEF_MEM_LEVEL,
-                            zlib.Z_DEFAULT_STRATEGY)
-    s1 = zobj.compress(payload)
-    s2 = zobj.flush()
-    s = s1 + s2
+    # totalSize is the total number of bytes needed for header
+    totalSize = reduce(operator.add, sizes)
+    if totalSize > HEADER_LEN:
+        raise MixError("Routing info won't fit in header")
 
-    # Now we check the 2 bytes of zlib header.  Strictly speaking,
-    # these are irrelevant, as are the 4 bytes of adler-32 checksum at
-    # the end.  Still, we can afford 6 bytes per payload, and
-    # reconstructing the checksum to keep zlib happy is a bit of a pain.
-    assert s[0] == '\x78' # deflate, 32K window
-    assert s[1] == '\xda' # no dict, max compression
-    return s
+    padding = HEADER_LEN-totalSize
+    # We can't underflow from the last header.  That means we *must* have
+    # enough space to pad the last routinginfo out to a public key size.
+    if padding+sizes[-1] < ENC_SUBHEADER_LEN:
+        raise MixError("Routing info won't fit in header")
 
-class CompressedDataTooLong(MixError):
-    """Exception: raised when try to uncompress data that turns out to be
-       longer than we had expected."""
-    pass
-
-def uncompressData(payload, maxLength=None):
-    """Uncompress a string 'payload'; raise ParseError if it is not
-       valid compressed data.  If the expanded data is longer than
-       maxLength, we raise 'CompressedDataTooLong'."""
-
-    if len(payload) < 6 or payload[0:2] != '\x78\xDA':
-        raise ParseError("Invalid zlib header")
-
-    # This code is necessary because versions of Python before 2.2 didn't
-    # support limited-size versions of zlib.decompress.  We use a helper
-    # function helpfully submitted by Zooko.
-    if sys.version_info[:3] < (2,2,0) and maxLength is not None:
-        try:
-            return zlibutil.safe_zlib_decompress_to_retval(payload,
-                                                      maxLength,
-                                                  max(maxLength*3, 1<<20))
-        except zlibutil.TooBigError:
-            raise CompressedDataTooLong()
-        except zlibutil.DecompressError, e:
-            raise ParseError("Error in compressed data: %s"%e)
-
-    try:
-        # We can't just call zlib.decompress(payload), since we may
-        # want to limit the output size.
-
-        zobj = zlib.decompressobj(zlib.MAX_WBITS)
-        # Decompress the payload.
-        if maxLength is None:
-            d = zobj.decompress(payload)
-        else:
-            # If we _do_ have Python 2.2, this is the easy way to do it.  It
-            # also uses less RAM in the failing case.
-            d = zobj.decompress(payload, maxLength)
-            if zobj.unconsumed_tail:
-                raise CompressedDataTooLong()
-
-        # Get any leftovers, which shouldn't exist.
-        nil = zobj.flush()
-        if nil != '':
-            raise ParseError("Error in compressed data")
-        return d
-    except zlib.error:
-        raise ParseError("Error in compressed data")
-
-def _validateZlib():
-    """Internal function:  Make sure that zlib is a recognized version, and
-       that it compresses things as expected.  (This check is important,
-       because using a zlib version that compressed differently from zlib1.1.4
-       would make senders partitionable by payload compression.)
-    """
-    global _ZLIB_LIBRARY_OK
-    ver = getattr(zlib, "ZLIB_VERSION")
-    if ver and ver < "1.1.2":
-        raise MixFatalError("Zlib version %s is not supported"%ver)
-
-    _ZLIB_LIBRARY_OK = 0.5
-    if ver in ("1.1.2", "1.1.3", "1.1.4"):
-        _ZLIB_LIBRARY_OK = 1
-        return
-
-    LOG.warn("Unrecognized zlib version: %r. Spot-checking output", ver)
-    # This test is inadequate, but it _might_ catch future incompatible
-    # changes.
-    _ZLIB_LIBRARY_OK = 0.5
-    good = '\x78\xda\xed\xc6A\x11\x00 \x08\x00\xb0l\xd4\xf0\x87\x02\xf6o'+\
-           '`\x0e\xef\xb6\xd7r\xed\x88S=7\xcd\xcc\xcc\xcc\xcc\xcc\xcc'+\
-           '\xcc\xcc\xcc\xcc\xcc\xcc\xcc\xcc\xcc\xcc\xcc\xcc\xbe\xdd\x03'+\
-           'q\x8d\n\x93'
-    if compressData("aZbAAcdefg"*1000) == good:
-        _ZLIB_LIBRARY_OK = 1
-    else:
-        _ZLIB_LIBRARY_OK = 0
-        raise MixFatalError("Zlib output not as exected.")
+    return routing, sizes, totalSize
