@@ -1,30 +1,44 @@
 # Copyright 2002-2003 Nick Mathewson.  See LICENSE for licensing information.
-# $Id: Common.py,v 1.48 2003/01/07 04:13:08 nickm Exp $
+# $Id: Common.py,v 1.89 2003/06/06 06:04:57 nickm Exp $
 
 """mixminion.Common
 
    Common functionality and utility code for Mixminion"""
 
-__all__ = [ 'IntervalSet', 'LOG', 'LogStream', 'MixError', 'MixFatalError',
-            'MixProtocolError', 'ceilDiv', 'checkPrivateDir',
-            'createPrivateDir', 'floorDiv', 'formatBase64', 'formatDate',
-            'formatFnameTime', 'formatTime', 'installSignalHandlers',
-            'isSMTPMailbox', 'onReset', 'onTerminate', 'openUnique',
-            'previousMidnight', 'readPossiblyGzippedFile', 'secureDelete',
-            'stringContains', 'waitForChildren' ]
+__all__ = [ 'IntervalSet', 'Lockfile', 'LOG', 'LogStream', 'MixError',
+            'MixFatalError', 'MixProtocolError', 'UIError', 'UsageError',
+            'armorText', 'ceilDiv', 'checkPrivateDir', 'checkPrivateFile',
+            'createPrivateDir', 'encodeBase64', 'floorDiv', 'formatBase64',
+            'formatDate', 'formatFnameTime', 'formatTime',
+            'installSIGCHLDHandler', 'isSMTPMailbox', 'openUnique',
+            'previousMidnight', 'readFile', 'readPickled',
+            'readPossiblyGzippedFile', 'secureDelete', 'stringContains',
+            'succeedingMidnight', 'tryUnlink', 'unarmorText',
+            'waitForChildren', 'writeFile', 'writePickled' ]
 
-import base64
+import binascii
 import bisect
 import calendar
+import cPickle
+import errno
+import fcntl
+import grp
 import gzip
 import os
+import pwd
 import re
 import signal
 import stat
 import statvfs
 import sys
+import threading
 import time
 import traceback
+# Imported here so we can get it in mixminion.server without being shadowed
+# by the old Queue.py file.
+from Queue import Queue
+MessageQueue = Queue
+del Queue
 
 from types import StringType
 
@@ -40,6 +54,29 @@ class MixProtocolError(MixError):
     """Exception class for MMTP protocol violations"""
     pass
 
+class MixProtocolReject(MixProtocolError):
+    """Exception class for server-rejected packets."""
+    pass
+
+class MixProtocolBadAuth(MixProtocolError):
+    """Exception class for failed authentication to a server."""
+    pass
+
+class UIError(MixError):
+    """Exception raised for an error that should be reported to the user,
+       not dumped as a stack trace."""
+    def dump(self):
+        if str(self): print >>sys.stderr, "ERROR:", str(self)
+    def dumpAndExit(self):
+        self.dump()
+        sys.exit(1)
+
+class UsageError(UIError):
+    """Exception raised for an error that should be reported to the user
+       along with a usage mesage.
+    """
+    pass
+
 #----------------------------------------------------------------------
 # Portability to future Python versions
 
@@ -97,11 +134,165 @@ def stripSpace(s, space=" \t\v\n"):
     """Remove all whitespace from s."""
     return s.translate(_ALLCHARS, space)
 
-def formatBase64(s):
-    """Convert 's' to a one-line base-64 representation."""
-    return base64.encodestring(s).replace("\n", "")
+if sys.version_info[0:3] >= (2,1,0):
+    def formatBase64(s):
+        """Convert 's' to a one-line base-64 representation."""
+        return binascii.b2a_base64(s).strip()
+else:
+    # Python 2.0 didin't allow a binascii to return more than one line.
+    def formatBase64(s):
+        """Convert 's' to a one-line base-64 representation."""
+        return encodeBase64(s, 64, 1)
+
+def encodeBase64(s, lineWidth=64, oneline=0):
+    """Convert 's' to a multiline base-64 representation.  Improves upon
+       base64.encodestring by having a variable line width.  Implementation
+       is based upon that function.
+    """
+    pieces = []
+    bytesPerLine = floorDiv(lineWidth, 4) * 3
+    for i in xrange(0, len(s), bytesPerLine):
+        chunk = s[i:i+bytesPerLine]
+        pieces.append(binascii.b2a_base64(chunk))
+    if oneline:
+        return "".join([ s.strip() for s in pieces ])
+    else:
+        return "".join(pieces)
 
 #----------------------------------------------------------------------
+# Functions to generate and parse OpenPGP-style ASCII armor
+
+# Matches a line that needs to be ascii-armored in plaintext mode.
+DASH_ARMOR_RE = re.compile('^-', re.M)
+
+def armorText(s, type, headers=(), base64=1):
+    """Given a string (s), string holding a message type (type), and a
+       list of key-value pairs for headers, generates an OpenPGP-style
+       ASCII-armored message of type 'type', with contents 's' and
+       headers 'header'.
+       
+       If base64 is false, uses cleartext armor."""
+    result = []
+    result.append("-----BEGIN %s-----\n" %type)
+    for k,v in headers:
+        result.append("%s: %s\n" %(k,v))
+    result.append("\n")
+    if base64:
+        result.append(encodeBase64(s, lineWidth=64))
+    else:
+        result.append(DASH_ARMOR_RE.sub('- -', s))
+    if not result[-1].endswith("\n"):
+        result.append("\n")
+    result.append("-----END %s-----\n" %type)
+
+    return "".join(result)
+
+# Matches a begin line.
+BEGIN_LINE_RE = re.compile(r'^-----BEGIN ([^-]+)-----[ \t]*$',re.M)
+
+# Matches a header line.
+ARMOR_KV_RE = re.compile(r'([^:\s]+): ([^\n]+)')
+def unarmorText(s, findTypes, base64=1, base64fn=None):
+    """Parse a list of OpenPGP-style ASCII-armored messages from 's',
+       and return a list of (type, headers, body) tuples, where 'headers'
+       is a list of key,val tuples.
+
+       s -- the string to parse.
+       findTypes -- a list of types to search for; others are ignored.
+       base64 -- if false, we do cleartext armor.
+       base64fn -- if provided, called with (type, headers) to tell whether
+          we do cleartext armor.
+    """
+    result = []
+    
+    while 1:
+        tp = None
+        fields = []
+        value = None
+
+        mBegin = BEGIN_LINE_RE.search(s)
+        if not mBegin:
+            return result
+
+        tp = mBegin.group(1)
+        endPat = r"^-----END %s-----[ \t]*$" % tp
+
+        endRE = re.compile(endPat, re.M)
+        mEnd = endRE.search(s, mBegin.start())
+        if not mEnd:
+            raise ValueError("Couldn't find end line for '%s'"%tp.lower())
+
+        if tp not in findTypes:
+            s = s[mEnd.end()+1:]
+            continue
+
+        idx = mBegin.end()+1
+        endIdx = mEnd.start()
+
+        assert s[idx-1] == s[endIdx-1] == '\n'
+        while idx < endIdx:
+            nl = s.index("\n", idx, endIdx)
+            line = s[idx:nl]
+            idx = nl+1
+            if ":" in line:
+                m = ARMOR_KV_RE.match(line)
+                if not m:
+                    raise ValueError("Bad header for '%s'"%tp.lower())
+                fields.append((m.group(1), m.group(2)))
+            elif line.strip() == '':
+                break
+
+        if base64fn:
+            base64 = base64fn(tp,fields)
+
+        if base64:
+            try:
+                value = binascii.a2b_base64(s[idx:endIdx])
+            except (TypeError, binascii.Incomplete, binascii.Error), e:
+                raise ValueError(str(e))
+        else:
+            v = s[idx:endIdx].split("\n")
+            for i in xrange(len(v)):
+                if v[i].startswith("- "):
+                    v[i] = v[i][2:]
+            value = "\n".join(v)
+
+        result.append((tp, fields, value))
+        
+        s = s[mEnd.end()+1:]
+
+    raise MixFatalError("Unreachable code somehow reached.")
+
+#----------------------------------------------------------------------
+def checkPrivateFile(fn, fix=1):
+    """Checks whether f is a file owned by this uid, set to mode 0600 or
+       0700, and all its parents pass checkPrivateDir.  Raises MixFatalError
+       if the assumtions are not met; else return None.  If 'fix' is true,
+       repair permissions on the file rather than raising MixFatalError."""
+    parent, _ = os.path.split(fn)
+    checkPrivateDir(parent)
+    try:
+        st = os.stat(fn)
+    except OSError, e:
+        if e.errno == errno.EEXIST:
+            raise MixFatalError("Nonexistant file %s" % fn)
+        else:
+            raise MixFatalError("Could't stat file %s: %s" % (fn, e))
+    if not st:
+        raise MixFatalError("Nonexistant file %s" % fn)
+    if not os.path.isfile(fn):
+        raise MixFatalError("%s is not a regular file" % fn)
+    me = os.getuid()
+    if st[stat.ST_UID] != me:
+        raise MixFatalError("File %s must have owner %s" % (fn, me))
+    mode = st[stat.ST_MODE] & 0777
+    if mode not in (0700, 0600):
+        if not fix:
+            raise MixFatalError("Bad mode %o on file %s" % (mode & 0777, fn))
+        newmode = {0:0600,0100:0700}[(mode & 0100)]
+        LOG.warn("Repairing permissions on file %s" % fn)
+        os.chmod(fn, newmode)
+
 def createPrivateDir(d, nocreate=0):
     """Create a directory, and all parent directories, checking permissions
        as we go along.  All superdirectories must be owned by root or us."""
@@ -110,19 +301,22 @@ def createPrivateDir(d, nocreate=0):
             raise MixFatalError("Nonexistent directory %s" % d)
         try:
             os.makedirs(d, 0700)
-        except OSError:
-            raise MixFatalError("Unable to create directory %s" % d)
+        except OSError, e:
+            raise MixFatalError("Unable to create directory %s: %s" % d, e)
 
     checkPrivateDir(d)
 
 _WARNED_DIRECTORIES = {}
+_VALID_DIRECTORIES = {}
+_TRUSTED_UIDS = [ 0 ]
 
 def checkPrivateDir(d, recurse=1):
-    """Return true iff d is a directory owned by this uid, set to mode
+    """Check whether d is a directory owned by this uid, set to mode
        0700. All of d's parents must not be writable or owned by anybody but
        this uid and uid 0.  If any of these conditions are unmet, raise
        MixFatalErrror.  Otherwise, return None."""
     me = os.getuid()
+    trusted_uids = _TRUSTED_UIDS + [ me ]
 
     if not os.path.isabs(d):
         d = os.path.abspath(d)
@@ -146,6 +340,8 @@ def checkPrivateDir(d, recurse=1):
     # Check permissions on parents.
     while 1:
         parent = os.path.split(d)[0]
+        if _VALID_DIRECTORIES.has_key(parent):
+            return
         if parent == d:
             return
         d = parent
@@ -153,18 +349,142 @@ def checkPrivateDir(d, recurse=1):
         st = os.stat(d)
         mode = st[stat.ST_MODE]
         owner = st[stat.ST_UID]
-        if owner not in (0, me):
+        if owner not in trusted_uids:
             raise MixFatalError("Bad owner (uid=%s) on directory %s"
                                 % (owner, d))
         if (mode & 02) and not (mode & stat.S_ISVTX):
-            raise MixFatalError("Bad mode (%o) on directory %s" %(mode, d))
+            raise MixFatalError("Bad mode (%o) on directory %s" %
+                                (mode&0777, d))
 
         if (mode & 020) and not (mode & stat.S_ISVTX):
             # FFFF We may want to give an even stronger error here.
             if not _WARNED_DIRECTORIES.has_key(d):
-                LOG.warn("Iffy mode %o on directory %s (Writable by gid %s)",
-                         mode, d, st[stat.ST_GID])
+                group = grp.getgrgid(st[stat.ST_GID])[0]
+                LOG.warn("Directory %s is writable by group %s (mode %o)",
+                         d, group, mode&0777)
             _WARNED_DIRECTORIES[d] = 1
+
+def configureTrustedUsers(config):
+    users = config['Host']['TrustedUser']
+    if not users:
+        return
+
+    for u in users:
+        u = u.strip()
+        try:
+            ent = pwd.getpwnam(u)
+        except KeyError:
+            LOG.warn("TrustedUser: No such user as %s", u)
+            continue
+
+        uid = ent[2]
+        _TRUSTED_UIDS.append(uid)
+
+#----------------------------------------------------------------------
+# File helpers
+class AtomicFile:
+    """Wrapper around open/write/rename to encapsulate writing to a temporary
+       file, then moving to the final filename on close.
+
+       NOTE 1: If you don't call 'close' or 'discard' on this object,
+       the temporary file it creates will stay around indefinitely.
+
+       NOTE 2: If multiple AtomicFiles are active for the same destination
+       file, the last one to close will win, and results for the others will
+       not be visible.
+       """
+    def __init__(self, fname, mode='w'):
+        self.fname = fname
+        self.tmpname = fname + ".tmp"
+        self.f, self.tmpname = openUnique(self.tmpname)
+
+    def write(self, s):
+        self.f.write(s)
+
+    def close(self):
+        """Close the underlying file and replace the destination file."""
+        os.rename(self.tmpname, self.fname)
+        self.f.close()
+        self.f = None
+
+    def discard(self):
+        """Discard changes to the temporary file."""
+        self.f.close()
+        os.unlink(self.tmpname)
+        self.f = None
+
+    def __del__(self):
+        if self.f:
+            LOG.error("Atomic file not closed/discarded: %s",self.tmpname)
+
+def readFile(fn, binary=0):
+    """Return the contents of the file named <fn>."""
+    f = open(fn, ['r', 'rb'][binary])
+    try:
+        return f.read()
+    finally:
+        f.close()
+
+def writeFile(fn, contents, mode=0600, binary=0):
+    """Atomically write a string <contents> into a file <file> with mode
+       <mode>.  If <binary>, binary mode will be used.
+
+       If the file exists, it will be replaced.
+
+       If two processes attempt to writeFile the same file at once,
+       the one finishing last wins.
+       """
+    tmpname = fn+".tmp"
+    f, tmpname = openUnique(tmpname, ['w','wb'][binary], mode)
+    try:
+        try:
+            f.write(contents)
+        finally:
+            f.close()
+    except:
+        if os.path.exists(tmpname): os.unlink(tmpname)
+        raise
+
+    os.rename(tmpname, fn)
+
+def readPickled(fn):
+    """Given the name of a file containing a pickled object, return the pickled
+       object."""
+    f = open(fn, 'rb')
+    try:
+        return cPickle.load(f)
+    finally:
+        f.close()
+
+def writePickled(fn, obj, mode=0600):
+    """Given a filename and an object to be pickled, pickles the object into
+       a temporary file, then replaces the file with the temporary file.
+    """
+    tmpname = fn + ".tmp"
+    f, tmpname = openUnique(tmpname, 'wb', mode)
+    try:
+        try:
+            cPickle.dump(obj, f, 1)
+        finally:
+            f.close()
+    except:
+        if os.path.exists(tmpname): os.unlink(tmpname)
+        raise
+
+    os.rename(tmpname, fn)
+
+def tryUnlink(fname):
+    """Try to remove the file named fname.  If the file is erased, return 1.
+       If the file didn't exist in the first place, return 0.  Otherwise
+       propagate an exception."""
+    try:
+        os.unlink(fname)
+        return 1
+    except OSError, e:
+        if e.errno == errno.ENOENT:
+            return 0
+        raise
+
 #----------------------------------------------------------------------
 # Secure filesystem operations.
 
@@ -182,7 +502,7 @@ def configureShredCommand(conf):
     global _SHRED_OPTS
     cmd, opts = None, None
     if conf is not None:
-        val = conf['Host'].get('ShredCommand', None)
+        val = conf['Host'].get('ShredCommand')
         if val is not None:
             cmd, opts = val
 
@@ -196,30 +516,33 @@ def configureShredCommand(conf):
     _SHRED_CMD, _SHRED_OPTS = cmd, opts
 
 
-# Size of a block on the filesystem we're overwriting on; If zero, we need
-# to determine it.
-_BLKSIZE = 0
-# A string of _BLKSIZE zeros
-_NILSTR = None
+# Map from parent directory to blocksize.  We only overwrite files in a few
+# locations, so this should be safe.
+_BLKSIZEMAP = {}
+# A string of max(_BLKSIZEMAP.values()) zeros
+_NILSTR = ""
 def _overwriteFile(f):
     """Overwrite f with zeros, rounding up to the nearest block.  This is
        used as the default implementation of secureDelete."""
-    global _BLKSIZE
     global _NILSTR
-    if not _BLKSIZE:
-        #???? this assumes that all filesystems we are using have the same
-        #???? block size.
+    parent = os.path.split(f)[0]
+    try:
+        sz = _BLKSIZEMAP[parent]
+    except KeyError:
         if hasattr(os, 'statvfs'):
-            _BLKSIZE = os.statvfs(f)[statvfs.F_BSIZE]
+            sz = os.statvfs(f)[statvfs.F_BSIZE]
         else:
-            _BLKSIZE = 8192 # ???? Safe guess?
-        _NILSTR = '\x00' * _BLKSIZE
+            sz = 8192 # Should be a safe guess? (????)
+        _BLKSIZEMAP[parent] = sz
+        if sz > len(_NILSTR):
+            _NILSTR = '\x00' * sz
+    nil = _NILSTR[:sz]
     fd = os.open(f, os.O_WRONLY)
     try:
         size = os.fstat(fd)[stat.ST_SIZE]
-        blocks = ceilDiv(size, _BLKSIZE)
+        blocks = ceilDiv(size, sz)
         for _ in xrange(blocks):
-            os.write(fd, _NILSTR)
+            os.write(fd, nil)
     finally:
         os.close(fd)
 
@@ -269,15 +592,17 @@ def secureDelete(fnames, blocking=0):
             os.unlink(f)
         return None
 
-    if blocking:
-        mode = os.P_WAIT
-    else:
-        mode = os.P_NOWAIT
-
     # Some systems are unhappy when you call them with too many options.
     for i in xrange(0, len(fnames), 250-len(_SHRED_OPTS)):
         files = fnames[i:i+250-len(_SHRED_OPTS)]
-        os.spawnl(mode, _SHRED_CMD, _SHRED_CMD, *(_SHRED_OPTS+files))
+        pid = os.spawnl(os.P_NOWAIT,
+                        _SHRED_CMD, _SHRED_CMD, *(_SHRED_OPTS+files))
+        if blocking:
+            try:
+                os.waitpid(pid, 0)
+            except OSError:
+                # sigchild handler might get to the pid first.
+                pass
 
 #----------------------------------------------------------------------
 # Logging
@@ -287,7 +612,9 @@ def secureDelete(fnames, blocking=0):
 
 def _logtime():
     'Helper function.  Returns current local time formatted for log.'
-    return time.strftime("%b %d %H:%M:%S", time.localtime(time.time()))
+    t = time.time()
+    return "%s.%03d"%(time.strftime("%b %d %H:%M:%S", time.localtime(t)),
+                      (t*1000)%1000)
 
 class _FileLogHandler:
     """Helper class for logging.  Represents a file on disk, and allows the
@@ -310,9 +637,9 @@ class _FileLogHandler:
             if not os.path.exists(parent):
                 createPrivateDir(parent)
             self.file = open(self.fname, 'a')
-        except OSError:
+        except OSError, e:
             self.file = None
-            raise MixError("Unable to open log file %r"%self.fname)
+            raise MixError("Unable to open log file %r: %s"%self.fname, e)
     def close(self):
         "Close the underlying file"
         self.file.close()
@@ -369,18 +696,22 @@ class Log:
            minSeverity, and sends its output to stderr."""
         self.configure(None)
         self.setMinSeverity(minSeverity)
+        self.__lock = threading.Lock()
 
-    def configure(self, config):
+    def configure(self, config, keepStderr=0):
         """Set up this Log object based on a ServerConfig or ClientConfig
-           object"""
+           object
 
+           If keepStderr is true, do not silence the console log, regardless
+           of the value of 'Daemon' or 'EchoMessages'.
+           """
         self.handlers = []
         if config == None or not config.has_section("Server"):
             self.setMinSeverity("WARN")
             self.addHandler(_ConsoleLogHandler(sys.stderr))
         else:
             self.setMinSeverity(config['Server'].get('LogLevel', "WARN"))
-            logfile = config['Server'].get('LogFile',None)
+            logfile = config['Server'].get('LogFile')
             if logfile is None:
                 homedir = config['Server']['Homedir']
                 if homedir:
@@ -391,6 +722,8 @@ class Log:
                     self.addHandler(_FileLogHandler(logfile))
                 except MixError, e:
                     self.error(str(e))
+                if keepStderr:
+                    return
                 if (config['Server'].get('Daemon',0) or
                     not config['Server'].get('EchoMessages',0)):
                     print "Silencing the console log; look in %s instead"%(
@@ -452,8 +785,12 @@ class Log:
         if _SEVERITIES.get(severity, 100) < self.severity:
             return
 
-        for h in self.handlers:
-            h.write(severity, m)
+        self.__lock.acquire()
+        try:
+            for h in self.handlers:
+                h.write(severity, m)
+        finally:
+            self.__lock.release()
 
     def trace(self, message, *args):
         "Write a trace (hyperverbose) message to the log"
@@ -514,11 +851,35 @@ class LogStream:
        We don't actually want to use these; but they keep us from dropping
        prints on the floor.
        """
+    # Fields:
+    #    name -- The name of this stream
+    #    severity -- The severity of log messages to generate.
+    #    buf -- A list of strings that have been written to this stream since
+    #        we most recently flushed the buffer to the log.
     def __init__(self, name, severity):
         self.name = name
         self.severity = severity
+        self.buf = []
     def write(self, s):
-        LOG.log(self.severity, "->%s: %s", self.name, s)
+        # This is inefficient, but we don't actually use this class if we can
+        # avoid it.  The basic idea is to generate a call to Log.log for every
+        # newline.  (This buffering is particularly necessary because typical
+        # uses of sys.stdout/stderr generate multiple calls to file.write per
+        # line.)
+        if "\n" not in s:
+            self.buf.append(s)
+            return
+
+        while "\n" in s:
+            idx = s.index("\n")
+            line = "%s%s" %("".join(self.buf), s[:idx])
+            LOG.log(self.severity, "->%s: %s", self.name, line)
+            del self.buf[:]
+            s = s[idx+1:]
+
+        if s:
+            self.buf.append(s)
+
     def flush(self): pass
     def close(self): pass
 
@@ -531,6 +892,21 @@ def previousMidnight(when):
     yyyy,MM,dd = time.gmtime(when)[0:3]
     return calendar.timegm((yyyy,MM,dd,0,0,0,0,0,0))
 
+def succeedingMidnight(when):
+    """Given a time_t 'when', return the smallest time_t > when that falls
+       on midnight, GMT."""
+    yyyy,MM,dd = time.gmtime(when)[0:3]
+    try:
+        return calendar.timegm((yyyy,MM,dd+1,0,0,0,0,0,0))
+    except ValueError:
+        # Python 2.3 seems to raise ValueError when dd is the last day of the
+        # month.
+        pass
+    if MM < 12:
+        return calendar.timegm((yyyy,MM+1,1,0,0,0,0,0,0))
+    else:
+        return calendar.timegm((yyyy+1,1,1,0,0,0,0,0,0))
+
 def formatTime(when,localtime=0):
     """Given a time in seconds since the epoch, returns a time value in the
        format used by server descriptors (YYYY/MM/DD HH:MM:SS) in GMT"""
@@ -554,6 +930,64 @@ def formatFnameTime(when=None):
         when = time.time()
     return time.strftime("%Y%m%d%H%M%S", time.localtime(when))
 
+#----------------------------------------------------------------------
+class Duration:
+    """A Duration is a number of time units, such as '1.5 seconds' or
+       '2 weeks'.  Durations are stored internally as a number of seconds.
+    """
+    ## Fields:
+    # seconds: the number of seconds in this duration
+    # unitName: the name of the units comprising this duration.
+    # nUnits: the number of units in this duration
+    def __init__(self, seconds, unitName=None, nUnits=None):
+        """Initialize a new Duration with a given number of seconds."""
+        self.seconds = seconds
+        if unitName:
+            self.unitName = unitName
+            self.nUnits = nUnits
+        else:
+            self.unitName = "second"
+            self.nUnits = seconds
+
+    def __str__(self):
+        s = ""
+        if self.nUnits != 1:
+            s = "s"
+        return "%s %s%s" % (self.nUnits, self.unitName, s)
+
+    def __repr__(self):
+        return "Duration(%r, %r, %r)" % (self.seconds, self.unitName,
+                                         self.nUnits)
+
+    def __float__(self):
+        """Return the number of seconds in this duration"""
+        return self.seconds
+
+    def __int__(self):
+        """Return the number of seconds in this duration"""
+        return int(self.seconds)
+
+    def getSeconds(self):
+        """Return the number of seconds in this duration"""
+        return self.seconds
+
+    def reduce(self):
+        """Change the representation of this object to its clearest form"""
+        s = self.seconds
+        for n,u in [(60*60*24*365,'year'),
+                    (60*60*24*30, 'month'),
+                    (60*60*24*7,  'week'),
+                    (60*60*24,    'day'),
+                    (60*60,       'hour'),
+                    (60,          'minute')]:
+            if s % n == 0:
+                self.nUnits = floorDiv(s,n)
+                self.unitName = u
+                return
+        self.nUnits = s
+        self.unitName = 'second'
+        return self
+
 #----------------------------------------------------------------------
 # InteralSet
 
@@ -742,30 +1176,18 @@ def isSMTPMailbox(s):
 #----------------------------------------------------------------------
 # Signal handling
 
-# List of 0-argument functions to call on SIGHUP
-resetHooks = []
-
-# List of 0-argument functions to call on SIGTERM
-terminateHooks = []
-
-def onReset(fn):
-    """Given a 0-argument function fn, cause fn to be invoked when
-       this process next receives a SIGHUP."""
-    resetHooks.append(fn)
-
-def onTerminate(fn):
-    """Given a 0-argument function fn, cause fn to be invoked when
-       this process next receives a SIGTERM."""
-    terminateHooks.append(fn)
-
-def waitForChildren(onceOnly=0):
+def waitForChildren(onceOnly=0, blocking=1):
     """Wait until all subprocesses have finished.  Useful for testing."""
+    if blocking:
+        options = 0
+    else:
+        options = os.WNOHANG
     while 1:
         try:
-            # FFFF This won't work on Windows.  What to do?
-            pid, status = os.waitpid(0, 0)
+            # WIN32 This won't work on Windows.  What to do?
+            pid, status = os.waitpid(-1, options)
         except OSError, e:
-            break
+            return
         except e:
             print e, repr(e), e.__class__
         if onceOnly:
@@ -775,12 +1197,12 @@ def _sigChldHandler(signal_num, _):
     '''(Signal handler for SIGCHLD)'''
     # Because of the peculiarities of Python's signal handling logic, I
     # believe we need to re-register ourself.
-    signal.signal(signal_num, _sigChldHandler)
+    signal.signal(signal.SIGCHLD, _sigChldHandler)
 
     while 1:
         try:
-            # This waitpid call won't work on Windows.  What to do?
-            pid, status = os.waitpid(0, os.WNOHANG)
+            # WIN32 This waitpid call won't work on Windows.  What to do?
+            pid, status = os.waitpid(-1, os.WNOHANG)
             if pid == 0:
                 break
         except OSError:
@@ -789,27 +1211,9 @@ def _sigChldHandler(signal_num, _):
     #outcome, core, sig = status & 0xff00, status & 0x0080, status & 0x7f
     # FFFF Log if outcome wasn't as expected.
 
-def _sigHandler(signal_num, _):
-    '''(Signal handler for SIGTERM and SIGHUP)'''
-    signal.signal(signal_num, _sigHandler)
-    if signal_num == signal.SIGTERM:
-        for hook in terminateHooks:
-            hook()
-        sys.exit(1)
-    else:
-        for hook in resetHooks:
-            hook()
-
-def installSignalHandlers(child=1,hup=1,term=1):
-    '''Register signal handlers for this process.  If 'child', registers
-       a handler for SIGCHLD.  If 'hup', registers a handler for SIGHUP.
-       If 'term', registes a handler for SIGTERM.'''
-    if child:
-        signal.signal(signal.SIGCHLD, _sigChldHandler)
-    if hup:
-        signal.signal(signal.SIGHUP, _sigHandler)
-    if term:
-        signal.signal(signal.SIGTERM, _sigHandler)
+def installSIGCHLDHandler():
+    '''Register sigchld handler for this process.'''
+    signal.signal(signal.SIGCHLD, _sigChldHandler)
 
 #----------------------------------------------------------------------
 # File helpers.
@@ -828,7 +1232,7 @@ def readPossiblyGzippedFile(fname, mode='r'):
         if f is not None:
             f.close()
 
-def openUnique(fname, mode='w'):
+def openUnique(fname, mode='w', perms=0600):
     """Helper function. Returns a file open for writing into the file named
        'fname'.  If fname already exists, opens 'fname.1' or 'fname.2' or
        'fname.3' or so on."""
@@ -836,9 +1240,75 @@ def openUnique(fname, mode='w'):
     idx = 0
     while 1:
         try:
-            fd = os.open(fname, os.O_WRONLY|os.O_CREAT|os.O_EXCL, 0600)
+            fd = os.open(fname, os.O_WRONLY|os.O_CREAT|os.O_EXCL, perms)
             return os.fdopen(fd, mode), fname
-        except OSError:
-            pass
+        except OSError, e:
+            if e.errno != errno.EEXIST:
+                raise
         idx += 1
         fname = os.path.join(base, "%s.%s"%(rest,idx))
+
+    raise MixFatalError("unreachable code")
+
+#----------------------------------------------------------------------
+class Lockfile:
+    """Class to implement a recursive advisory lock, using flock on a
+       'well-known' filename."""
+    ## Fields:
+    # filename--the name of the file to lock
+    # count--the recursion depth of the lock; 0 is unlocked.
+    # fd--If fd>1, a file descriptor open to 'filename'.  Otherwise, None.
+    def __init__(self, filename):
+        """Create a new Lockfile object to acquire and release a lock on
+           'filename'"""
+        self.filename = filename
+        self.count = 0
+        self.fd = None
+
+    def getContents(self):
+        """Return the contents of the lock file."""
+        return readFile(self.filename)
+
+    def acquire(self, contents="", blocking=0):
+        """Acquire this lock.  If we're acquiring the lock for the first time,
+           write 'contents' to the lockfile.  If 'blocking' is true, wait until
+           we can acquire the lock.  If 'blocking' is false, raise IOError if
+           we can't acquire the lock."""
+        if self.count > 0:
+            self.count += 1
+            return
+
+        assert self.fd is None
+        self.fd = os.open(self.filename, os.O_RDWR|os.O_CREAT, 0600)
+        try:
+            if blocking:
+                fcntl.flock(self.fd, fcntl.LOCK_EX)
+            else:
+                fcntl.flock(self.fd, fcntl.LOCK_EX|fcntl.LOCK_NB)
+            self.count += 1
+            os.write(self.fd, contents)
+        except:
+            os.close(self.fd)
+            self.fd = None
+            raise
+
+    def release(self):
+        """Release the lock."""
+        assert self.fd is not None
+        self.count -= 1
+        if self.count > 0:
+            return
+        try:
+            os.unlink(self.filename)
+        except OSError:
+            pass
+        try:
+            fcntl.flock(self.fd, fcntl.LOCK_UN)
+        except OSError:
+            pass
+        try:
+            os.close(self.fd)
+        except OSError:
+            pass
+
+        self.fd = None
