@@ -1,5 +1,5 @@
 # Copyright 2002-2003 Nick Mathewson.  See LICENSE for licensing information.
-# $Id: Config.py,v 1.33 2003/01/05 13:19:53 nickm Exp $
+# $Id: Config.py,v 1.47 2003/06/05 18:41:40 nickm Exp $
 
 """Configuration file parsers for Mixminion client and server
    configuration.
@@ -60,7 +60,8 @@ from cStringIO import StringIO
 
 import mixminion.Common
 import mixminion.Crypto
-from mixminion.Common import MixError, LOG, isPrintingAscii, stripSpace
+from mixminion.Common import MixError, LOG, ceilDiv, isPrintingAscii, \
+   stripSpace, stringContains
 
 class ConfigError(MixError):
     """Thrown when an error is found in a configuration file."""
@@ -99,8 +100,8 @@ def _parseServerMode(mode):
     return s
 
 # re to match strings of the form '9 seconds', '1 month', etc.
-_interval_re = re.compile(r'''(\d+\.?\d*|\.\d+)\s+
-                     (sec|second|min|minute|hour|day|week|mon|month|year)s?''',
+_interval_re = re.compile(r'''^(\d+\.?\d*|\.\d+)\s+
+                     (sec|second|min|minute|hour|day|week|mon|month|year)s?$''',
                           re.X)
 _seconds_per_unit = {
     'second': 1,
@@ -117,15 +118,46 @@ _seconds_per_unit = {
 _canonical_unit_names = { 'sec' : 'second', 'min': 'minute', 'mon' : 'month' }
 def _parseInterval(interval):
     """Validation function.  Converts a config value to an interval of time,
-       in the format (number of units, name of unit, total number of seconds).
-       Raises ConfigError on failure."""
+       returning a Duration object. Raises ConfigError on failure."""
     inter = interval.strip().lower()
     m = _interval_re.match(inter)
     if not m:
         raise ConfigError("Unrecognized interval %r" % inter)
-    num, unit = float(m.group(1)), m.group(2)
-    nsec = num * _seconds_per_unit[unit]
-    return num, _canonical_unit_names.get(unit,unit), nsec
+    num, unit = m.group(1), m.group(2)
+    if '.' in num:
+        num = float(num)
+    else:
+        num = int(num)
+    nsec = int(num * _seconds_per_unit[unit])
+    return mixminion.Common.Duration(nsec,
+                    _canonical_unit_names.get(unit,unit), num)
+
+def _parseIntervalList(s):
+    """Validation functions. Parse a list of comma-separated intervals
+       in the form ((every)? INTERVAL for INTERVAL)|INTERVAL into a list
+       of interval lengths in seconds."""
+    items = s.strip().lower().split(",")
+    ilist = []
+    for item in items:
+        item = item.strip()
+        if stringContains(item, " for "):
+            if item.startswith("every "):
+                item = item[6:]
+            interval, duration = item.split(" for ", 1)
+            interval = int(_parseInterval(interval))
+            duration = int(_parseInterval(duration))
+            if interval < 1:
+                raise ConfigError("Repeated interval too small in %s"%s)
+
+            ilist += [interval] * ceilDiv(duration, interval)
+        elif item.startswith("every "):
+            raise ConfigError(
+                "Bad syntax on interval %s. (Did you mean %s for X days?)",
+                item, item)
+        else:
+            interval = int(_parseInterval(item))
+            ilist.append(interval)
+    return ilist
 
 def _parseInt(integer):
     """Validation function.  Converts a config value to an int.
@@ -137,7 +169,7 @@ def _parseInt(integer):
         raise ConfigError("Expected an integer but got %r" % (integer))
 
 # Regular expression to match a dotted quad.
-_ip_re = re.compile(r'\d+\.\d+\.\d+\.\d+')
+_ip_re = re.compile(r'^\d+\.\d+\.\d+\.\d+$')
 
 def _parseIP(ip):
     """Validation function.  Converts a config value to an IP address.
@@ -158,12 +190,12 @@ def _parseIP(ip):
 
 # Regular expression to match 'address sets' as used in Allow/Deny
 # configuration lines. General format is "<IP|*> ['/'MASK] [PORT['-'PORT]]"
-_address_set_re = re.compile(r'''(\d+\.\d+\.\d+\.\d+|\*)
+_address_set_re = re.compile(r'''^(\d+\.\d+\.\d+\.\d+|\*)
                                  \s*
                                  (?:/\s*(\d+\.\d+\.\d+\.\d+))?\s*
                                  (?:(\d+)\s*
                                            (?:-\s*(\d+))?
-                                        )?''',re.X)
+                                        )?$''',re.X)
 def _parseAddressSet_allow(s, allowMode=1):
     """Validation function.  Converts an address set string of the form
        'IP/mask port-port' into a tuple of (IP, Mask, Portmin, Portmax).
@@ -256,7 +288,7 @@ def _parsePublicKey(s):
     return key
 
 # Regular expression to match YYYY/MM/DD
-_date_re = re.compile(r"(\d\d\d\d)/(\d\d)/(\d\d)")
+_date_re = re.compile(r"^(\d\d\d\d)/(\d\d)/(\d\d)$")
 def _parseDate(s):
     """Validation function.  Converts from YYYY/MM/DD format to a (long)
        time value for midnight on that date."""
@@ -271,7 +303,7 @@ def _parseDate(s):
     return calendar.timegm((yyyy,MM,dd,0,0,0,0,0,0))
 
 # Regular expression to match YYYY/MM/DD HH:MM:SS
-_time_re = re.compile(r"(\d\d\d\d)/(\d\d)/(\d\d) (\d\d):(\d\d):(\d\d)")
+_time_re = re.compile(r"^(\d\d\d\d)/(\d\d)/(\d\d) (\d\d):(\d\d):(\d\d)$")
 def _parseTime(s):
     """Validation function.  Converts from YYYY/MM/DD HH:MM:SS format
        to a (float) time value for GMT."""
@@ -289,12 +321,12 @@ def _parseTime(s):
 
 _NICKNAME_CHARS = ("ABCDEFGHIJKLMNOPQRSTUVWXYZ"+
                    "abcdefghijklmnopqrstuvwxyz"+
-                   "0123456789_.!@#-")
+                   "0123456789_.@-")
 MAX_NICKNAME = 128
 def _parseNickname(s):
     """Validation function.  Returns true iff s contains a valoid
        server nickname-- that is, a string of 1..128 characters,
-       containing only the characters [A-Za-z0-9_.!@#] and '-'.
+       containing only the characters [A-Za-z0-9_@], '.' or '-'.
        """
     s = s.strip()
     bad = s.translate(mixminion.Common._ALLCHARS, _NICKNAME_CHARS)
@@ -306,6 +338,17 @@ def _parseNickname(s):
         raise ConfigError("Nickname is too short")
     return s
 
+def _parseFilename(s):
+    """Validation function.  Matches a filename, expanding tildes as
+       appropriate"""
+    s = s.strip()
+    if s[0] in "\"'":
+        if s[-1] != s[0]:
+            raise ConfigError("Mismatched quotes")
+        s = s[1:-1]
+
+    return os.path.expanduser(s)
+
 #----------------------------------------------------------------------
 
 # Regular expression to match a section header.
@@ -333,7 +376,6 @@ def _readConfigFile(contents):
     if not isPrintingAscii(contents):
         raise ConfigError("Invalid characters in file")
 
-    #FFFF We should really use xreadlines or something if we have a file.
     fileLines = contents.split("\n")
     if fileLines[-1] == '':
         del fileLines[-1]
@@ -453,6 +495,8 @@ class _ConfigFile:
     #                                 default, ) }
     #     _restrictFormat is 1/0: do we allow full RFC822ness, or do
     #         we insist on a tight data format?
+    #     _restrictKeys is 1/0: do we raise a ConfigError when we see an
+    #         unrecognized key, or do we simply generate a warning?
 
     ## Validation rules:
     # A key without a corresponding entry in _syntax gives an error.
@@ -468,6 +512,7 @@ class _ConfigFile:
 
     _syntax = None
     _restrictFormat = 0
+    _restrictKeys = 1
 
     def __init__(self, filename=None, string=None, assumeValid=0):
         """Create a new _ConfigFile.  If <filename> is set, read from
@@ -479,40 +524,22 @@ class _ConfigFile:
            If <assumeValid> is true, skip all unnecessary validation
            steps.  (Use this to load a file that's already been checked as
            valid.)"""
-        assert filename is None or string is None
+        assert (filename is None) != (string is None)
+
         if not hasattr(self, '_callbacks'):
             self._callbacks = {}
 
         self.assumeValid = assumeValid
-        self.fname = filename
+
         if filename:
-            self.reload()
-        elif string:
-            self.__reload(None, string)
+            contents = mixminion.Common.readPossiblyGzippedFile(filename)
+            self.__load(contents)
         else:
-            self.clear()
+            assert string is not None
+            self.__load(string)
 
-    def clear(self):
-        """Remove all sections from this _ConfigFile object."""
-        self._sections = {}
-        self._sectionEntries = {}
-        self._sectionNames = []
-
-    def reload(self):
-        """Reload this _ConfigFile object from disk.  If the object is no
-           longer present and correctly formatted, raise an error, but leave
-           the contents of this object unchanged."""
-        if not self.fname:
-            return
-
-        contents = mixminion.Common.readPossiblyGzippedFile(self.fname)
-        self.__reload(None, contents)
-
-    def __reload(self, file, fileContents):
+    def __load(self, fileContents):
         """As in .reload(), but takes an open file object _or_ a string."""
-        if fileContents is None:
-            fileContents = file.read()
-            file.close()
 
         fileContents = _abnormal_line_ending_re.sub("\n", fileContents)
 
@@ -521,27 +548,27 @@ class _ConfigFile:
         else:
             sections = _readConfigFile(fileContents)
 
-        # These will become self.(_sections,_sectionEntries,_sectionNames)
-        # if we are successful.
-        self_sections = {}
-        self_sectionEntries = {}
-        self_sectionNames = []
+        sections = self.prevalidate(sections)
+
+        self._sections = {}
+        self._sectionEntries = {}
+        self._sectionNames = []
         sectionEntryLines = {}
 
         for secName, secEntries in sections:
-            self_sectionNames.append(secName)
+            self._sectionNames.append(secName)
 
-            if self_sections.has_key(secName):
+            if self._sections.has_key(secName):
                 raise ConfigError("Duplicate section [%s]" %secName)
 
             section = {}
             sectionEntries = []
             entryLines = []
-            self_sections[secName] = section
-            self_sectionEntries[secName] = sectionEntries
+            self._sections[secName] = section
+            self._sectionEntries[secName] = sectionEntries
             sectionEntryLines[secName] = entryLines
 
-            secConfig = self._syntax.get(secName, None)
+            secConfig = self._syntax.get(secName)
 
             if not secConfig:
                 LOG.warn("Skipping unrecognized section %s", secName)
@@ -553,8 +580,12 @@ class _ConfigFile:
                 try:
                     rule, parseFn, default = secConfig[k]
                 except KeyError:
-                    raise ConfigError("Unrecognized key %s on line %s" %
-                                      (k, line))
+                    if self._restrictKeys:
+                        raise ConfigError("Unrecognized key %s on line %s" %
+                                          (k, line))
+                    else:
+                        LOG.warn("Unrecognized key %s on line %s", k, line)
+                        continue
 
                 # Parse and validate the value of this entry.
                 if parseFn is not None:
@@ -602,7 +633,7 @@ class _ConfigFile:
                             assert rule == 'ALLOW*'
                             section[k] = map(parseFn,default)
 
-            cb = self._callbacks.get(secName, None)
+            cb = self._callbacks.get(secName)
             if cb:
                 cb(section, sectionEntries)
 
@@ -611,20 +642,15 @@ class _ConfigFile:
         for secName, secConfig in self._syntax.items():
             secRule = secConfig.get('__SECTION__', ('ALLOW',None,None))
             if (secRule[0] == 'REQUIRE'
-                and not self_sections.has_key(secName)):
+                and not self._sections.has_key(secName)):
                 raise ConfigError("Section [%s] not found." %secName)
-            elif not self_sections.has_key(secName):
-                self_sections[secName] = {}
-                self_sectionEntries[secName] = []
+            elif not self._sections.has_key(secName):
+                self._sections[secName] = {}
+                self._sectionEntries[secName] = []
 
         if not self.assumeValid:
             # Call our validation hook.
-            self.validate(self_sections, self_sectionEntries,
-                          sectionEntryLines, fileContents)
-
-        self._sections = self_sections
-        self._sectionEntries = self_sectionEntries
-        self._sectionNames = self_sectionNames
+            self.validate(sectionEntryLines, fileContents)
 
     def _addCallback(self, section, cb):
         """For use by subclasses.  Adds a callback for a section"""
@@ -632,8 +658,14 @@ class _ConfigFile:
             self._callbacks = {}
         self._callbacks[section] = cb
 
-    def validate(self, sections, sectionEntries, entryLines,
-                 fileContents):
+    def prevalidate(self, contents):
+        """Given a list of (SECTION-NAME, [(KEY, VAL, LINENO)]), makes
+           decision on whether to parse sections.  Subclasses should
+           override.  Returns a revised version of its input.
+        """
+        return contents
+
+    def validate(self, entryLines, fileContents):
         """Check additional semantic properties of a set of configuration
            data before overwriting old data.  Subclasses should override."""
         pass
@@ -671,25 +703,31 @@ class ClientConfig(_ConfigFile):
     _syntax = {
         'Host' : { '__SECTION__' : ('ALLOW', None, None),
                    'ShredCommand': ('ALLOW', _parseCommand, None),
-                   'EntropySource': ('ALLOW', None, "/dev/urandom"),
+                   'EntropySource': ('ALLOW', _parseFilename, "/dev/urandom"),
+                   'TrustedUser': ('ALLOW*', None, None),
                    },
         'DirectoryServers' :
                    { '__SECTION__' : ('REQUIRE', None, None),
                      'ServerURL' : ('ALLOW*', None, None),
                      'MaxSkew' : ('ALLOW', _parseInterval, "10 minutes") },
-        'User' : { 'UserDir' : ('ALLOW', None, "~/.mixminion" ) },
+        'User' : { 'UserDir' : ('ALLOW', _parseFilename, "~/.mixminion" ) },
         'Security' : { 'PathLength' : ('ALLOW', _parseInt, "8"),
                        'SURBAddress' : ('ALLOW', None, None),
-                       'SURBPathLength' : ('ALLOW', _parseInt, "8"),
-                       'SURBLifetime' : ('ALLOW', _parseInterval, "7 days") },
+                       'SURBPathLength' : ('ALLOW', _parseInt, "4"),
+                       'SURBLifetime' : ('ALLOW', _parseInterval, "7 days"),
+                       'ForwardPath' : ('ALLOW', None, "*"),
+                       'ReplyPath' : ('ALLOW', None, "*"),
+                       'SURBPath' : ('ALLOW', None, "*"),
+                       },
+        'Network' : { 'ConnectionTimeout' : ('ALLOW', _parseInterval, None) }
         }
     def __init__(self, fname=None, string=None):
         _ConfigFile.__init__(self, fname, string)
 
-    def validate(self, sections, entries, lines, contents):
-        _validateHostSection(sections.get('Host', {}))
+    def validate(self, lines, contents):
+        _validateHostSection(self['Host'])
 
-        security = sections.get('Security', {})
+        security = self['Security']
         p = security.get('PathLength', 8)
         if not 0 < p <= 16:
             raise ConfigError("Path length must be between 1 and 16")
@@ -697,9 +735,18 @@ class ClientConfig(_ConfigFile):
             LOG.warn("Your default path length is frighteningly low."
                           "  I'll trust that you know what you're doing.")
 
+        t = self['Network'].get('ConnectionTimeout')
+        if t:
+            if int(t) < 5:
+                LOG.warn("Very short connection timeout")
+            elif int(t) > 60:
+                LOG.warn("Very long connection timeout")
+
 def _validateHostSection(sec):
     """Helper function: Makes sure that the shared [Host] section is correct;
        raise ConfigError if it isn't"""
     # For now, we do nothing here.  EntropySource and ShredCommand are checked
     # in configure_trng and configureShredCommand, respectively.
-    pass
+
+    # Host is checked in setupTrustedUIDs.
+
