@@ -1,5 +1,5 @@
 # Copyright 2002-2003 Nick Mathewson.  See LICENSE for licensing information.
-# $Id: Crypto.py,v 1.35 2003/01/05 13:19:53 nickm Exp $
+# $Id: Crypto.py,v 1.45 2003/06/05 18:41:40 nickm Exp $
 """mixminion.Crypto
 
    This package contains all the cryptographic primitives required
@@ -8,11 +8,13 @@
    the functions in mixminion.Crypto, and not call _minionlib's crypto
    functionality themselves."""
 
+import binascii
 import copy_reg
+import errno
 import os
 import stat
 import sys
-import binascii
+import threading
 from types import StringType
 
 import mixminion._minionlib as _ml
@@ -49,7 +51,8 @@ def init_crypto(config=None):
     except MixFatalError:
         raise
     except:
-        raise MixFatalError("Error initializing entropy source")
+        info = sys.exc_info()
+        raise MixFatalError("Error initializing entropy source: %s", info[0])
     openssl_seed(40)
 
 def sha1(s):
@@ -270,7 +273,8 @@ def pk_fingerprint(key):
 def pk_PEM_save(rsa, filename, password=None):
     """Save a PEM-encoded private key to a file.  If <password> is provided,
        encrypt the key using the password."""
-    f = open(filename, 'w')
+    fd = os.open(filename, os.O_WRONLY|os.O_CREAT, 0600)
+    f = os.fdopen(fd, 'wb')
     if password:
         rsa.PEM_write_key(f, 0, password)
     else:
@@ -478,6 +482,7 @@ class RNG:
            bytes at a time."""
         self.bytes = ""
         self.chunksize = chunksize
+
     def getBytes(self, n):
         """Returns a string of 'n' random bytes."""
 
@@ -560,6 +565,35 @@ class RNG:
         #return o / float(0x7fffffff)
         return o / 2147483647.0
 
+    def openNewFile(self, dir, prefix="", binary=1):
+        """Generate a new random filename within a directory with a given
+           prefix within a directory, and open a new file within the directory
+           with that filename.  Return 2-tuple of a file object and the
+           random portion of the filename.
+
+           Random portions are generated by choosing 8 random characters
+           from the set 'A-Za-z0-9+-'.
+           """
+        flags = os.O_WRONLY|os.O_CREAT|os.O_EXCL
+        mode = "w"
+        if binary:
+            flags |= getattr(os, 'O_BINARY', 0)
+            mode = "wb"
+        while 1:
+            bytes = self.getBytes(6)
+            base = binascii.b2a_base64(bytes).strip().replace("/","-")
+            fname = os.path.join(dir, "%s%s"%(prefix,base))
+            try:
+                fd = os.open(fname, flags, 0600)
+                return os.fdopen(fd, mode), base
+            except OSError, e:
+                if e.errno != errno.EEXIST:
+                    raise e
+                # If the file exists (a rare event!) we pass through, and
+                # try again.  This paranoia is brought to you by user
+                # request. :)
+        raise MixFatalError("Unreachable") # appease pychecker.
+
     def _prng(self, n):
         """Abstract method: Must be overridden to return n bytes of fresh
            entropy."""
@@ -588,13 +622,15 @@ class AESCounterPRNG(RNG):
             raise MixFatalError("Exhausted period of PRNG.")
         return prng(self.key,n,c)
 
-_theSharedPRNG = None
 def getCommonPRNG():
     '''Returns a general-use AESCounterPRNG, initializing it if necessary.'''
-    global _theSharedPRNG
-    if _theSharedPRNG is None:
-        _theSharedPRNG = AESCounterPRNG()
-    return _theSharedPRNG
+    # We create one PRNG per thread.
+    thisThread = threading.currentThread()
+    try:
+        return thisThread.minion_shared_PRNG
+    except AttributeError:
+        thisThread.minion_shared_PRNG = AESCounterPRNG()
+        return thisThread.minion_shared_PRNG
 
 #----------------------------------------------------------------------
 # TRNG implementation
@@ -619,7 +655,7 @@ def configure_trng(config):
        none is provided, tries some sane defaults."""
     global _TRNG_FILENAME
     if config is not None:
-        requestedFile = config['Host'].get('EntropySource', None)
+        requestedFile = config['Host'].get('EntropySource')
     else:
         requestedFile = None
 
@@ -666,6 +702,8 @@ class _TrueRNG(RNG):
         """Creates a TrueRNG to retrieve data from our underlying RNG 'n'
            bytes at a time"""
         RNG.__init__(self,n)
+        self.__lock = threading.Lock()
+
     def _prng(self,n):
         "Returns n fresh bytes from our true RNG."
         if _TRNG_FILENAME is None:
@@ -676,6 +714,18 @@ class _TrueRNG(RNG):
         f.close()
         return d
 
+    def getBytes(self, n):
+        # We need to synchronize this method, since a single TRNG instance
+        # is shared by all threads.
+        self.__lock.acquire()
+        b = RNG.getBytes(self, n)
+        self.__lock.release()
+        return b
+
 # Global _TrueRNG instance, for use by trng().
 _theTrueRNG = _TrueRNG(1024)
 
+# Return the shared instance of the true RNG.
+def getTrueRNG():
+    """Return the shared instance of the true RNG."""
+    return _theTrueRNG
