@@ -1,5 +1,5 @@
-# Copyright 2002 Nick Mathewson.  See LICENSE for licensing information.
-# $Id: Packet.py,v 1.22 2002/12/16 02:40:11 nickm Exp $
+# Copyright 2002-2003 Nick Mathewson.  See LICENSE for licensing information.
+# $Id: Packet.py,v 1.46 2003/06/06 06:04:57 nickm Exp $
 """mixminion.Packet
 
    Functions, classes, and constants to parse and unparse Mixminion
@@ -9,26 +9,36 @@
    packets, see BuildMessage.py.  For functions that handle
    server-side processing of packets, see PacketHandler.py."""
 
-__all__ = [ 'ParseError', 'Message', 'Header', 'Subheader',
-            'parseMessage', 'parseHeader', 'parseSubheader',
-            'getTotalBlocksForRoutingInfoLen', 'parsePayload',
-            'SingletonPayload', 'FragmentPayload', 'ReplyBlock',
-            'IPV4Info', 'SMTPInfo', 'MBOXInfo', 'parseIPV4Info',
-            'parseSMTPInfo', 'parseMBOXInfo', 'ReplyBlock',
-            'parseReplyBlock', 'ENC_SUBHEADER_LEN', 'HEADER_LEN',
-            'PAYLOAD_LEN', 'MAJOR_NO', 'MINOR_NO', 'SECRET_LEN', 'TAG_LEN',
-            'SINGLETON_PAYLOAD_OVERHEAD', 'OAEP_OVERHEAD',
-            'FRAGMENT_PAYLOAD_OVERHEAD', 'ENC_FWD_OVERHEAD',
-            'DROP_TYPE', 'FWD_TYPE', 'SWAP_FWD_TYPE',
-            'SMTP_TYPE', 'MBOX_TYPE', 'MIN_EXIT_TYPE'
-]
+__all__ = [ 'compressData', 'CompressedDataTooLong', 'DROP_TYPE',
+            'ENC_FWD_OVERHEAD', 'ENC_SUBHEADER_LEN',
+            'FRAGMENT_PAYLOAD_OVERHEAD', 'FWD_TYPE', 'FragmentPayload',
+            'HEADER_LEN', 'IPV4Info', 'MAJOR_NO', 'MBOXInfo',
+            'MBOX_TYPE', 'MINOR_NO', 'MIN_EXIT_TYPE',
+            'MIN_SUBHEADER_LEN', 'Message',
+            'OAEP_OVERHEAD', 'PAYLOAD_LEN', 'ParseError', 'ReplyBlock',
+            'ReplyBlock', 'SECRET_LEN', 'SINGLETON_PAYLOAD_OVERHEAD',
+            'SMTPInfo', 'SMTP_TYPE', 'SWAP_FWD_TYPE', 'SingletonPayload',
+            'Subheader', 'TAG_LEN', 'TextEncodedMessage',
+            'parseHeader', 'parseIPV4Info',
+            'parseMBOXInfo', 'parseMessage', 'parsePayload', 'parseReplyBlock',
+            'parseReplyBlocks', 'parseSMTPInfo', 'parseSubheader',
+            'parseTextEncodedMessages', 'parseTextReplyBlocks', 'uncompressData'            ]
 
+import binascii
+import re
 import struct
+import sys
+import zlib
 from socket import inet_ntoa, inet_aton
-from mixminion.Common import MixError, floorDiv, isSMTPMailbox
+from mixminion.Common import MixError, MixFatalError, encodeBase64, \
+     floorDiv, formatTime, isSMTPMailbox, LOG, armorText, unarmorText
+from mixminion.Crypto import sha1
+
+if sys.version_info[:3] < (2,2,0):
+    import mixminion._zlibutil as zlibutil
 
 # Major and minor number for the understood packet format.
-MAJOR_NO, MINOR_NO = 0,1
+MAJOR_NO, MINOR_NO = 0,3
 
 # Length of a Mixminion message
 MESSAGE_LEN = 1 << 15
@@ -41,12 +51,13 @@ PAYLOAD_LEN = MESSAGE_LEN - HEADER_LEN*2
 OAEP_OVERHEAD = 42
 
 # Length of a subheader, once RSA-encoded.
-ENC_SUBHEADER_LEN = 128
+ENC_SUBHEADER_LEN = 256
 # Smallest possible size for a subheader
 MIN_SUBHEADER_LEN = 42
 # Most information we can fit into a subheader before padding
 MAX_SUBHEADER_LEN = ENC_SUBHEADER_LEN - OAEP_OVERHEAD
-# Longest routing info that will fit in the main subheader
+# Longest routing info that will fit into the RSA-encrypted portion of
+# the subheader.
 MAX_ROUTING_INFO_LEN = MAX_SUBHEADER_LEN - MIN_SUBHEADER_LEN
 
 # Length of a digest
@@ -56,9 +67,6 @@ SECRET_LEN = 16
 # Length of end-to-end message tag
 TAG_LEN = 20
 
-# Most info that fits in a single extened subheader
-ROUTING_INFO_PER_EXTENDED_SUBHEADER = ENC_SUBHEADER_LEN
-
 #----------------------------------------------------------------------
 # Values for the 'Routing type' subheader field
 # Mixminion types
@@ -108,35 +116,7 @@ def parseHeader(s):
     if len(s) != HEADER_LEN:
         raise ParseError("Bad header length")
 
-    return Header(s)
-
-class Header:
-    """Represents a 2K Mixminion header, containing up to 16 subheaders."""
-    def __init__(self, contents):
-        """Initialize a new header from its contents"""
-        self.contents = contents
-
-    def __getitem__(self, i):
-        """header[i] -> str
-
-           Returns the i'th encoded subheader of this header, for i in 0..15"""
-        if i < 0: i = 16+i
-        return self.contents[i*ENC_SUBHEADER_LEN:
-                             (i+1)*ENC_SUBHEADER_LEN]
-
-    def __getslice__(self, i, j):
-        """header[i:j] -> str
-
-           Returns a slice of the i-j'th subheaders of this header."""
-        if j > 16: j = 16
-        if i < 0: i += 16
-        if j < 0: j += 16
-        return self.contents[i*ENC_SUBHEADER_LEN:
-                             j*ENC_SUBHEADER_LEN]
-
-    def __len__(self):
-        """Return the number of subheaders in this header (always 16)"""
-        return 16
+    return s
 
 # A subheader begins with: a major byte, a minor byte, SECRET_LEN secret
 # bytes, DIGEST_LEN digest bytes, a routing_len short, and a routing_type
@@ -156,34 +136,30 @@ def parseSubheader(s):
     except struct.error:
         raise ParseError("Misformatted subheader")
     ri = s[MIN_SUBHEADER_LEN:]
+    underflow = ""
     if rlen < len(ri):
-        ri = ri[:rlen]
+        ri, underflow = ri[:rlen], ri[rlen:]
     if rt >= MIN_EXIT_TYPE and rlen < 20:
         raise ParseError("Subheader missing tag")
-    return Subheader(major,minor,secret,digest,rt,ri,rlen)
-
-def getTotalBlocksForRoutingInfoLen(bytes):
-    """Return the number of subheaders that will be needed for a hop
-       whose routinginfo is (bytes) long."""
-    if bytes <= MAX_ROUTING_INFO_LEN:
-        return 1
-    else:
-        extraBytes = bytes - MAX_ROUTING_INFO_LEN
-        return 2 + floorDiv(extraBytes,ROUTING_INFO_PER_EXTENDED_SUBHEADER)
+    return Subheader(major,minor,secret,digest,rt,ri,rlen,underflow)
 
 class Subheader:
-    """Represents a decoded Mixminion subheader
+    """Represents a decoded Mixminion subheader.
 
        Fields: major, minor, secret, digest, routinglen, routinginfo,
                routingtype.
 
        A Subheader can exist in a half-initialized state where routing
-       info has been read from the first header, but not from the
-       extened headers.  If this is so, routinglen will be > len(routinginfo).
-       """
+       info has been read from the first RSA-encrypted data, but not
+       from the symmetrically encrypted data in the rest of the
+       header.  If this is so, routinglen will be > len(routinginfo).
 
+       If 'underflow' is present, it contains material that does not
+       belong to this subheader that was provided to 'parseSubheader'
+       anyway.
+       """
     def __init__(self, major, minor, secret, digest, routingtype,
-                 routinginfo, routinglen=None):
+                 routinginfo, routinglen=None, underflow=""):
         """Initialize a new subheader"""
         self.major = major
         self.minor = minor
@@ -195,6 +171,7 @@ class Subheader:
             self.routinglen = routinglen
         self.routingtype = routingtype
         self.routinginfo = routinginfo
+        self.underflow = underflow
 
     def __repr__(self):
         return ("Subheader(major=%(major)r, minor=%(minor)r, "+
@@ -221,33 +198,34 @@ class Subheader:
         self.routinginfo = info
         self.routinglen = len(info)
 
-    def isExtended(self):
-        """Return true iff the routinginfo is too long to fit in a single
-           subheader."""
-        return self.routinglen > MAX_ROUTING_INFO_LEN
+    def appendOverflow(self, data):
+        """Given a string containing additional routing info, add it
+           to the routinginfo of this object.  """
+        self.routinginfo += data
+        assert len(self.routinginfo) <= self.routinglen
 
-    def getNExtraBlocks(self):
-        """Return the number of extra blocks that will be needed to fit
-           the routinginfo."""
-        return getTotalBlocksForRoutingInfoLen(self.routinglen)-1
-
-    def appendExtraBlocks(self, data):
-        """Given a string containing additional (decoded) blocks of
-           routing info, add them to the routinginfo of this
-           object.
+    def getUnderflowLength(self):
+        """Return the number of bytes from the rest of the header that should
+           be included in the RSA-encrypted part of the header.
         """
-        nBlocks = self.getNExtraBlocks()
-        assert len(data) == nBlocks * ENC_SUBHEADER_LEN
-        raw = [self.routinginfo]
-        for i in range(nBlocks):
-            block = data[i*ENC_SUBHEADER_LEN:(i+1)*ENC_SUBHEADER_LEN]
-            raw.append(block)
-        self.routinginfo = ("".join(raw))[:self.routinglen]
+        return max(0, MAX_ROUTING_INFO_LEN - self.routinglen)
+
+    def getOverflowLength(self):
+        """Return the length of the data from routinginfo that will
+           not fit in the RSA-encrypted part of the header.
+        """
+        return max(0, self.routinglen - MAX_ROUTING_INFO_LEN)
+
+    def getOverflow(self):
+        """Return the portion of routinginfo that doesn't fit into the
+           RSA-encrypted part of the header.
+        """
+        return self.routinginfo[MAX_ROUTING_INFO_LEN:]
 
     def pack(self):
         """Return the (unencrypted) string representation of this Subhead.
 
-           Does not include extra blocks"""
+           Does not include overflow or underflow"""
         assert self.routinglen == len(self.routinginfo)
         assert len(self.digest) == DIGEST_LEN
         assert len(self.secret) == SECRET_LEN
@@ -257,22 +235,6 @@ class Subheader:
                            self.major,self.minor,self.secret,self.digest,
                            self.routinglen, self.routingtype)+info
 
-    def getExtraBlocks(self):
-        """Return a list of (unencrypted) blocks of extra routing info."""
-        if not self.isExtended():
-            return []
-        else:
-            info = self.routinginfo[MAX_ROUTING_INFO_LEN:]
-            result = []
-            for i in range(self.getNExtraBlocks()):
-                content = info[i*ROUTING_INFO_PER_EXTENDED_SUBHEADER:
-                               (i+1)*ROUTING_INFO_PER_EXTENDED_SUBHEADER]
-                missing = ROUTING_INFO_PER_EXTENDED_SUBHEADER-len(content)
-                if missing > 0:
-                    content += '\000'*missing
-                result.append(content)
-            return result
-
 #----------------------------------------------------------------------
 # UNENCRYPTED PAYLOADS
 
@@ -392,9 +354,58 @@ class FragmentPayload(_Payload):
 #   routingInfo for the last server.
 RB_UNPACK_PATTERN = "!4sBBL%dsHH%ss" % (HEADER_LEN, SECRET_LEN)
 MIN_RB_LEN = 30+HEADER_LEN
+RB_ARMOR_NAME = "TYPE III REPLY BLOCK"
+
+def parseTextReplyBlocks(s):
+    """Given a string holding one or more text-encoded reply blocks,
+       return a list containing the reply blocks.  Raise ParseError on
+       failure."""
+
+    try:
+        res = unarmorText(s, (RB_ARMOR_NAME,), base64=1)
+    except ValueError, e:
+        raise ParseError(str(e))
+    blocks = []
+    for tp, fields, value in res:
+        d = {}
+        for k,v in fields:
+            d[k]=v
+        if not d.get("Version") == '0.2':
+            LOG.warn("Skipping SURB with bad version: %r", d.get("Version"))
+        blocks.append(parseReplyBlock(value))
+    return blocks
+
+def parseReplyBlocks(s):
+    """Given a string containing a list of concatenated encoded reply blocks,
+       return list of reply blocks corresponding to those in the string.
+       Raise ParseError on failure.
+    """
+    blocks = []
+    while 1:
+        # Skip over any whitespace before or after the reply blocks.
+        while s and s[0] in ' \t\r\n':
+            s = s[1:]
+        if not s:
+            break
+        block, length = _parseReplyBlock(s)
+        blocks.append(block)
+        s = s[length:]
+    return blocks
 
 def parseReplyBlock(s):
-    """Return a new ReplyBlock object for an encoded reply block"""
+    """Return a new ReplyBlock object for an encoded reply block.
+
+       Raise ParseError on failure.
+    """
+    block, length = _parseReplyBlock(s)
+    if length > len(s):
+        raise ParseError("Misformatted reply block: extra data.")
+    return block
+
+def _parseReplyBlock(s):
+    """Helper function: Given a string containing one or more (binary) reply
+       blocks, extract the first one.  On success, return a tuple containing
+       a ReplyBlock, and the length of 's' taken up by the reply block."""
     if len(s) < MIN_RB_LEN:
         raise ParseError("Reply block too short")
     try:
@@ -411,10 +422,11 @@ def parseReplyBlock(s):
                          major,minor)
 
     ri = s[MIN_RB_LEN:]
-    if len(ri) != rlen:
-        raise ParseError("Misformatted reply block")
+    length = rlen + MIN_RB_LEN
+    ri = ri[:rlen]
 
-    return ReplyBlock(header, timestamp, rt, ri, key)
+    surb =  ReplyBlock(header, timestamp, rt, ri, key)
+    return surb, length
 
 class ReplyBlock:
     """A mixminion reply block, including the address of the first hop
@@ -428,6 +440,17 @@ class ReplyBlock:
         self.routingInfo = ri
         self.encryptionKey = key
 
+    def format(self):
+        hash = binascii.b2a_hex(sha1(self.pack()))
+        expiry = formatTime(self.timestamp)
+        if self.routingType == SWAP_FWD_TYPE:
+            server = parseIPV4Info(self.routingInfo).format()
+        else:
+            server = "????"
+        return """Reply block hash: %s
+Expires at: %s GMT
+First server is: %s""" % (hash, expiry, server)
+
     def pack(self):
         """Returns the external representation of this reply block"""
         return struct.pack(RB_UNPACK_PATTERN,
@@ -435,6 +458,11 @@ class ReplyBlock:
                            len(self.routingInfo), self.routingType,
                            self.encryptionKey) + self.routingInfo
 
+    def packAsText(self):
+        """Returns the external text representation of this reply block"""
+        return armorText(self.pack(), RB_ARMOR_NAME,
+                         headers=(("Version", "0.2"),))
+
 #----------------------------------------------------------------------
 # Routing info
 
@@ -446,7 +474,7 @@ def parseIPV4Info(s):
     """Converts routing info for an IPV4 address into an IPV4Info object,
        suitable for use by FWD or SWAP_FWD modules."""
     if len(s) != 4+2+DIGEST_LEN:
-        raise ParseError("IPV4 information with wrong length")
+        raise ParseError("IPV4 information with wrong length (%d)" % len(s))
     try:
         ip, port, keyinfo = struct.unpack(IPV4_PAT, s)
     except struct.error:
@@ -467,18 +495,31 @@ class IPV4Info:
         self.port = port
         self.keyinfo = keyinfo
 
+    def format(self):
+        return "%s:%s (keyid=%s)"%(self.ip, self.port,
+                                   binascii.b2a_hex(self.keyinfo))
+
+
     def pack(self):
         """Return the routing info for this address"""
         assert len(self.keyinfo) == DIGEST_LEN
         return struct.pack(IPV4_PAT, inet_aton(self.ip),
                            self.port, self.keyinfo)
 
+    def __repr__(self):
+        return "IPV4Info(%r, %r, %r)"%(self.ip, self.port, self.keyinfo)
+
     def __hash__(self):
         return hash(self.pack())
 
-    def __eq__(self, other):
-        return (type(self) == type(other) and self.ip == other.ip and
-                self.port == other.port and self.keyinfo == other.keyinfo)
+    def __cmp__(self, other):
+        r = cmp(type(self), type(other))
+        if r: return r
+        r = cmp(self.ip, other.ip)
+        if r: return r
+        r = cmp(self.port, other.port)
+        if r: return r
+        return cmp(self.keyinfo, other.keyinfo)
 
 def parseSMTPInfo(s):
     """Convert the encoding of an SMTP exitinfo into an SMTPInfo object."""
@@ -513,3 +554,228 @@ class MBOXInfo:
     def pack(self):
         """Return the external representation of this routing info."""
         return self.user
+
+#----------------------------------------------------------------------
+# Ascii-encoded packets
+#
+# A packet is an ASCII-armored object, with the following headers:
+#     Message-type: (plaintext|encrypted|binary|overcompressed)
+#    [Decoding-handle: base64-encoded-stuff]
+#
+# If Message-type is 'plaintext', the body is dash-escaped.  Otherwise,
+# the body is base64-encoded.
+
+MESSAGE_ARMOR_NAME = "TYPE III ANONYMOUS MESSAGE"
+ 
+def parseTextEncodedMessages(msg,force=0):
+
+    """Given a text-encoded Type III packet, return a list of
+       TextEncodedMessage objects or raise ParseError.
+       
+          force -- uncompress the message even if it's overcompressed.
+    """
+    
+    def isBase64(t,f):
+        for k,v in f:
+            if k == "Message-type":
+                if v != 'plaintext':
+                    return 1
+        return 0
+    
+    unarmored = unarmorText(msg, (MESSAGE_ARMOR_NAME,), base64fn=isBase64)
+    res = []
+    for tp,fields,val in unarmored:
+        d = {}
+        for k,v in fields:
+            d[k] = v
+        if d.get("Message-type", "plaintext") == "plaintext":
+            msgType = 'TXT'
+        elif d['Message-type'] == 'overcompressed':
+            msgType = "LONG"
+        elif d['Message-type'] == 'binary':
+            msgType = "BIN"
+        elif d['Message-type'] == 'encrypted':
+            msgType = "ENC"
+        else:
+            raise ParseError("Unknown message type: %r"%d["Message-type"])
+
+        ascTag = d.get("Decoding-handle")
+        if ascTag:
+            msgType = "ENC"
+
+        if msgType == 'LONG' and force:
+            msg = uncompressData(msg)
+            
+        if msgType in ('TXT','BIN','LONG'):
+            res.append(TextEncodedMessage(val, msgType))
+        else:
+            assert msgType == 'ENC'
+            try:
+                tag = binascii.a2b_base64(ascTag)
+            except (TypeError, binascii.Incomplete, binascii.Error), e:
+                raise ParseError("Error in base64 encoding: %s"%e)
+            if len(tag) != TAG_LEN:
+                raise ParseError("Impossible tag length: %s"%len(tag))
+            res.append(TextEncodedMessage(val, 'ENC', tag))
+
+    return res
+
+class TextEncodedMessage:
+    """A TextEncodedMessage object holds a Type III message as delivered
+       over a text-based medium."""
+    def __init__(self, contents, messageType, tag=None):
+        """Create a new TextEncodedMessage given a set of contents, a
+           messageType ('TXT', 'ENC', 'LONG', or 'BIN'), and optionally
+           a tag."""
+        assert messageType in ('TXT', 'ENC', 'LONG', 'BIN')
+        assert tag is None or (messageType == 'ENC' and len(tag) == 20)
+        self.contents = contents
+        self.messageType = messageType
+        self.tag = tag
+    def isBinary(self):
+        """Return true iff this is a binary plaintext packet."""
+        return self.messageType == 'BIN'
+    def isText(self):
+        """Return true iff this is a text plaintext packet."""
+        return self.messageType == 'TXT'
+    def isEncrypted(self):
+        """Return true iff this is an encrypted packet."""
+        return self.messageType == 'ENC'
+    def isOvercompressed(self):
+        """Return true iff this is an overcompressed plaintext packet."""
+        return self.messageType == 'LONG'
+    def getContents(self):
+        """Return the (unencoded) contents of this packet."""
+        return self.contents
+    def getTag(self):
+        """Return the (unencoded) decoding handle for this packet, or None."""
+        return self.tag
+    def pack(self):
+        """Return the text representation of this message."""
+        c = self.contents
+        fields = [("Message-type",
+                   { 'TXT' : "plaintext",
+                     'LONG' : "overcompressed",
+                     'BIN' : "binary",
+                     'ENC' : "encrypted" }[self.messageType]),
+                  ]
+        if self.messageType == 'ENC':
+            fields.append(("Decoding-handle",
+                           binascii.b2a_base64(self.tag).strip()))
+
+        return armorText(c, MESSAGE_ARMOR_NAME, headers=fields,
+                         base64=(self.messageType!='TXT'))
+
+#----------------------------------------------------------------------
+# COMPRESSION FOR PAYLOADS
+
+# Global: contains 0 if we haven't validated zlib; 1 if we have, and 0.5
+#    if we're in the middle of validation.
+_ZLIB_LIBRARY_OK = 0
+
+def compressData(payload):
+    """Given a string 'payload', compress it with the 'deflate' method
+       as specified in the remailer spec and in RFC1951."""
+    if not _ZLIB_LIBRARY_OK:
+        _validateZlib()
+
+    # Don't change any of these options; if different Mixminion clients
+    # compress their data differently, an adversary could distinguish
+    # messages generated by them.
+    zobj = zlib.compressobj(zlib.Z_BEST_COMPRESSION, zlib.DEFLATED,
+                            zlib.MAX_WBITS, zlib.DEF_MEM_LEVEL,
+                            zlib.Z_DEFAULT_STRATEGY)
+    s1 = zobj.compress(payload)
+    s2 = zobj.flush()
+    s = s1 + s2
+
+    # Now we check the 2 bytes of zlib header.  Strictly speaking,
+    # these are irrelevant, as are the 4 bytes of adler-32 checksum at
+    # the end.  Still, we can afford 6 bytes per payload, and
+    # reconstructing the checksum to keep zlib happy is a bit of a pain.
+    assert s[0] == '\x78' # deflate, 32K window
+    assert s[1] == '\xda' # no dict, max compression
+    return s
+
+class CompressedDataTooLong(MixError):
+    """Exception: raised when try to uncompress data that turns out to be
+       longer than we had expected."""
+    pass
+
+def uncompressData(payload, maxLength=None):
+    """Uncompress a string 'payload'; raise ParseError if it is not
+       valid compressed data.  If the expanded data is longer than
+       maxLength, we raise 'CompressedDataTooLong'."""
+
+    if len(payload) < 6 or payload[0:2] != '\x78\xDA':
+        raise ParseError("Invalid zlib header")
+
+    # This code is necessary because versions of Python before 2.2 didn't
+    # support limited-size versions of zlib.decompress.  We use a helper
+    # function helpfully submitted by Zooko.
+    if sys.version_info[:3] < (2,2,0) and maxLength is not None:
+        try:
+            return zlibutil.safe_zlib_decompress_to_retval(payload,
+                                                      maxLength,
+                                                  max(maxLength*3, 1<<20))
+        except zlibutil.TooBigError:
+            raise CompressedDataTooLong()
+        except zlibutil.DecompressError, e:
+            raise ParseError("Error in compressed data: %s"%e)
+        except (IOError, ValueError), e:
+            raise ParseError("Error in compressed data: %s"%e)
+
+    try:
+        # We can't just call zlib.decompress(payload), since we may
+        # want to limit the output size.
+
+        zobj = zlib.decompressobj(zlib.MAX_WBITS)
+        # Decompress the payload.
+        if maxLength is None:
+            d = zobj.decompress(payload)
+        else:
+            # If we _do_ have Python 2.2, this is the easy way to do it.  It
+            # also uses less RAM in the failing case.
+            d = zobj.decompress(payload, maxLength)
+            if zobj.unconsumed_tail:
+                raise CompressedDataTooLong()
+
+        # Get any leftovers, which shouldn't exist.
+        nil = zobj.flush()
+        if nil != '':
+            raise ParseError("Error in compressed data")
+        return d
+    except zlib.error:
+        raise ParseError("Error in compressed data")
+    except (IOError, ValueError), e:
+        raise ParseError("Error in compressed data: %s"%e)
+
+def _validateZlib():
+    """Internal function:  Make sure that zlib is a recognized version, and
+       that it compresses things as expected.  (This check is important,
+       because using a zlib version that compressed differently from zlib1.1.4
+       would make senders partitionable by payload compression.)
+    """
+    global _ZLIB_LIBRARY_OK
+    ver = getattr(zlib, "ZLIB_VERSION")
+    if ver and ver < "1.1.2":
+        raise MixFatalError("Zlib version %s is not supported"%ver)
+
+    _ZLIB_LIBRARY_OK = 0.5
+    if ver in ("1.1.2", "1.1.3", "1.1.4"):
+        _ZLIB_LIBRARY_OK = 1
+        return
+
+    LOG.warn("Unrecognized zlib version: %r. Spot-checking output", ver)
+    # This test is inadequate, but it _might_ catch future incompatible
+    # changes.
+    _ZLIB_LIBRARY_OK = 0.5
+    good = '\x78\xda\xed\xc6A\x11\x00 \x08\x00\xb0l\xd4\xf0\x87\x02\xf6o'+\
+           '`\x0e\xef\xb6\xd7r\xed\x88S=7\xcd\xcc\xcc\xcc\xcc\xcc\xcc'+\
+           '\xcc\xcc\xcc\xcc\xcc\xcc\xcc\xcc\xcc\xcc\xcc\xcc\xbe\xdd\x03'+\
+           'q\x8d\n\x93'
+    if compressData("aZbAAcdefg"*1000) == good:
+        _ZLIB_LIBRARY_OK = 1
+    else:
+        _ZLIB_LIBRARY_OK = 0
+        raise MixFatalError("Zlib output not as exected.")
