@@ -1,5 +1,5 @@
 # Copyright 2002-2003 Nick Mathewson.  See LICENSE for licensing information.
-# $Id: ServerInfo.py,v 1.36 2003/01/06 07:03:24 nickm Exp $
+# $Id: ServerInfo.py,v 1.49 2003/06/09 00:03:37 nickm Exp $
 
 """mixminion.ServerInfo
 
@@ -26,12 +26,16 @@ from mixminion.Crypto import CryptoError, DIGEST_LEN, pk_check_signature
 MAX_CONTACT = 256
 # Longest allowed Comments field
 MAX_COMMENTS = 1024
+# Longest allowed Contact-Fingerprint field
+MAX_FINGERPRINT = 128
 # Shortest permissible identity key
 MIN_IDENTITY_BYTES = 2048 >> 3
 # Longest permissible identity key
 MAX_IDENTITY_BYTES = 4096 >> 3
 # Length of packet key
-PACKET_KEY_BYTES = 1024 >> 3
+PACKET_KEY_BYTES = 2048 >> 3
+# Length of MMTP key
+MMTP_KEY_BYTES = 1024 >> 3
 
 # tmp alias to make this easier to spell.
 C = mixminion.Config
@@ -43,10 +47,10 @@ class ServerInfo(mixminion.Config._ConfigFile):
 
     """A ServerInfo object holds a parsed server descriptor."""
     _restrictFormat = 1
+    _restrictKeys = 0
     _syntax = {
         "Server" : { "__SECTION__": ("REQUIRE", None, None),
                      "Descriptor-Version": ("REQUIRE", None, None),
-                     "IP": ("REQUIRE", C._parseIP, None),
                      "Nickname": ("REQUIRE", C._parseNickname, None),
                      "Identity": ("REQUIRE", C._parsePublicKey, None),
                      "Digest": ("REQUIRE", C._parseBase64, None),
@@ -57,9 +61,15 @@ class ServerInfo(mixminion.Config._ConfigFile):
                      "Contact": ("ALLOW", None, None),
                      "Comments": ("ALLOW", None, None),
                      "Packet-Key": ("REQUIRE", C._parsePublicKey, None),
+                     "Contact-Fingerprint": ("ALLOW", None, None),
+                     # XXXX010 change these next few to "REQUIRE".
+                     "Packet-Formats": ("ALLOW", None, None),
+                     "Software": ("ALLOW", None, None),
+                     "Secure-Configuration": ("ALLOW", C._parseBoolean, None),
                      },
         "Incoming/MMTP" : {
                      "Version": ("REQUIRE", None, None),
+                     "IP": ("REQUIRE", C._parseIP, None),
                      "Port": ("REQUIRE", C._parseInt, None),
                      "Key-Digest": ("REQUIRE", C._parseBase64, None),
                      "Protocols": ("REQUIRE", None, None),
@@ -95,20 +105,29 @@ class ServerInfo(mixminion.Config._ConfigFile):
         self._validatedDigests = validatedDigests
         mixminion.Config._ConfigFile.__init__(self, fname, string, assumeValid)
         del self._validatedDigests
-        LOG.trace("Reading server descriptor %s from %s",
-                       self['Server']['Nickname'],
-                       fname or "<string>")
 
-    def validate(self, sections, entries, lines, contents):
+    def prevalidate(self, contents):
+        for name, ents in contents:
+            if name == 'Server':
+                for k,v,_ in ents:
+                    if k == 'Descriptor-Version' and v.strip() != '0.2':
+                        raise ConfigError("Unrecognized descriptor version: %s"
+                                          % v.strip())
+            # FFFF005 Remove sections with unrecognized versions.
+
+        return contents
+
+
+    def validate(self, lines, contents):
         ####
         # Check 'Server' section.
-        server = sections['Server']
-        if server['Descriptor-Version'] != '0.1':
+        server = self['Server']
+        if server['Descriptor-Version'] != '0.2':
             raise ConfigError("Unrecognized descriptor version %r",
                               server['Descriptor-Version'])
 
         ####
-        # Check the igest of file
+        # Check the digest of file
         digest = getServerInfoDigest(contents)
         if digest != server['Digest']:
             raise ConfigError("Invalid digest")
@@ -132,11 +151,14 @@ class ServerInfo(mixminion.Config._ConfigFile):
             raise ConfigError("Contact too long")
         if server['Comments'] and len(server['Comments']) > MAX_COMMENTS:
             raise ConfigError("Comments too long")
+        if server['Contact-Fingerprint'] and \
+               len(server['Contact-Fingerprint']) > MAX_FINGERPRINT:
+            raise ConfigError("Contact-Fingerprint too long")
+
         packetKeyBytes = server['Packet-Key'].get_modulus_bytes()
         if packetKeyBytes != PACKET_KEY_BYTES:
             raise ConfigError("Invalid length on packet key")
 
-
         ####
         # Check signature
         try:
@@ -148,7 +170,7 @@ class ServerInfo(mixminion.Config._ConfigFile):
             raise ConfigError("Signed digest is incorrect")
 
         ## Incoming/MMTP section
-        inMMTP = sections['Incoming/MMTP']
+        inMMTP = self['Incoming/MMTP']
         if inMMTP:
             if inMMTP['Version'] != '0.1':
                 raise ConfigError("Unrecognized MMTP descriptor version %s"%
@@ -158,7 +180,7 @@ class ServerInfo(mixminion.Config._ConfigFile):
                                   formatBase64(inMMTP['Key-Digest']))
 
         ## Outgoing/MMTP section
-        outMMTP = sections['Outgoing/MMTP']
+        outMMTP = self['Outgoing/MMTP']
         if outMMTP:
             if outMMTP['Version'] != '0.1':
                 raise ConfigError("Unrecognized MMTP descriptor version %s"%
@@ -180,7 +202,7 @@ class ServerInfo(mixminion.Config._ConfigFile):
 
     def getAddr(self):
         """Returns this server's IP address"""
-        return self['Server']['IP']
+        return self['Incoming/MMTP']['IP']
 
     def getPort(self):
         """Returns this server's IP port"""
@@ -205,14 +227,14 @@ class ServerInfo(mixminion.Config._ConfigFile):
     def getCaps(self):
         # FFFF refactor this once we have client addresses.
         caps = []
-        if not self['Incoming/MMTP'].get('Version',None):
+        if not self['Incoming/MMTP'].get('Version'):
             return caps
-        if self['Delivery/MBOX'].get('Version', None):
+        if self['Delivery/MBOX'].get('Version'):
             caps.append('mbox')
-        if self['Delivery/SMTP'].get('Version', None):
+        if self['Delivery/SMTP'].get('Version'):
             caps.append('smtp')
         # XXXX This next check is highly bogus.
-        if self['Outgoing/MMTP'].get('Version',None):
+        if self['Outgoing/MMTP'].get('Version'):
             caps.append('relay')
         return caps
 
@@ -287,7 +309,11 @@ class ServerDirectory:
        become very inefficient when directories get big, but we won't have
        that problem for a while."""
     ##Fields:
-    # servers: list of validated ServerInfo objects, in no particular order.
+    # allServers: list of validated ServerInfo objects, in no particular order.
+    # servers: sub-list of self.allServers, containing all of the servers
+    #    that are recommended.
+    # goodServerNames: list of lowercased nicknames for the recommended
+    #    servers in this directory.
     # header: a _DirectoryHeader object for the non-serverinfo part of this
     #    directory.
     def __init__(self, string=None, fname=None, validatedDigests=None):
@@ -316,17 +342,35 @@ class ServerDirectory:
         servercontents = [ "[Server]\n%s"%s for s in sections[1:] ]
 
         self.header = _DirectoryHeader(headercontents, digest)
-        self.servers = [ ServerInfo(string=s,
-                                    validatedDigests=validatedDigests)
-                         for s in servercontents ]
+        self.goodServerNames = [name.strip().lower() for name in
+                   self.header['Directory']['Recommended-Servers'].split(",")]
+        servers = [ ServerInfo(string=s,
+                               validatedDigests=validatedDigests)
+                    for s in servercontents ]
+        self.allServers = servers[:]
+        self.servers = [ s for s in servers
+                         if s.getNickname().lower() in self.goodServerNames ]
 
     def getServers(self):
-        """Return a list of ServerInfo objects in this directory"""
+        """Return a list of recommended ServerInfo objects in this directory"""
         return self.servers
 
+    def getAllServers(self):
+        """Return a list of all (even unrecommended) ServerInfo objects in
+           this directory."""
+        return self.allServers
+
+    def getRecommendedNicknames(self):
+        """Return a list of the (lowercased) nicknames of all of the
+           recommended servers in this directory."""
+        return self.goodServerNames
+
     def __getitem__(self, item):
         return self.header[item]
 
+    def get(self, item, default=None):
+        return self.header.get(item, default)
+
 class _DirectoryHeader(mixminion.Config._ConfigFile):
     """Internal object: used to parse, validate, and store fields in a
        directory's header sections.
@@ -335,18 +379,23 @@ class _DirectoryHeader(mixminion.Config._ConfigFile):
     # expectedDigest: the 20-byte digest we expect to find in this
     #    directory's header.
     _restrictFormat = 1
+    _restrictKeys = 0
     _syntax = {
         'Directory': { "__SECTION__": ("REQUIRE", None, None),
                        "Version": ("REQUIRE", None, None),
                        "Published": ("REQUIRE", C._parseTime, None),
                        "Valid-After": ("REQUIRE", C._parseDate, None),
                        "Valid-Until": ("REQUIRE", C._parseDate, None),
+                       "Recommended-Servers": ("REQUIRE", None, None),
                        },
         'Signature': {"__SECTION__": ("REQUIRE", None, None),
                  "DirectoryIdentity": ("REQUIRE", C._parsePublicKey, None),
                  "DirectoryDigest": ("REQUIRE", C._parseBase64, None),
                  "DirectorySignature": ("REQUIRE", C._parseBase64, None),
-                      }
+                      },
+        'Recommended-Software': {"__SECTION__": ("ALLOW", None, None),
+                "MixminionClient": ("ALLOW", None, None),
+                "MixminionServer": ("ALLOW", None, None), }
         }
     def __init__(self, contents, expectedDigest):
         """Parse a directory header out of a provided string; validate it
@@ -355,16 +404,25 @@ class _DirectoryHeader(mixminion.Config._ConfigFile):
         self.expectedDigest = expectedDigest
         mixminion.Config._ConfigFile.__init__(self, string=contents)
 
-    def validate(self, sections, entries, lines, contents):
-        direc = sections['Directory']
-        if direc['Version'] != "0.1":
+    def prevalidate(self, contents):
+        for name, ents in contents:
+            if name == 'Directory':
+                for k,v,_ in ents:
+                    if k == 'Version' and v.strip() != '0.2':
+                        raise ConfigError("Unrecognized directory version")
+
+        return contents
+
+    def validate(self, lines, contents):
+        direc = self['Directory']
+        if direc['Version'] != "0.2":
             raise ConfigError("Unrecognized directory version")
         if direc['Published'] > time.time() + 600:
             raise ConfigError("Directory published in the future")
         if direc['Valid-Until'] <= direc['Valid-After']:
             raise ConfigError("Directory is never valid")
 
-        sig = sections['Signature']
+        sig = self['Signature']
         identityKey = sig['DirectoryIdentity']
         identityBytes = identityKey.get_modulus_bytes()
         if not (MIN_IDENTITY_BYTES <= identityBytes <= MAX_IDENTITY_BYTES):
@@ -403,7 +461,7 @@ def _cleanForDigest(s):
     s = _abnormal_line_ending_re.sub("\n", s)
     s = _trailing_whitespace_re.sub("", s)
     s = _leading_whitespace_re.sub("", s)
-    if s[-1] != "\n":
+    if not s.endswith("\n"):
         s += "\n"
     return s
 
