@@ -1,5 +1,5 @@
 # Copyright 2002-2003 Nick Mathewson.  See LICENSE for licensing information.
-# $Id: benchmark.py,v 1.26 2003/01/05 13:19:53 nickm Exp $
+# $Id: benchmark.py,v 1.42 2003/06/06 07:54:46 nickm Exp $
 
 """mixminion.benchmark
 
@@ -13,20 +13,24 @@
 __pychecker__ = 'no-funcdoc no-reimport'
 __all__ = [ 'timeAll', 'testLeaks1', 'testLeaks2' ]
 
+import gc
 import os
 import stat
 import cPickle
+import threading
 from time import time
 
 import mixminion._minionlib as _ml
+import mixminion.server.ServerQueue
+
 from mixminion.BuildMessage import _buildHeader, buildForwardMessage, \
-     compressData, uncompressData
-from mixminion.Common import secureDelete, installSignalHandlers, \
-     waitForChildren, formatBase64
+     compressData, uncompressData, _encodePayload, decodePayload
+from mixminion.Common import secureDelete, installSIGCHLDHandler, \
+     waitForChildren, formatBase64, Lockfile
 from mixminion.Crypto import *
 from mixminion.Crypto import OAEP_PARAMETER
 from mixminion.Crypto import _add_oaep_padding, _check_oaep_padding
-from mixminion.Packet import SMTP_TYPE
+from mixminion.Packet import SMTP_TYPE, CompressedDataTooLong, IPV4Info
 from mixminion.ServerInfo import ServerInfo
 from mixminion.server.HashLog import HashLog
 from mixminion.server.PacketHandler import PacketHandler
@@ -52,7 +56,7 @@ def timeit_(fn, iters, ov=1):
     if iters < 1: iters = 1
     nones = [None]*iters
     if ov:
-        overhead = loop_overhead.get(iters, None)
+        overhead = loop_overhead.get(iters)
         if overhead is None:
             overhead = loop_overhead[iters] = timeit_((
                 lambda:(lambda:None)()), iters, 0)
@@ -81,9 +85,9 @@ def timeit(fn,times):
 
 def spacestr(n):
     """Converts number of bytes to readable representation)"""
-    if abs(n) < 1e4:
-        return "%d bytes" %n
-    elif abs(n) < 1e7:
+    if abs(n) < 1024:
+        return "%d B" %n
+    elif abs(n) < 1048576:
         return "%d KB" % (n >> 10)
     elif abs(n) < 1e10:
         return "%d MB" % (n >> 20)
@@ -221,6 +225,8 @@ def cryptoTiming():
     print "bear D (32K)", timeit((
         lambda bkey=bkey: bear_decrypt(s32K, bkey)), 100)
 
+def rsaTiming():
+    c = AESCounterPRNG()
     if hasattr(_ml, 'add_oaep_padding'):
         print "OAEP_add (70->128B) (C)",
         print timeit((lambda: _ml.add_oaep_padding(s70b,OAEP_PARAMETER,128)),
@@ -263,7 +269,7 @@ def cryptoTiming():
 
     print "RSA generate (1024 bit,e=65535)", timeit((lambda: pk_generate(1024,
                                                                   65535)),10)
-    rsa = pk_generate()
+    rsa = pk_generate(1024,65535)
     print "Pad+RSA public encrypt",
     print timeit((lambda rsa=rsa: pk_encrypt(s70b, rsa)),1000)
     enc = pk_encrypt(s70b, rsa)
@@ -272,7 +278,17 @@ def cryptoTiming():
 
     print "RSA generate (1024 bit,e=3)", timeit((lambda: pk_generate(1024,
                                                                   3)),10)
-    rsa = pk_generate()
+    rsa = pk_generate(1024,3)
+    print "Pad+RSA public encrypt",
+    print timeit((lambda rsa=rsa: pk_encrypt(s70b, rsa)),1000)
+    enc = pk_encrypt(s70b, rsa)
+    print "Pad+RSA private decrypt", \
+          timeit((lambda enc=enc,rsa=rsa: pk_decrypt(enc, rsa)),100)
+
+    print "RSA generate (1024 bit,e=100073471)", timeit(
+        lambda: pk_generate(1024, 100073471), 10)
+                             
+    rsa = pk_generate(1024, 100073471)
     print "Pad+RSA public encrypt",
     print timeit((lambda rsa=rsa: pk_encrypt(s70b, rsa)),1000)
     enc = pk_encrypt(s70b, rsa)
@@ -320,6 +336,7 @@ def _hashlogTiming(fname, load):
     h = HashLog(fname, "A")
     hashes = [ prng.getBytes(20) for _ in xrange(load) ]
 
+    # XXXX Check under different circumstances -- different sync patterns.
     t = time()
     for n in xrange(len(hashes)):
         h.logHash(hashes[n])
@@ -378,7 +395,7 @@ IP: 1.1.1.1
     keyring = ServerKeyring(config)
     keyring.getIdentityKey()
     print "Create and sign server descriptor", timeit(keyring.createKeys, 10)
-    liveKey = keyring.getServerKeyset()
+    liveKey = keyring.getServerKeysets()[0]
     descFile = liveKey.getDescriptorFileName()
     desc = open(descFile).read()
 ##     for _ in xrange(2000):
@@ -403,7 +420,7 @@ IP: 1.1.1.1
 
 def buildMessageTiming():
     print "#================= BUILD MESSAGE ====================="
-    pk = pk_generate()
+    pk = pk_generate(2048)
 
     for payload in "Hello!!!"*128, "Hello!!!"*(128*28):
         print "Compress %sK" % (len(payload)/1024), \
@@ -459,6 +476,70 @@ def buildMessageTiming():
     bm(8,8,20)
     bm(16,16,10)
 
+#----------------------------------------------------------------------
+def serverQueueTiming():
+    print "#================= SERVER QUEUES ====================="
+    Queue = mixminion.server.ServerQueue.Queue
+    DeliveryQueue = mixminion.server.ServerQueue.DeliveryQueue
+    d1 = mix_mktemp()
+    q1 = Queue(d1, create=1)
+
+    d2 = mix_mktemp()
+    os.mkdir(d2,0700)
+    getCommonPRNG().getBytes(1)
+    
+    #for ln,it in (32*1024,100),(128,400),(1024,400), (32*1024,100):
+    for ln,it in ():
+        msg = "z"*ln
+        def y(msg=msg,idx=[0],d2=d2):
+            fn = os.path.join(d2,"k_"+str(idx[0]))
+            idx[0] += 1
+            f = open(fn, 'wb')
+            f.write(msg)
+            f.close()
+        def x(msg=msg,d2=d2):
+            f,b=getCommonPRNG().openNewFile(d2,"k_",1)
+            f.write(msg)
+            f.close()
+        print "Base: write %s file: %s" %(
+            spacestr(ln), timestr(timeit_(y, it)))
+        for p in os.listdir(d2):
+            os.unlink(os.path.join(d2,p))
+        print "Base: write %s file with random name: %s" %(
+            spacestr(ln), timestr(timeit_(x, it)))
+        for p in os.listdir(d2):
+            os.unlink(os.path.join(d2,p))
+
+        tm = timeit_(lambda q1=q1,msg=msg:q1.queueMessage(msg),  it)
+        print "Queue %s message: %s" %(spacestr(ln), timestr(tm))
+        t2 = time()
+        q1.removeAll()
+        q1.cleanQueue()
+        t2 = time() - t2
+        print "Scrub %s message: %s" %(spacestr(ln), timestr(t2/float(it)))
+
+        msg = [ 123, 414, msg ]
+        tm = timeit_(lambda q1=q1,msg=msg:q1.queueObject(msg), it)
+        print "Pickle %s message: %s" %(spacestr(ln), timestr(tm))
+        q1.removeAll()
+        q1.cleanQueue()
+
+    for ln,it in (128,400),(1024,400), (32*1024,100):
+        q1 = DeliveryQueue(d1, [100,100,100,100])
+        msg = "z"*ln
+        print "Delivery queue: %s message: %s" %(
+            spacestr(ln),
+            timeit(lambda q1=q1,msg=msg: q1.queueDeliveryMessage(msg), it))
+        print "            (repOK):", \
+              timeit(lambda q1=q1: q1._repOk(), it*10)
+#        q1._bs2()
+#        print "          (set metadata 2):", \
+#              timeit(lambda q1=q1: q1._saveState2(), it)
+
+        for p in os.listdir(d1):
+            os.unlink(os.path.join(d1,p))
+
+
 #----------------------------------------------------------------------
 class DummyLog:
     def seenHash(self,h): return 0
@@ -467,9 +548,9 @@ class DummyLog:
 def serverProcessTiming():
     print "#================= SERVER PROCESS ====================="
 
-    pk = pk_generate()
+    pk = pk_generate(2048)
     server = FakeServerInfo("127.0.0.1", 1, pk, "X"*20)
-    sp = PacketHandler(pk, DummyLog())
+    sp = PacketHandler([pk], [DummyLog()])
 
     m_noswap = buildForwardMessage("Hello world", SMTP_TYPE, "f@invalid",
                                    [server, server], [server, server])
@@ -483,6 +564,31 @@ def serverProcessTiming():
     print "Server process (swap, no log)", timeit(
         lambda sp=sp, m_swap=m_swap: sp.processMessage(m_swap), 100)
 
+def encodingTiming():
+    print "#=============== END-TO-END ENCODING =================="
+    shortP = "hello world"
+    prng = AESCounterPRNG()
+    p = _encodePayload(shortP, 0, prng)
+    t = prng.getBytes(20)
+    print "Decode short payload", timeit(
+        lambda p=p,t=t: decodePayload(p, t), 1000)
+
+    k20 = prng.getBytes(20*1024)
+    p = _encodePayload(k20, 0, prng)
+    t = prng.getBytes(20)
+    print "Decode 20K payload", timeit(
+        lambda p=p,t=t: decodePayload(p, t), 1000)
+
+    comp = "x"*(20*1024)
+    p = _encodePayload(comp, 0, prng)
+    t = prng.getBytes(20)
+    def decode(p=p,t=t):
+        try:
+            decodePayload(p,t)
+        except CompressedDataTooLong:
+            pass
+    print "Decode overcompressed payload", timeit(decode, 1000)
+
 #----------------------------------------------------------------------
 def timeEfficiency():
     print "#================= ACTUAL v. IDEAL ====================="
@@ -522,7 +628,7 @@ def timeEfficiency():
     print "              short xor: %3.1f%%" % (100*2*strxor_20b/lioness_e)
 
     ##### SERVER PROCESS
-    pk = pk_generate(1024)
+    pk = pk_generate(2048)
 
     # Typical (no swap) server process is:
     #  pk_decrypt (128b)
@@ -552,7 +658,7 @@ def timeEfficiency():
     prng_128b = timeit_((lambda k=aeskey: prng(k,128)),10000)
 
     server = FakeServerInfo("127.0.0.1", 1, pk, "X"*20)
-    sp = PacketHandler(pk, DummyLog())
+    sp = PacketHandler([pk], [DummyLog()])
 
     m_noswap = buildForwardMessage("Hello world", SMTP_TYPE, "f@invalid",
                                    [server, server], [server, server])
@@ -580,11 +686,19 @@ def timeEfficiency():
 
 def fileOpsTiming():
     print "#================= File ops ====================="
-    installSignalHandlers(child=1,hup=0,term=0)
+    installSIGCHLDHandler()
     dname = mix_mktemp(".d")
 
-
     os.mkdir(dname)
+
+    lockfile = Lockfile(os.path.join("dname"))
+    t1 = time()
+    for _ in xrange(2000):
+        lockfile.acquire(blocking=1)
+        lockfile.release()
+    t = time()-t1
+    print "Lockfile: lock+unlock", timestr(t/2000.)
+
     for i in xrange(200):
         f = open(os.path.join(dname, str(i)), 'wb')
         f.write(s32K)
@@ -618,6 +732,7 @@ def testLeaks1():
     keytxt="a"*16
     key = _ml.aes_key(keytxt)
     while 1:
+        _ml.aes_key(keytxt)
         _ml.sha1(s20k)
         _ml.aes_ctr128_crypt(key,s20k,0)
         _ml.aes_ctr128_crypt(key,s20k,2000)
@@ -656,23 +771,216 @@ def testLeaks2():
     p = pk_generate(512)
     n,e = p.get_public_key()
 
+    f = open("/dev/null", 'w')
     while 1:
-        if 1:
+        if 0:
             p = pk_generate(512)
+        if 1:
             pk_decrypt(pk_encrypt(s20,p),p)
             for public in (0,1):
                 x = p.encode_key(public)
                 _ml.rsa_decode_key(x,public)
             p.get_public_key()
             _ml.rsa_make_public_key(n,e)
+            p.get_modulus_bytes()
+            p.get_exponent()
+        if 1:
+            p.PEM_write_key(f, 1)
+            p.PEM_write_key(f, 0)
+            p.PEM_write_key(f, 0, "Z")
+        if 1:
+            x = p.crypt("A"*64, 1, 1)
+            p.crypt(x, 0, 0)
+
+def testLeaks3():
+    print "Trying to leak (certgen)"
+    p = pk_generate(512)
+    p2 = pk_generate(512)
+    fn = mix_mktemp()
+    while 1:
+        _ml.generate_cert(fn, p, p2, "A", "B", 100, 10000)
+
+def testLeaks4():
+    print "Trying to leak (SSL)"
+
+    p = pk_generate(512)
+    p2 = pk_generate(512)
+    fn = mix_mktemp()
+    dh = mix_mktemp()
+    _ml.generate_cert(fn, p, p2, "A", "B", 100, 10000)
+    dh_fname = os.environ.get("MM_TEST_DHPARAMS")
+    if dh_fname and os.path.exists(dh_fname):
+        dh = dh_fname
+    elif dh_fname:
+        _ml.generate_dh_parameters(dh_fname, 1, 512)
+        dh = dh_fname
+    else:
+        _ml.generate_dh_parameters(dh, 1, 512)
+    print "OK"
+    context = _ml.TLSContext_new(fn, p, dh)
+    while 1:
+        if 1:
+            context = _ml.TLSContext_new(fn, p, dh)
+            _ = context.sock(0, 0)
+            _ = context.sock(0, 1)
+
+def testLeaks5():
+    from mixminion.test import _getMMTPServer
+    server, listener, messagesIn, keyid = _getMMTPServer(1)
+    #t = threading.Thread(None, testLeaks5_send,
+    #                     args=(keyid,))
+    #t.start()
+
+    while 1:
+        server.process(0.5)
+        #if messagesIn:
+        #    print "Connections"
+        del messagesIn[:]
+    #t.join()
+
+def testLeaks5_send():
+    from mixminion.test import TEST_PORT
+    import mixminion.MMTPClient
+    routing = IPV4Info("127.0.0.1", TEST_PORT, None)
+
+    #msg = "X" * 32 * 1024
+    n = 0
+    while 1:
+        mixminion.MMTPClient.sendMessages(routing, ["Z"*(32*1024)])
+        n += 1
+        print n, "sent"
+
+
+
+def testLeaks5_send2():
+    from mixminion.test import _getMMTPServer
+    from mixminion.test import TEST_PORT, _getTLSContext
+    import mixminion.MMTPClient
+
+    #msg = "X" * 32 * 1024
+    n = 0
+    server, listener, messagesIn, keyid = _getMMTPServer(1,port=(TEST_PORT+1))
+    #t = threading.Thread(None, testLeaks5_send,
+    #                     args=(keyid,))
+    #t.start()
+
+    sending = [0]
+    def sentHook(sending=sending):
+        sending[0]=0
+
+    certcache = mixminion.MMTPClient.PeerCertificateCache()
+
+    print len(gc.get_objects())
+
+    import socket
+
+    context = _getTLSContext(0)
+
+    i = 0
+    while 0:
+        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+        sock.setblocking(1)
+        sock.connect(("127.0.0.1", TEST_PORT))
+        tls = _getTLSContext(0).sock(sock)
+        tls.connect()
+        #tls.check_cert_alive()
+        #tls.verify_cert_and_get_identity_pk()
+        #tls.get_peer_cert_pk()
+        certcache.check(tls, keyid, ("127.0.0.1", TEST_PORT))
+        #print certcache.cache
+        
+        tls.shutdown()
+        sock.close()
+            
+
+
+    while 1:
+        clientcon = mixminion.server.MMTPServer.MMTPClientConnection(
+            _getTLSContext(0), "127.0.0.1", TEST_PORT, keyid,
+            ["X"*(32*1024), "JUNK"], ["z", None],
+            finishedCallback=sentHook, certCache=certcache)
+        clientcon.register(server)
+        i += 1
+        sending[0] = 1
+        print "Sending",i
+        while sending[0]:
+            server.process(0.5)
+
+        #pprint.pprint( clientcon.__dict__ )
+        old = clientcon
+        clientcon = None
+        gc.collect()
+        #print len(certcache.cache)
+        print len(gc.get_objects())
+        print len(server.readers), len(server.writers)
+        print gc.get_referrers(old)
+
+
+def testLeaks6():
+    import socket
+    p = pk_generate(512)
+    p2 = pk_generate(512)
+    fn = mix_mktemp()
+    dh = mix_mktemp()
+    _ml.generate_cert(fn, p, p2, "A", "B", 100, 10000)
+    dh_fname = os.environ.get("MM_TEST_DHPARAMS", None)
+    if dh_fname and os.path.exists(dh_fname):
+        dh = dh_fname
+    elif dh_fname:
+        _ml.generate_dh_parameters(dh_fname, 1, 512)
+        dh = dh_fname
+    else:
+        _ml.generate_dh_parameters(dh, 1, 512)
+    print "OK"
+    context = _ml.TLSContext_new(fn, p)#XXXX, dh)
+
+    listenSock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+    listenSock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
+    listenSock.bind(("127.0.0.1", 48999))
+    listenSock.listen(5)
+    while 1:
+        con, address = listenSock.accept()
+        tls = context.sock(con, serverMode=1)
+        tls.accept()
+        while 1:
+            r = tls.read(50)
+            if r == 0:
+                break
+        while 1:
+            r = tls.shutdown()
+            if r == 1:
+                break
+        con.close()
+
+def testLeaks6_2():
+    import socket
+    context = _ml.TLSContext_new()
+    m = "X"*99*1024
+    while 1:
+        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+        sock.connect(("127.0.0.1", 48999))
+        tls = context.sock(sock)
+        tls.connect()
+        tls.write(m)
+        while 1:
+            r = tls.shutdown()
+            if r: break
+        tls.shutdown()
+        sock.close()
 
 #----------------------------------------------------------------------
-
 def timeAll(name, args):
+    if 0:
+        testLeaks5_send()
+        return
+    
     cryptoTiming()
+    rsaTiming()
     buildMessageTiming()
     directoryTiming()
     fileOpsTiming()
+    encodingTiming()
+    serverQueueTiming()
     serverProcessTiming()
     hashlogTiming()
     timeEfficiency()
