@@ -1,5 +1,5 @@
 # Copyright 2002-2003 Nick Mathewson.  See LICENSE for licensing information.
-# $Id: MMTPServer.py,v 1.12.2.1 2003/01/10 19:44:46 nickm Exp $
+# $Id: MMTPServer.py,v 1.39 2003/06/06 07:17:35 nickm Exp $
 """mixminion.MMTPServer
 
    This package implements the Mixminion Transfer Protocol as described
@@ -8,9 +8,7 @@
    of the protocol.
 
    If you just want to send messages into the system, use MMTPClient.
-
-   FFFF As yet unsupported are: Session resumption, key renegotiation,
-   FFFF: Also unsupported: timeouts."""
+   """
 
 # NOTE FOR THE CURIOUS: The 'asyncore' module in the standard library
 #    is another general select/poll wrapper... so why are we using our
@@ -29,9 +27,11 @@ import time
 from types import StringType
 
 import mixminion._minionlib as _ml
-from mixminion.Common import MixError, MixFatalError, LOG, stringContains
-from mixminion.Crypto import sha1
+from mixminion.Common import MixError, MixFatalError, MixProtocolError, \
+     LOG, stringContains
+from mixminion.Crypto import sha1, getCommonPRNG
 from mixminion.Packet import MESSAGE_LEN, DIGEST_LEN
+from mixminion.MMTPClient import PeerCertificateCache
 
 __all__ = [ 'AsyncServer', 'ListenConnection', 'MMTPServerConnection',
             'MMTPClientConnection' ]
@@ -91,12 +91,12 @@ class AsyncServer:
     def hasReader(self, reader):
         """Return true iff 'reader' is a reader on this server."""
         fd = reader.fileno()
-        return self.readers.get(fd, None) is reader
+        return self.readers.get(fd) is reader
 
     def hasWriter(self, writer):
         """Return true iff 'writer' is a writer on this server."""
         fd = writer.fileno()
-        return self.writers.get(fd, None) is writer
+        return self.writers.get(fd) is writer
 
     def registerReader(self, reader):
         """Register a connection as a reader.  The connection's 'handleRead'
@@ -131,10 +131,12 @@ class AsyncServer:
         if r: del self.readers[fd]
         if w: del self.writers[fd]
 
-    def tryTimeout(self, now):
+    def tryTimeout(self, now=None):
         """Timeout any connection that is too old."""
         if self._timeout is None:
             return
+        if now is None:
+            now = time.time()
         # All connections older than 'cutoff' get purged.
         cutoff = now - self._timeout
         # Maintain a set of filenos for connections we've checked, so we don't
@@ -233,8 +235,9 @@ class SimpleTLSConnection(Connection):
     #    __con: an underlying TLS object
     #    __state: a callback to use whenever we get a read or a write. May
     #           throw _ml.TLSWantRead or _ml.TLSWantWrite.  See __acceptFn,
-    #           __connectFn, __shutdownFn, __readFn, __writeFn.
-    #    __server: an AsyncServer.
+    #           __connectFn, __shutdownFn, __readFn, __writeFn.  If __state
+    #           is None, the connection should close immediately.
+    #    __server: an AsyncServer.  If None, the connection is closed.
     #    __inbuf: A list of strings that we've read since the last expectRead.
     #    __inbuflen: The total length of all the strings in __inbuf
     #    __expectReadLen: None, or the number of bytes to read before
@@ -242,6 +245,11 @@ class SimpleTLSConnection(Connection):
     #    __terminator: None, or a string which will terminate the current read.
     #    __outbuf: None, or the remainder of the string we're currently
     #           writing.
+    #    __servermode: If true, we're the server side of the connection.
+    #           Else, we're the client side.
+    #    __connection: Are we currently trying to start a connection? (boolean)
+    #    __failed: Have we given up on this connection?
+
     def __init__(self, sock, tls, serverMode, address=None):
         """Create a new SimpleTLSConnection.
 
@@ -253,16 +261,22 @@ class SimpleTLSConnection(Connection):
         self.__con = tls
         self.fd = self.__con.fileno()
         self.lastActivity = time.time()
+        self.__serverMode = serverMode
+        self.__failed = 0
+        self.__inbuf = []
+        self.__inbuflen = 0
 
         if serverMode:
+            self.__connecting = 0
             self.__state = self.__acceptFn
         else:
+            self.__connecting = 1
             self.__state = self.__connectFn
 
         if address is not None:
-            self.address = address
+            self.address = "%s (fd %s)" % (address, self.fd)
         else:
-            self.address = "remote host"
+            self.address = "remote host (fd %s)" % self.fd
 
     def isShutdown(self):
         """Returns true iff this connection is finished shutting down"""
@@ -287,7 +301,7 @@ class SimpleTLSConnection(Connection):
            'bytes' bytes.  If terminator is provided, we read until we
            encounter the terminator, but give up after 'bytes' bytes.
         """
-        self.__inbuf = []
+        del self.__inbuf[:]
         self.__inbuflen = 0
         self.__expectReadLen = bytes
         self.__terminator = terminator
@@ -314,6 +328,7 @@ class SimpleTLSConnection(Connection):
         """Hook to implement client-side handshake."""
         self.__con.connect() #may throw want*
         self.__server.unregister(self)
+        self.__connecting = 0
         self.finished()
 
     def __shutdownFn(self):
@@ -326,9 +341,8 @@ class SimpleTLSConnection(Connection):
         # ourself.
         if self.__con.shutdown() == 1: #may throw want*
             #trace("Got a 1 on shutdown (fd %s)", self.fd)
-            self.__server.unregister(self)
-            self.__state = None
             self.__sock.close()
+            self.__state = None
             self.shutdownFinished()
             return
 
@@ -389,17 +403,31 @@ class SimpleTLSConnection(Connection):
         if len(out) == 0:
             self.finished()
 
+    def __handshakeFn(self):
+        """Callback used when we're renegotiating the connection key.  Must
+           only be used from client mode."""
+        assert not self.__serverMode
+        self.__con.do_handshake() #may throw want*
+        self.__server.unregister(self)
+        self.finished()
+
+    def startRenegotiate(self):
+        """Begin renegotiation the connection key.  Must only be called from
+           client mode."""
+        self.__con.renegotiate() # Succeeds immediately.
+        self.__state = self.__handshakeFn
+        self.__server.registerBoth(self) #????
+
     def tryTimeout(self, cutoff):
         if self.lastActivity <= cutoff:
-            warn("Socket %s to %s timed out", self.fd, self.address)
+            warn("Connection to %s timed out", self.address)
             # ????     I'm not sure this is right.  Instead of just killing
             # ???? the socket, should we shut down the SSL too?
-            # ????     Also, should we handle timeout as a separate kind of
-            # ???? error from a hooks point of view.
-            self.__server.unregister(self)
-            self.__state = None
             self.__sock.close()
-            self.handleFail(1)
+            if not self.__failed:
+                self.__failed = 1
+                self.handleFail(1)
+            self.remove()
 
     def handleRead(self):
         self.__handleAll()
@@ -426,22 +454,32 @@ class SimpleTLSConnection(Connection):
         except _ml.TLSWantRead:
             self.__server.registerReader(self)
         except _ml.TLSClosed:
-            warn("Unexpectedly closed connection to %s", self.address)
-            self.handleFail(retriable=1)
+            if self.__connecting:
+                warn("Couldn't connect to server %s", self.address)
+            else:
+                warn("Unexpectedly closed connection to %s", self.address)
             self.__sock.close()
-            self.__server.unregister(self)
+            if not self.__failed:
+                self.__failed = 1
+                self.handleFail(retriable=1)
+            self.remove()
         except _ml.TLSError, e:
             if self.__state != self.__shutdownFn:
                 warn("Unexpected error: %s. Closing connection to %s.",
                      e, self.address)
                 self.shutdown(err=1, retriable=1)
+                self.__handleAll() # Try another round of the loop.
             else:
                 warn("Error while shutting down: closing connection to %s",
                      self.address)
-                self.__server.unregister(self)
+                self.__sock.close()
+                if not self.__failed:
+                    self.__failed = 1
+                    self.handleFail(1)
+                self.remove()
         else:
-            # We are in no state at all.
-            self.__server.unregister(self)
+            # We are in no state at all; disconnect
+            self.remove()
 
     def finished(self):
         """Called whenever a connect, accept, read, or write is finished."""
@@ -451,10 +489,18 @@ class SimpleTLSConnection(Connection):
         """Called when this connection is successfully shut down."""
         pass
 
+    def shutdownFailed(self):
+        """Called when this connection goes down hard."""
+        pass
+
     def shutdown(self, err=0, retriable=0):
         """Begin a shutdown on this connection"""
-        if err:
+        if err and not self.__failed:
             self.handleFail(retriable)
+            #self.__sock.close()
+            #self.__state = None
+            #return
+            
         self.__state = self.__shutdownFn
 
     def fileno(self):
@@ -464,18 +510,40 @@ class SimpleTLSConnection(Connection):
         """Returns the current contents of the input buffer."""
         return "".join(self.__inbuf)
 
-    def getPeerPK(self):
-        return self.__con.get_peer_cert_pk()
+    def pullInput(self):
+        """Returns the current contents of the input buffer, and clears the
+           input buffer."""
+        inp = "".join(self.__inbuf)
+        del self.__inbuf[:]
+        self.__inbuflen = 0
+        return inp
+
+    def getTLSConnection(self):
+        return self.__con
 
     def handleFail(self, retriable=0):
         """Called when we shutdown with an error."""
         pass
 
+    def remove(self):
+        """Called when this connection is shut down successfully or closed
+           with an error.  Removes all state associated with this connection.
+           """
+        self.__server.unregister(self)
+        self.__server = None
+
+        # Under heavy loads, having circular references through __state and
+        # __finished can keep the connection object alive for many garbage
+        # collections.  Let's nuke those so it is deleted right away.
+        self.__state = None
+        self.finished = None
+
+
 #----------------------------------------------------------------------
 # Implementation for MMTP.
 
 # The protocol string to send.
-PROTOCOL_STRING      = "MMTP 0.1\r\n"
+PROTOCOL_STRING      = "MMTP 0.3\r\n"
 # The protocol specification to expect.
 PROTOCOL_RE          = re.compile("MMTP ([^\s\r\n]+)\r\n")
 # Control line for sending a message.
@@ -484,6 +552,8 @@ SEND_CONTROL         = "SEND\r\n"
 JUNK_CONTROL         = "JUNK\r\n"
 # Control line for acknowledging a message
 RECEIVED_CONTROL     = "RECEIVED\r\n"
+# Control line for refusing a message
+REJECTED_CONTROL     = "REJECTED\r\n"
 SEND_CONTROL_LEN     = len(SEND_CONTROL)
 RECEIVED_CONTROL_LEN = len(RECEIVED_CONTROL)
 SEND_RECORD_LEN      = len(SEND_CONTROL) + MESSAGE_LEN + DIGEST_LEN
@@ -496,14 +566,30 @@ class MMTPServerConnection(SimpleTLSConnection):
     # messageConsumer: a function to call with all received messages.
     # finished: callback when we're done with a read or write; see
     #     SimpleTLSConnection.
-    def __init__(self, sock, tls, consumer):
+    # protocol: The MMTP protocol version we're currently using, or None
+    #     if negotiation hasn't completed.
+    # PROTOCOL_VERSIONS: (static) a list of protocol versions we allow,
+    #     in decreasing order of preference.
+    # junkCallback: a no-arguments function called when we receive a
+    #     junk packet.
+    # rejectCallback: a no-arguments function called when we reject a packet.
+    # rejectPackets: a flag: should we reject incoming packets?
+    PROTOCOL_VERSIONS = [ '0.3' ]
+    def __init__(self, sock, tls, consumer, rejectPackets=0):
         """Create an MMTP connection to receive messages sent along a given
            socket.  When valid packets are received, pass them to the
-           function 'consumer'."""
+           function 'consumer'.  If rejectPackets is true, then instead of
+           accepting packets, we refuse them instead--for example, if the
+           because the disk is full."""
         SimpleTLSConnection.__init__(self, sock, tls, 1,
                                      "%s:%s"%sock.getpeername())
+        
         self.messageConsumer = consumer
+        self.junkCallback = lambda : None
+        self.rejectCallback = lambda : None
         self.finished = self.__setupFinished
+        self.protocol = None
+        self.rejectPackets = rejectPackets
 
     def __setupFinished(self):
         """Called once we're done accepting.  Begins reading the protocol
@@ -516,49 +602,61 @@ class MMTPServerConnection(SimpleTLSConnection):
         """Called once we're done reading the protocol string.  Either
            rejects, or sends our response.
         """
-        trace("done w/ client sendproto (fd %s)", self.fd)
-        inp = self.getInput()
+        trace("done w/ client sendproto to %s", self.address)
+        inp = self.pullInput()
         m = PROTOCOL_RE.match(inp)
 
         if not m:
-            warn("Bad protocol list.  Closing connection to %s", self.address)
-            self.shutdown(err=1)
-            return
-        protocols = m.group(1).split(",")
-        if "0.1" not in protocols:
-            warn("Unsupported protocol list.  Closing connection to %s",
+            warn("Bad protocol list: %r.  Closing connection to %s", inp,
                  self.address)
             self.shutdown(err=1)
             return
-        else:
-            trace("protocol ok (fd %s)", self.fd)
-            self.finished = self.__sentProtocol
-            self.beginWrite(PROTOCOL_STRING)
+        protocols = m.group(1).split(",")
+        for p in self.PROTOCOL_VERSIONS:
+            if p in protocols:
+                trace("Using protocol %s with %s", p, self.address)
+                self.protocol = p
+                self.finished = self.__sentProtocol
+                self.beginWrite("MMTP %s\r\n"% p)
+                return
+
+        warn("Unsupported protocol list.  Closing connection to %s",
+             self.address)
+        self.shutdown(err=1)
+        return
 
     def __sentProtocol(self):
         """Called once we're done sending our protocol response.  Begins
            reading a packet from the line.
         """
-        trace("done w/ server sendproto (fd %s)", self.fd)
+        trace("done w/ server sendproto to %s", self.address)
         self.finished = self.__receivedMessage
         self.expectRead(SEND_RECORD_LEN)
 
     def __receivedMessage(self):
         """Called once we've read a message from the line.  Checks the
            digest, and either rejects or begins sending an ACK."""
-        data = self.getInput()
+        data = self.pullInput()
         msg = data[SEND_CONTROL_LEN:-DIGEST_LEN]
         digest = data[-DIGEST_LEN:]
 
         if data.startswith(JUNK_CONTROL):
             expectedDigest = sha1(msg+"JUNK")
             replyDigest = sha1(msg+"RECEIVED JUNK")
+            replyControl = RECEIVED_CONTROL
+            isJunk = 1
         elif data.startswith(SEND_CONTROL):
             expectedDigest = sha1(msg+"SEND")
-            replyDigest = sha1(msg+"RECEIVED")
+            if self.rejectPackets:
+                replyDigest = sha1(msg+"REJECTED")
+                replyControl = REJECTED_CONTROL
+            else:
+                replyDigest = sha1(msg+"RECEIVED")
+                replyControl = RECEIVED_CONTROL
+            isJunk = 0
         else:
-            warn("Unrecognized command from %s.  Closing connection.",
-                 self.address)
+            warn("Unrecognized command (%r) from %s.  Closing connection.",
+                 data[:4], self.address)
             self.shutdown(err=1)
             return
         if expectedDigest != digest:
@@ -567,49 +665,102 @@ class MMTPServerConnection(SimpleTLSConnection):
             self.shutdown(err=1)
             return
         else:
-            debug("%s packet received from %s; Checksum valid.",
-                  data[:4], self.address)
+            if isJunk:
+                debug("Link padding received from %s; Checksum valid.",
+                      self.address)
+            else:
+                debug("Packet received from %s; Checksum valid.", self.address)
             self.finished = self.__sentAck
-            self.beginWrite(RECEIVED_CONTROL+replyDigest)
-            self.messageConsumer(msg)
+            self.beginWrite(replyControl+replyDigest)
+            if isJunk:
+                self.junkCallback()
+            elif self.rejectPackets:
+                self.rejectCallback()
+            else:
+                self.messageConsumer(msg)
 
     def __sentAck(self):
         """Called once we're done sending an ACK.  Begins reading a new
            message."""
-        debug("Send ACK for message from %s (fd %s)", self.address, self.fd)
-        #FFFF Rehandshake
+        debug("Send ACK for message from %s", self.address)
         self.finished = self.__receivedMessage
         self.expectRead(SEND_RECORD_LEN)
 
+    def remove(self):
+        self.messageConsumer = None
+        self.finished = None
+        self.junkCallback = None
+        self.rejectCallback = None
+        
+        SimpleTLSConnection.remove(self)
+
 #----------------------------------------------------------------------
 
-# FFFF We need to note retriable situations better.
+NULL_KEYID = "\x00"*20
+
 class MMTPClientConnection(SimpleTLSConnection):
     """Asynchronious implementation of the sending ("client") side of a
        mixminion connection."""
     ## Fields:
-    # ip, port, keyID, messageList, handleList, sendCallback, failCallback:
-    #   As described in the docstring for __init__ below.
+    # ip, port, keyID, messageList, handleList, sendCallback, failCallback,
+    # finishedCallback, certCache:
+    #   As described in the docstring for __init__ below.  We remove entries
+    #   from the front of messageList/handleList as we begin sending them.
+    # _curMessage, _curHandle: Correspond to the message and handle
+    #     that we are currently trying to deliver.
+    # junk: A list of 32KB padding chunks that we're going to send.  We
+    #   pregenerate these to avoid timing attacks.  They correspond to
+    #   the 'JUNK' entries in messageList.
+    # isJunk: flag.  Is the current chunk padding?
+    # expectedDigest: The digest we expect to receive in response to the
+    #   current chunk.
+    # protocol: The MMTP protocol version we're currently using, or None
+    #     if negotiation hasn't completed.
+    # PROTOCOL_VERSIONS: (static) a list of protocol versions we allow,
+    #     in the order we offer them.
+    # active: Are we currently able to send messages to a server?  Boolean.
+
+    PROTOCOL_VERSIONS = [ '0.3' ]
     def __init__(self, context, ip, port, keyID, messageList, handleList,
-                 sentCallback=None, failCallback=None):
+                 sentCallback=None, failCallback=None, finishedCallback=None,
+                 certCache=None):
         """Create a connection to send messages to an MMTP server.
            Raises socket.error if the connection fails.
-        
+
            ip -- The IP of the destination server.
            port -- The port to connect to.
            keyID -- None, or the expected SHA1 hash of the server's public key
-           messageList -- a list of message payloads to send
-           handleList -- a list of objects corresponding to the messages in
+           messageList -- a list of message payloads and control strings.
+               The control string "JUNK" sends 32KB of padding; the control
+               string "RENEGOTIATE" renegotiates the connection key.
+           handleList -- a list of objects corresponding to the entries in
               messageList.  Used for callback.
            sentCallback -- None, or a function of (msg, handle) to be called
               whenever a message is successfully sent.
            failCallback -- None, or a function of (msg, handle, retriable)
-              to be called when messages can't be sent."""
+              to be called when messages can't be sent.
+           finishedCallback -- None, or a function to be called when this
+              connection is closed.
+           certCache -- an instance of PeerCertificateCache to use for
+              checking server certificates.
+        """
+        # Generate junk before connecting to avoid timing attacks
+        self.junk = []
+        self.messageList = []
+        self.handleList = []
+        self.active = 1
+
+        self.addMessages(messageList, handleList)
+
+        if certCache is None:
+            certCache = PeerCertificateCache()
+        self.certCache = certCache
 
         sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
         sock.setblocking(0)
         self.keyID = keyID
         self.ip = ip
+        self.port = port
         try:
             sock.connect((ip, port))
         except socket.error, e:
@@ -621,30 +772,57 @@ class MMTPClientConnection(SimpleTLSConnection):
         tls = context.sock(sock)
 
         SimpleTLSConnection.__init__(self, sock, tls, 0, "%s:%s"%(ip,port))
-        self.messageList = messageList
-        self.handleList = handleList
         self.finished = self.__setupFinished
         self.sentCallback = sentCallback
         self.failCallback = failCallback
+        self.finishedCallback = finishedCallback
+        self.protocol = None
+        self._curMessage = self._curHandle = None
+        
+        debug("Opening client connection to %s", self.address)
 
-        debug("Opening client connection (fd %s)", self.fd)
+    def isActive(self):
+        """Return true iff messages added to this connection via addMessages
+           will be delivered.  isActive() will return false if, for example,
+           the connection is currently shutting down."""
+        return self.active
+
+    def addMessages(self, messages, handles):
+        """Given a list of messages and handles, as given to
+           MMTPServer.__init__, cause this connection to deliver that new
+           set of messages after it's done with those it's currently sending.
+        """
+        assert self.active
+        assert len(messages) == len(handles)
+        for m,h in zip(messages, handles):
+            if m in ("JUNK", "RENEGOTIATE"):
+                assert h is None
+        for m in messages:
+            if m == "JUNK":
+                self.junk.append(getCommonPRNG().getBytes(MESSAGE_LEN))
+        self.messageList.extend(messages)
+        self.handleList.extend(handles)
+
+    def getAddr(self):
+        """Return an (ip,port,keyID) tuple for this connection"""
+        return self.ip, self.port, self.keyID
 
     def __setupFinished(self):
         """Called when we're done with the client side negotations.
            Begins sending the protocol string.
         """
-        keyID = sha1(self.getPeerPK().encode_key(public=1))
-        if self.keyID is not None:
-            if keyID != self.keyID:
-                warn("Got unexpected Key ID from %s; shutting down connection",
-                     self.address)
-                # The keyid may start being good in a while.
-                self.shutdown(err=1,retriable=1)
-                return
-            else:
-                debug("KeyID from %s is valid", self.address)
+        try:
+            self.certCache.check(self.getTLSConnection(), self.keyID,
+                                 self.address)
+        except MixProtocolError, e:
+            warn("Certificate error: %s.  Shutting down connection to %s",
+                 e, self.address)
+            self.shutdown(err=1,retriable=1)
+            return
+        else:
+            debug("KeyID is valid from %s", self.address)
 
-        self.beginWrite(PROTOCOL_STRING)
+        self.beginWrite("MMTP %s\r\n"%(",".join(self.PROTOCOL_VERSIONS)))
         self.finished = self.__sentProtocol
 
     def __sentProtocol(self):
@@ -658,26 +836,50 @@ class MMTPClientConnection(SimpleTLSConnection):
         """Called when we're done receiving the protocol string.  Begins
            sending a packet, or exits if we're done sending.
         """
-        inp = self.getInput()
-        if inp != PROTOCOL_STRING:
-            warn("Invalid protocol.  Closing connection to %s", self.address)
-            # This isn't retriable; we don't talk to servers we don't
-            # understand.
-            self.shutdown(err=1,retriable=0)
-            return
+        inp = self.pullInput()
 
-        self.beginNextMessage()
+        for p in self.PROTOCOL_VERSIONS:
+            if inp == 'MMTP %s\r\n'%p:
+                trace("Speaking MMTP version %s with %s", p, self.address)
+                self.protocol = inp
+                self.beginNextMessage()
+                return
+
+        warn("Invalid protocol.  Closing connection to %s", self.address)
+             
+        # This isn't retriable; we don't talk to servers we don't
+        # understand.
+        self.shutdown(err=1,retriable=0)
+        return
 
     def beginNextMessage(self):
         """Start writing a message to the connection."""
         if not self.messageList:
             self.shutdown(0)
             return
-        msg = self.messageList[0]
-        self.expectedDigest = sha1(msg+"RECEIVED")
-        msg = SEND_CONTROL+msg+sha1(msg+"SEND")
-        assert len(msg) == SEND_RECORD_LEN
 
+        msg = self._curMessage = self.messageList[0]
+        self._curHandle = self.handleList[0]
+        del self.messageList[0]
+        del self.handleList[0]
+        if msg == 'RENEGOTIATE':
+            self.finished = self.beginNextMessage
+            self.startRenegotiate()
+            return
+        elif msg == 'JUNK':
+            msg = self.junk[0]
+            del self.junk[0]
+            self.expectedDigest = sha1(msg+"RECEIVED JUNK")
+            self.rejectDigest = sha1(msg+"REJECTED")
+            msg = JUNK_CONTROL+msg+sha1(msg+"JUNK")
+            self.isJunk = 1
+        else:
+            self.expectedDigest = sha1(msg+"RECEIVED")
+            self.rejectDigest = sha1(msg+"REJECTED")
+            msg = SEND_CONTROL+msg+sha1(msg+"SEND")
+            self.isJunk = 0
+
+        assert len(msg) == SEND_RECORD_LEN
         self.beginWrite(msg)
         self.finished = self.__sentMessage
 
@@ -685,7 +887,7 @@ class MMTPClientConnection(SimpleTLSConnection):
         """Called when we're done sending a message.  Begins reading the
            server's ACK."""
 
-        debug("Message delivered to %s (fd %s)", self.address, self.fd)
+        debug("Message delivered to %s", self.address)
         self.finished = self.__receivedAck
         self.expectRead(RECEIVED_RECORD_LEN)
 
@@ -698,52 +900,84 @@ class MMTPClientConnection(SimpleTLSConnection):
           If there are more messages to send, begins sending the next.
           Otherwise, begins shutting down.
        """
-       trace("received ack (fd %s)", self.fd)
-       # FFFF Rehandshake
-       inp = self.getInput()
-       if inp != (RECEIVED_CONTROL+self.expectedDigest):
+       trace("received ack from %s", self.address)
+       inp = self.pullInput()
+       rejected = 0
+       if inp == REJECTED_CONTROL+self.rejectDigest:
+           debug("Message rejected from %s", self.address)
+           rejected = 1
+       elif inp != (RECEIVED_CONTROL+self.expectedDigest):
            # We only get bad ACKs if an adversary somehow subverts TLS's
            # checksumming.  That's not fixable.
            self.shutdown(err=1,retriable=0)
            return
+       else:
+           debug("Received valid ACK for message from %s", self.address)
 
-       debug("Received valid ACK for message from %s", self.address)
-       justSent = self.messageList[0]
-       justSentHandle = self.handleList[0]
-       del self.messageList[0]
-       del self.handleList[0]
-       if self.sentCallback is not None:
-           self.sentCallback(justSent, justSentHandle)
+       if not self.isJunk:
+           if not rejected and self.sentCallback is not None:
+               self.sentCallback(self._curMessage, self._curHandle)
+           elif rejected and self.failCallback is not None:
+               self.failCallback(self._curMessage, self._curHandle,
+                                 retriable=1)
+
+       self._curMessage = self._curHandle = None
 
        self.beginNextMessage()
 
     def handleFail(self, retriable):
-        """Invoked when a message is not deliverable."""
+        """Invoked when we shutdown with an error."""
         if self.failCallback is not None:
+            if self._curHandle is not None:
+                self.failCallback(self._curMessage, self._curHandle, retriable)
             for msg, handle in zip(self.messageList, self.handleList):
+                if handle is None:
+                    continue
                 self.failCallback(msg,handle,retriable)
+        self._messageList = self.handleList = []
+        self._curMessage = self._curHandle = None
 
-LISTEN_BACKLOG = 10 # ???? Is something else more reasonable?
+    def shutdown(self, err=0, retriable=0):
+        self.active = 0
+        SimpleTLSConnection.shutdown(self, err=err, retriable=retriable)
+
+    def remove(self):
+        self.active = 0
+        if self.finishedCallback is not None:
+            self.finishedCallback()
+        self.finishedCallback = None
+        self.failCallback = None
+        self.sentCallback = None
+        
+        SimpleTLSConnection.remove(self)
+
+
+LISTEN_BACKLOG = 128
 class MMTPAsyncServer(AsyncServer):
     """A helper class to invoke AsyncServer, MMTPServerConnection, and
        MMTPClientConnection, with a function to add new connections, and
        callbacks for message success and failure."""
     ##
+    # context: a TLSContext object to use for newly received connections.
+    # clientConByAddr: A map from 3-tuples returned by MMTPClientConnection.
+    #     getAddr, to MMTPClientConnection objects.
+    # certificateCache: A PeerCertificateCache object.
+    # listener: A ListenConnection object.
+    # _timeout: The number of seconds of inactivity to allow on a connection
+    #     before formerly shutting it down.
     def __init__(self, config, tls):
         AsyncServer.__init__(self)
 
         self.context = tls
         # FFFF Don't always listen; don't always retransmit!
-        # FFFF Support listening on specific IPs
+        # FFFF Support listening on multiple IPs
 
-        if config['Incoming/MMTP'].get('ListenIP',None) is not None:
-            IP = config['Incoming/MMTP']['ListenIP']
-        else:
+        IP = config['Incoming/MMTP'].get('ListenIP')
+        if IP is None:
             IP = config['Incoming/MMTP']['IP']
 
-        if config['Incoming/MMTP'].get('ListenPort',None) is not None:
-            port = config['Incoming/MMTP']['ListenPort']
-        else:
+        port =  config['Incoming/MMTP'].get('ListenPort')
+        if port is None:
             port = config['Incoming/MMTP']['Port']
 
         self.listener = ListenConnection(IP, port,
@@ -751,11 +985,20 @@ class MMTPAsyncServer(AsyncServer):
                                          self._newMMTPConnection)
         #self.config = config
         self.listener.register(self)
-        self._timeout = config['Server']['Timeout'][2]
+        self._timeout = config['Server']['Timeout'].getSeconds()
+        self.clientConByAddr = {}
+        self.certificateCache = PeerCertificateCache()
 
-    def getNextTimeoutTime(self, now):
+    def setContext(self, context):
+        """Change the TLS context used for newly received connections.
+           Used to rotate keys."""
+        self.context = context
+
+    def getNextTimeoutTime(self, now=None):
         """Return the time at which we next purge connections, if we have
            last done so at time 'now'."""
+        if now is None:
+            now = time.time()
         return now + self._timeout
 
     def _newMMTPConnection(self, sock):
@@ -773,29 +1016,64 @@ class MMTPAsyncServer(AsyncServer):
 
     def sendMessages(self, ip, port, keyID, messages, handles):
         """Begin sending a set of messages to a given server."""
-        # ???? Can we remove these asserts yet?
+
         for m,h in zip(messages, handles):
+            if m in ("JUNK", "RENEGOTIATE"):
+                assert h is None
+                continue
             assert len(m) == MESSAGE_LEN
             assert len(h) < 32
 
         try:
+            # Is there an existing connection open to the right server?
+            con = self.clientConByAddr[(ip,port,keyID)]
+            # If so, is that connection currently sending messages?
+            if con.isActive():
+                LOG.debug("Queueing %s messages on open connection to %s",
+                          len(messages), con.address)
+                con.addMessages(messages, handles)
+                return
+        except KeyError:
+            pass
+
+        try:
+            # There isn't any connection to the right server. Open one...
+            addr = (ip, port, keyID)
+            finished = lambda addr=addr, self=self: self.__clientFinished(addr)
             con = MMTPClientConnection(self.context,
-                                       ip, port, keyID, messages, handles,
-                                       self.onMessageSent,
-                                       self.onMessageUndeliverable)
+                                     ip, port, keyID, messages, handles,
+                                     sentCallback=self.onMessageSent,
+                                     failCallback=self.onMessageUndeliverable,
+                                     finishedCallback=finished,
+                                     certCache=self.certificateCache)
             con.register(self)
+            # ...and register it in clientConByAddr
+            assert addr == con.getAddr()
+            self.clientConByAddr[addr] = con
         except socket.error, e:
             LOG.error("Unexpected socket error connecting to %s:%s: %s",
                       ip, port, e)
             for m,h in zip(messages, handles):
                 self.onMessageUndeliverable(m,h,1)
 
+    def __clientFinished(self, addr):
+        """Called when a client connection runs out of messages to send."""
+        try:
+            del self.clientConByAddr[addr]
+        except KeyError:
+            warn("Didn't find client connection to %s in address map",
+                 addr)
+
     def onMessageReceived(self, msg):
+        """Abstract function.  Called when we get a message"""
         pass
 
     def onMessageUndeliverable(self, msg, handle, retriable):
+        """Abstract function: Called when an attempt to deliver a
+           message fails."""
         pass
 
     def onMessageSent(self, msg, handle):
+        """Abstract function: Called when an attempt to deliver a
+           message succeeds."""
         pass
-
