@@ -1,5 +1,5 @@
 # Copyright 2002-2003 Nick Mathewson.  See LICENSE for licensing information.
-# $Id: Modules.py,v 1.20 2003/01/08 03:43:21 nickm Exp $
+# $Id: Modules.py,v 1.42 2003/06/05 18:41:40 nickm Exp $
 
 """mixminion.server.Modules
 
@@ -18,7 +18,7 @@ import re
 import sys
 import smtplib
 import socket
-import base64
+import threading
 
 if sys.version_info[:2] >= (2,3):
     import textwrap
@@ -28,12 +28,14 @@ else:
 import mixminion.BuildMessage
 import mixminion.Config
 import mixminion.Packet
-import mixminion.server.Queue
-from mixminion.Config import ConfigError, _parseBoolean, _parseCommand
+import mixminion.server.ServerQueue
+import mixminion.server.ServerConfig
+import mixminion.server.EventStats as EventStats
+from mixminion.Config import ConfigError, _parseBoolean, _parseCommand, \
+     _parseIntervalList
 from mixminion.Common import LOG, createPrivateDir, MixError, isSMTPMailbox, \
-     isPrintingAscii
-from mixminion.BuildMessage import CompressedDataTooLong
-from mixminion.Packet import ParseError
+     isPrintingAscii, readFile, waitForChildren
+from mixminion.Packet import ParseError, CompressedDataTooLong
 
 # Return values for processMessage
 DELIVER_OK = 1
@@ -58,12 +60,17 @@ class DeliveryModule:
         "Zero-argument constructor, as required by Module protocol."
         pass
 
+    def getRetrySchedule(self):
+        """Return a retry schedule for this module's queue, as specified
+           in ServerQueue.DeliveryQueue.setRetrySchedule."""
+        return None
+
     def getConfigSyntax(self):
         """Return a map from section names to section syntax, as described
            in Config.py"""
         raise NotImplementedError("getConfigSyntax")
 
-    def validateConfig(self, sections, entries, lines, contents):
+    def validateConfig(self, config, lines, contents):
         """See mixminion.Config.validate"""
         pass
 
@@ -101,11 +108,11 @@ class DeliveryModule:
            Otherwise, the message is either a reply or an encrypted
            forward message.
            """
-        return SimpleModuleDeliveryQueue(self, queueDir)
+        return SimpleModuleDeliveryQueue(self, queueDir,
+                                   retrySchedule=self.getRetrySchedule())
 
-    def processMessage(self, message, tag, exitType, exitInfo):
-        """Given a message with a given exitType and exitInfo, try to deliver
-           it.  'tag' is as decribed in createDeliveryQueue. Return one of:
+    def processMessage(self, packet):
+        """Given a DeliveryPacket object, try to delier it.  Return one of:
             DELIVER_OK (if the message was successfully delivered),
             DELIVER_FAIL_RETRY (if the message wasn't delivered, but might be
               deliverable later), or
@@ -118,60 +125,117 @@ class DeliveryModule:
 class ImmediateDeliveryQueue:
     """Helper class usable as delivery queue for modules that don't
        actually want a queue.  Such modules should have very speedy
-       processMessage() methods, and should never have deliery fail."""
+       processMessage() methods, and should never have delivery fail."""
     ##Fields:
     #  module: the underlying DeliveryModule object.
     def __init__(self, module):
         self.module = module
 
-    def queueDeliveryMessage(self, (exitType, address, tag), message):
+    def queueDeliveryMessage(self, packet, retry=0, lastAttempt=0):
         """Instead of queueing our message, pass it directly to the underlying
            DeliveryModule."""
         try:
-            res = self.module.processMessage(message, tag, exitType, address)
+            EventStats.log.attemptedDelivery() #FFFF
+            res = self.module.processMessage(packet)
             if res == DELIVER_OK:
-                return
+                EventStats.log.successfulDelivery() #FFFF
             elif res == DELIVER_FAIL_RETRY:
                 LOG.error("Unable to retry delivery for message")
+                EventStats.log.unretriableDelivery() #FFFF
             else:
                 LOG.error("Unable to deliver message")
+                EventStats.log.unretriableDelivery() #FFFF
         except:
             LOG.error_exc(sys.exc_info(),
                                "Exception delivering message")
+            EventStats.log.unretriableDeliery() #FFFF
+
+        return "<nil>"
 
     def sendReadyMessages(self):
         # We do nothing here; we already delivered the messages
         pass
 
-    def cleanQueue(self):
+    def cleanQueue(self, deleteFn=None):
         # There is no underlying queue to worry about here; do nothing.
         pass
 
-class SimpleModuleDeliveryQueue(mixminion.server.Queue.DeliveryQueue):
+class SimpleModuleDeliveryQueue(mixminion.server.ServerQueue.DeliveryQueue):
     """Helper class used as a default delivery queue for modules that
        don't care about batching messages to like addresses."""
     ## Fields:
     # module: the underlying module.
-    def __init__(self, module, directory):
-        mixminion.server.Queue.DeliveryQueue.__init__(self, directory)
+    def __init__(self, module, directory, retrySchedule=None):
+        mixminion.server.ServerQueue.DeliveryQueue.__init__(self, directory,
+                                                            retrySchedule)
         self.module = module
+        
 
     def _deliverMessages(self, msgList):
-        for handle, addr, message, n_retries in msgList:
+        for handle, packet in msgList:
             try:
-                exitType, address, tag = addr
-                result = self.module.processMessage(message,tag,exitType,address)
+                EventStats.log.attemptedDelivery() #FFFF
+                result = self.module.processMessage(packet)
                 if result == DELIVER_OK:
+                    LOG.debug("Successfully delivered message MOD:%s", handle)
                     self.deliverySucceeded(handle)
+                    EventStats.log.successfulDelivery() #FFFF
                 elif result == DELIVER_FAIL_RETRY:
+                    LOG.debug("Unable to deliver message MOD:%s; will retry",
+                              handle)
                     self.deliveryFailed(handle, 1)
+                    EventStats.log.failedDelivery() #FFFF
                 else:
-                    LOG.error("Unable to deliver message")
+                    assert result == DELIVER_FAIL_NORETRY
+                    LOG.error("Unable to deliver message MOD:%s; giving up",
+                              handle)
                     self.deliveryFailed(handle, 0)
+                    EventStats.log.unretriableDelivery() #FFFF
             except:
                 LOG.error_exc(sys.exc_info(),
                                    "Exception delivering message")
                 self.deliveryFailed(handle, 0)
+                EventStats.log.unretriableDelivery() #FFFF
+
+class DeliveryThread(threading.Thread):
+    """A thread object used by ModuleManager to send messages in the
+       background; delegates to ModuleManager._sendReadyMessages."""
+    ## Fields:
+    # moduleManager -- a ModuleManager object.
+    # event -- an Event that is set when we have messages to deliver, or
+    #    when we're stopping.
+    # __stoppingEvent -- an event that is set when we're shutting down.
+    def __init__(self, moduleManager):
+        """Create a new DeliveryThread."""
+        threading.Thread.__init__(self)
+        self.moduleManager = moduleManager
+        self.event = threading.Event()
+        self.__stoppingevent = threading.Event()
+
+    def beginSending(self):
+        """Tell this thread that there are messages ready to be sent."""
+        self.event.set()
+
+    def shutdown(self):
+        """Tell this thread to shut down after sending further messages."""
+        LOG.info("Telling delivery thread to shut down.")
+        self.__stoppingevent.set()
+        self.event.set()
+
+    def run(self):
+        try:
+            while 1:
+                self.event.wait()
+                self.event.clear()
+                stop = self.__stoppingevent.isSet()
+                if stop:
+                    LOG.info("Delivery thread shutting down.")
+                    return
+                self.moduleManager._sendReadyMessages()
+                waitForChildren(blocking=0)
+        except:
+            LOG.error_exc(sys.exc_info(),
+                          "Exception in delivery; shutting down thread.")
 
 class ModuleManager:
     """A ModuleManager knows about all of the server modules in the system.
@@ -202,6 +266,7 @@ class ModuleManager:
     #            queueMessage and sendReadyMessages as in DeliveryQueue.)
     #    _isConfigured: flag: has this modulemanager's configure method been
     #            called?
+    #    thread: None, or a DeliveryThread object.
 
     def __init__(self):
         "Create a new ModuleManager"
@@ -221,6 +286,13 @@ class ModuleManager:
         self.registerModule(MixmasterSMTPModule())
 
         self._isConfigured = 0
+        self.thread = None
+
+    def startThreading(self):
+        """Begin delivering messages in a separte thread.  Should only
+           be called once."""
+        self.thread = DeliveryThread(self)
+        self.thread.start()
 
     def isConfigured(self):
         """Return true iff this object's configure method has been called"""
@@ -251,7 +323,7 @@ class ModuleManager:
     def setPath(self, path):
         """Sets the search path for Python modules"""
         if path:
-            self.path = path.split(":")
+            self.path = [ os.path.expanduser(fn) for fn in path.split(":") ]
         else:
             self.path = []
 
@@ -281,10 +353,10 @@ class ModuleManager:
         except Exception, e:
             raise MixError("Error initializing module %s" %className)
 
-    def validate(self, sections, entries, lines, contents):
+    def validate(self, config, lines, contents):
         # (As in ServerConfig)
         for m in self.modules:
-            m.validateConfig(sections, entries, lines, contents)
+            m.validateConfig(config, lines, contents)
 
     def configure(self, config):
         self._setQueueRoot(os.path.join(config['Server']['Homedir'],
@@ -312,14 +384,14 @@ class ModuleManager:
         self.queues[module.getName()] = queue
         self.enabled[module.getName()] = 1
 
-    def cleanQueues(self):
+    def cleanQueues(self, deleteFn=None):
         """Remove trash messages from all internal queues."""
         for queue in self.queues.values():
-            queue.cleanQueue()
+            queue.cleanQueue(deleteFn)
 
     def disableModule(self, module):
         """Unmaps all the types for a module object."""
-        LOG.info("Disabling module %s", module.getName())
+        LOG.debug("Disabling module %s", module.getName())
         for t in module.getExitTypes():
             if (self.typeToModule.has_key(t) and
                 self.typeToModule[t].getName() == module.getName()):
@@ -329,36 +401,46 @@ class ModuleManager:
         if self.enabled.has_key(module.getName()):
             del self.enabled[module.getName()]
 
-    def queueMessage(self, message, tag, exitType, address):
-        """Queue a message for delivery."""
-        # FFFF Support non-exit messages.
-        mod = self.typeToModule.get(exitType, None)
+    def queueDecodedMessage(self, packet):
+        """Given a packet of type DeliveryPacket, try to find an appropriate
+           exit module, and queue the packet for delivey by that exit module.
+        """
+        exitType = packet.getExitType()
+
+        mod = self.typeToModule.get(exitType)
         if mod is None:
             LOG.error("Unable to handle message with unknown type %s",
-                           exitType)
-            return
+                      exitType)
+            return "<nil>"
         queue = self.queues[mod.getName()]
         LOG.debug("Delivering message %r (type %04x) via module %s",
-                       message[:8], exitType, mod.getName())
-        payload = None
-        try:
-            payload = mixminion.BuildMessage.decodePayload(message, tag)
-        except CompressedDataTooLong:
-            contents = mixminion.Packet.parsePayload(message).getContents()
-            queue.queueDeliveryMessage((exitType, address, 'long'), contents)
-            return
-        except MixError:
-            queue.queueDeliveryMessage((exitType, address, 'err'), message)
-            return
+                  packet.getContents()[:8], exitType, mod.getName())
 
-        if payload is None:
-            # enrypted message
-            queue.queueDeliveryMessage((exitType, address, tag), message)
-        else:
-            # forward message
-            queue.queueDeliveryMessage((exitType, address, None), payload)
+        return queue.queueDeliveryMessage(packet)
+
+    def shutdown(self):
+        """Tell the delivery thread (if any) to stop."""
+        if self.thread is not None:
+            self.thread.shutdown()
+
+    def join(self):
+        """Wait for the delivery thread (if any) to finish shutting down."""
+        if self.thread is not None:
+            self.thread.join()
 
     def sendReadyMessages(self):
+        """Begin message delivery, either by telling every module's queue to
+           try sending its pending messages, or by telling the delivery
+           thread to do so if we're threading."""
+        if self.thread is not None:
+            self.thread.beginSending()
+        else:
+            self._sendReadyMessages()
+
+    def _sendReadyMessages(self):
+        """Actual implementation of message delivery. Tells every module's
+           queue to send pending messages.  This is called directly if
+           we aren't threading, and from the delivery thread if we are."""
         for name, queue in self.queues.items():
             queue.sendReadyMessages()
 
@@ -371,6 +453,8 @@ class DropModule(DeliveryModule):
     """Null-object pattern: drops all messages it receives."""
     def getConfigSyntax(self):
         return { }
+    def getRetrySchedule(self):
+        return [ ]
     def getServerInfoBlock(self):
         return ""
     def configure(self, config, manager):
@@ -381,7 +465,7 @@ class DropModule(DeliveryModule):
         return [ mixminion.Packet.DROP_TYPE ]
     def createDeliveryQueue(self, directory):
         return ImmediateDeliveryQueue(self)
-    def processMessage(self, message, tag, exitType, exitInfo):
+    def processMessage(self, packet):
         LOG.debug("Dropping padding message")
         return DELIVER_OK
 
@@ -404,8 +488,8 @@ class EmailAddressSet:
                or any parent domain thereof.  "Deny allhosts fred.com" matches
                "bob@fred.com" and "bob@host.fred.com", but not "bob@com".
              pattern: match the email address if the provided regex appears
-               anywhere in it.  "Pattern /./" matches everything;
-               "Pattern /(..)*/" matches all addresses with an even number
+               anywhere in it.  "Deny pattern /./" matches everything;
+               "Deny pattern /(..)*/" matches all addresses with an even number
                of characters.
     """
     ## Fields
@@ -419,9 +503,7 @@ class EmailAddressSet:
     def __init__(self, fname=None, string=None, includeStr="deny"):
         """Read the address set from a file or a string."""
         if string is None:
-            f = open(fname, 'r')
-            string = f.read()
-            f.close()
+            string = readFile(fname)
 
         self.addresses = {}
         self.domains = {}
@@ -505,7 +587,7 @@ class EmailAddressSet:
         domparts = dom.split(".")
         for idx in range(len(domparts)):
             subdom = ".".join(domparts[idx:])
-            if self.domains.get(subdom,None) == 'SUB':
+            if self.domains.get(subdom) == 'SUB':
                 return 1
 
         # Does it match any patterns?
@@ -539,19 +621,24 @@ class MBoxModule(DeliveryModule):
         DeliveryModule.__init__(self)
         self.addresses = {}
 
+    def getRetrySchedule(self):
+        return self.retrySchedule
+
     def getConfigSyntax(self):
         # FFFF There should be some way to say that fields are required
         # FFFF if the module is enabled.
         return { "Delivery/MBOX" :
                  { 'Enabled' : ('REQUIRE',  _parseBoolean, "no"),
+                   'Retry': ('ALLOW', _parseIntervalList,
+                             "7 hours for 6 days"),
                    'AddressFile' : ('ALLOW', None, None),
                    'ReturnAddress' : ('ALLOW', None, None),
                    'RemoveContact' : ('ALLOW', None, None),
                    'SMTPServer' : ('ALLOW', None, 'localhost') }
                  }
 
-    def validateConfig(self, sections, entries, lines, contents):
-        sec = sections['Delivery/MBOX']
+    def validateConfig(self, config, lines, contents):
+        sec = config['Delivery/MBOX']
         if not sec.get('Enabled'):
             return
         for field in ['AddressFile', 'ReturnAddress', 'RemoveContact',
@@ -566,6 +653,8 @@ class MBoxModule(DeliveryModule):
                 LOG.warn("Value of %s (%s) doesn't look like an email address",
                          field, sec[field])
 
+        config.validateRetrySchedule("Delivery/MBOX")
+
     def configure(self, config, moduleManager):
         if not config['Delivery/MBOX'].get("Enabled", 0):
             moduleManager.disableModule(self)
@@ -576,6 +665,7 @@ class MBoxModule(DeliveryModule):
         self.addressFile = sec['AddressFile']
         self.returnAddress = sec['ReturnAddress']
         self.contact = sec['RemoveContact']
+        self.retrySchedule = sec['Retry']
         # validate should have caught these.
         assert (self.server and self.addressFile and self.returnAddress
                 and self.contact)
@@ -588,26 +678,26 @@ class MBoxModule(DeliveryModule):
         # Parse the address file.
         self.addresses = {}
         f = open(self.addressFile)
-        address_line_re = re.compile(r'\s*([^\s:=]+)\s*[:=]\s*(\S+)')
         try:
-            lineno = 0
-            while 1:
-                line = f.readline()
-                if not line:
-                    break
-                line = line.strip()
-                lineno += 1
-                if line == '' or line[0] == '#':
-                    continue
-                m = address_line_re.match(line)
-                if not m:
-                    raise ConfigError("Bad address on line %s of %s"%(
-                        lineno,self.addressFile))
-                self.addresses[m.group(1)] = m.group(2)
-                LOG.trace("Mapping MBOX address %s -> %s", m.group(1),
-                               m.group(2))
+            lines = f.readlines()
         finally:
             f.close()
+        
+        address_line_re = re.compile(r'([^\s:=]+)\s*[:=]\s*(\S+)')
+
+        lineno = 0
+        for line in lines:
+            line = line.strip()
+            lineno += 1
+            if line == '' or line[0] == '#':
+                continue
+            m = address_line_re.match(line)
+            if not m:
+                raise ConfigError("Bad address on line %s of %s"%(
+                    lineno,self.addressFile))
+            self.addresses[m.group(1)] = m.group(2)
+            LOG.trace("Mapping MBOX address %s -> %s", m.group(1),
+                           m.group(2))
 
         moduleManager.enableModule(self)
 
@@ -623,11 +713,11 @@ class MBoxModule(DeliveryModule):
     def getExitTypes(self):
         return [ mixminion.Packet.MBOX_TYPE ]
 
-    def processMessage(self, message, tag, exitType, address):
+    def processMessage(self, packet): #message, tag, exitType, address):
         # Determine that message's address;
-        assert exitType == mixminion.Packet.MBOX_TYPE
+        assert packet.getExitType() == mixminion.Packet.MBOX_TYPE
         LOG.debug("Received MBOX message")
-        info = mixminion.Packet.parseMBOXInfo(address)
+        info = mixminion.Packet.parseMBOXInfo(packet.getAddress())
         try:
             address = self.addresses[info.user]
         except KeyError:
@@ -635,7 +725,7 @@ class MBoxModule(DeliveryModule):
             return DELIVER_FAIL_NORETRY
 
         # Escape the message if it isn't plaintext ascii
-        msg = _escapeMessageForEmail(message, tag)
+        msg = _escapeMessageForEmail(packet)
 
         # Generate the boilerplate (FFFF Make this configurable)
         fields = { 'user': address,
@@ -647,7 +737,8 @@ class MBoxModule(DeliveryModule):
         msg = """\
 To: %(user)s
 From: %(return)s
-Subject: Anonymous Mixminion message
+Subject: Type III anonymous message
+X-Anonymous: yes
 
 THIS IS AN ANONYMOUS MESSAGE.  The mixminion server '%(nickname)s' at
 %(addr)s has been configured to deliver messages to your address.
@@ -684,9 +775,14 @@ class DirectSMTPModule(SMTPModule):
     def __init__(self):
         SMTPModule.__init__(self)
 
+    def getRetrySchedule(self):
+        return self.retrySchedule
+
     def getConfigSyntax(self):
         return { "Delivery/SMTP" :
                  { 'Enabled' : ('REQUIRE', _parseBoolean, "no"),
+                   'Retry': ('ALLOW', _parseIntervalList,
+                             "7 hours for 6 days"),
                    'BlacklistFile' : ('ALLOW', None, None),
                    'SMTPServer' : ('ALLOW', None, 'localhost'),
                    'Message' : ('ALLOW', None, ""),
@@ -697,8 +793,8 @@ class DirectSMTPModule(SMTPModule):
                    }
                  }
 
-    def validateConfig(self, sections, entries, lines, contents):
-        sec = sections['Delivery/SMTP']
+    def validateConfig(self, config, lines, contents):
+        sec = config['Delivery/SMTP']
         if not sec.get('Enabled'):
             return
         for field in 'SMTPServer', 'ReturnAddress':
@@ -711,6 +807,8 @@ class DirectSMTPModule(SMTPModule):
             LOG.warn("Return address (%s) doesn't look like an email address",
                      sec['ReturnAddress'])
 
+        config.validateRetrySchedule("Delivery/SMTP")
+
     def configure(self, config, manager):
         sec = config['Delivery/SMTP']
         if not sec.get('Enabled'):
@@ -718,6 +816,7 @@ class DirectSMTPModule(SMTPModule):
             return
 
         self.server = sec['SMTPServer']
+        self.retrySchedule = sec['Retry']
         if sec['BlacklistFile']:
             self.blacklist = EmailAddressSet(fname=sec['BlacklistFile'])
         else:
@@ -726,19 +825,20 @@ class DirectSMTPModule(SMTPModule):
         subject = sec['SubjectLine']
         self.returnAddress = sec['ReturnAddress']
 
-        self.header = "From: %s\nSubject: %s\n\n%s" %(
+        self.header = "From: %s\nSubject: %s\nX-Anonymous: yes\n\n%s" %(
             self.returnAddress, subject, message)
 
         manager.enableModule(self)
 
-    def processMessage(self, message, tag, exitType, address):
-        assert exitType == mixminion.Packet.SMTP_TYPE
+    def processMessage(self, packet):
+        assert packet.getExitType() == mixminion.Packet.SMTP_TYPE
         LOG.debug("Received SMTP message")
         # parseSMTPInfo will raise a parse error if the mailbox is invalid.
         try:
-            address = mixminion.Packet.parseSMTPInfo(address).email
+            address = mixminion.Packet.parseSMTPInfo(packet.getAddress()).email
         except ParseError:
-            LOG.warn("Dropping SMTP message to invalid address %r", address)
+            LOG.warn("Dropping SMTP message to invalid address %r",
+                     packet.getAddress())
             return DELIVER_FAIL_NORETRY
 
         # Now, have we blacklisted this address?
@@ -747,7 +847,7 @@ class DirectSMTPModule(SMTPModule):
             return DELIVER_FAIL_NORETRY
 
         # Decode and escape the message, and get ready to send it.
-        msg = _escapeMessageForEmail(message, tag)
+        msg = _escapeMessageForEmail(packet)
         msg = "To: %s\n%s\n\n%s"%(address, self.header, msg)
 
         # Send the message.
@@ -757,10 +857,9 @@ class MixmasterSMTPModule(SMTPModule):
     """Implements SMTP by relaying messages via Mixmaster nodes.  This
        is kind of unreliable and kludgey, but it does allow us to
        test mixminion by usingg Mixmaster nodes as exits."""
-    # FFFF Mixmaster has tons of options.  Maybe we should use 'em...
-    # FFFF ... or maybe we should deliberately ignore them, since
-    # FFFF this is only a temporary workaround until enough people
-    # FFFF are running SMTP exit nodes
+    # (Mixmaster has tons of options, but we ignore them, since
+    #  this is only a temporary workaround until enough people
+    #  are running SMTP exit nodes.)
     ## Fields:
     # server: The path (usually a single server) to use for outgoing messages.
     #    Multiple servers should be separated by commas.
@@ -773,9 +872,14 @@ class MixmasterSMTPModule(SMTPModule):
     def __init__(self):
         SMTPModule.__init__(self)
 
+    def getRetrySchedule(self):
+        return self.retrySchedule
+
     def getConfigSyntax(self):
         return { "Delivery/SMTP-Via-Mixmaster" :
                  { 'Enabled' : ('REQUIRE', _parseBoolean, "no"),
+                   'Retry': ('ALLOW', _parseIntervalList,
+                             "7 hours for 6 days"),
                    'MixCommand' : ('REQUIRE', _parseCommand, None),
                    'Server' : ('REQUIRE', None, None),
                    'SubjectLine' : ('ALLOW', None,
@@ -783,9 +887,12 @@ class MixmasterSMTPModule(SMTPModule):
                    }
                  }
 
-    def validateConfig(self, sections, entries, lines, contents):
-        # Currently, we accept any configuration options that the config allows
-        pass
+    def validateConfig(self, config, lines, contents):
+        #FFFF write more
+        sec = config['Delivery/SMTP-Via-Mixmaster']
+        if not sec.get("Enabled"):
+            return
+        config.validateRetrySchedule("Delivery/SMTP-Via-Mixmaster")
 
     def configure(self, config, manager):
         sec = config['Delivery/SMTP-Via-Mixmaster']
@@ -795,6 +902,7 @@ class MixmasterSMTPModule(SMTPModule):
         cmd = sec['MixCommand']
         self.server = sec['Server']
         self.subject = sec['SubjectLine']
+        self.retrySchedule = sec['Retry']
         self.command = cmd[0]
         self.options = tuple(cmd[1]) + ("-l", self.server,
                                         "-s", self.subject)
@@ -806,21 +914,22 @@ class MixmasterSMTPModule(SMTPModule):
     def createDeliveryQueue(self, queueDir):
         # We create a temporary queue so we can hold files there for a little
         # while before passing their names to mixmaster.
-        self.tmpQueue = mixminion.server.Queue.Queue(queueDir+"_tmp", 1, 1)
+        self.tmpQueue = mixminion.server.ServerQueue.Queue(queueDir+"_tmp", 1, 1)
         self.tmpQueue.removeAll()
         return _MixmasterSMTPModuleDeliveryQueue(self, queueDir)
 
-    def processMessage(self, message, tag, exitType, smtpAddress):
+    def processMessage(self, packet):
         """Insert a message into the Mixmaster queue"""
-        assert exitType == mixminion.Packet.SMTP_TYPE
+        assert packet.getExitType() == mixminion.Packet.SMTP_TYPE
         # parseSMTPInfo will raise a parse error if the mailbox is invalid.
         try:
-            info = mixminion.Packet.parseSMTPInfo(smtpAddress)
+            info = mixminion.Packet.parseSMTPInfo(packet.getAddress())
         except ParseError:
-            LOG.warn("Dropping SMTP message to invalid address %r",smtpAddress)
+            LOG.warn("Dropping SMTP message to invalid address %r",
+                     packet.getAddress())
             return DELIVER_FAIL_NORETRY
 
-        msg = _escapeMessageForEmail(message, tag)
+        msg = _escapeMessageForEmail(packet)
         handle = self.tmpQueue.queueMessage(msg)
 
         cmd = self.command
@@ -836,7 +945,7 @@ class MixmasterSMTPModule(SMTPModule):
            should be called after invocations of processMessage."""
         cmd = self.command
         LOG.debug("Flushing Mixmaster pool")
-        os.spawnl(os.P_NOWAIT, cmd, cmd, "-S")
+        os.spawnl(os.P_WAIT, cmd, cmd, "-S")
 
 class _MixmasterSMTPModuleDeliveryQueue(SimpleModuleDeliveryQueue):
     """Delivery queue for _MixmasterSMTPModule.  Same as
@@ -873,89 +982,34 @@ def sendSMTPMessage(server, toList, fromAddr, message):
 
 #----------------------------------------------------------------------
 
-def _escapeMessageForEmail(msg, tag):
+def _escapeMessageForEmail(packet):
     """Helper function: Given a message and tag, escape the message if
        it is not plaintext ascii, and wrap it in some standard
        boilerplate.  Add a disclaimer if the message is not ascii.
 
-          msg -- A (possibly decoded) message
-          tag -- One of: a 20-byte decoding tag [if the message is encrypted
-                            or a reply]
-                         None [if the message is in plaintext]
-                         'err' [if the message was invalid.]
-                         'long' [if the message might be a zlib bomb'].
+          packet -- an instance of DeliveryPacket
 
        Returns None on an invalid message."""
-    m = _escapeMessage(msg, tag, text=1)
-    if m is None:
+    if packet.isError():
         return None
-    code, msg, tag = m
 
-    if code == 'ENC':
+    if packet.isEncrypted():
         junk_msg = """\
 This message is not in plaintext.  It's either 1) a reply; 2) a forward
 message encrypted to you; or 3) junk.\n\n"""
-    elif code == 'ZB':
+    elif packet.isOvercompressed():
         junk_msg = """\
 This message is compressed with zlib.  Ordinarily, I would have decompressed
 it, but it was compressed by more than a factor of 20, which makes me nervous.
 \n"""
-    elif code == 'BIN':
+    elif not packet.isPrintingAscii():
+        assert packet.isPlaintext()
         junk_msg = """\
 This message contains nonprinting characters, so I encoded it with Base64
 before sending it to you.\n\n"""
     else:
+        assert packet.isPlaintext()
         junk_msg = ""
 
-    if tag is not None:
-        tag = "Decoding handle: "+tag+"\n"
-    else:
-        tag = ""
-
-    if msg and msg[-1] != '\n':
-        extra_newline = "\n"
-    else:
-        extra_newline = ""
-
-    return """\
-%s======= TYPE III ANONYMOUS MESSAGE BEGINS ========
-%s%s%s======== TYPE III ANONYMOUS MESSAGE ENDS =========
-""" %(junk_msg, tag, msg, extra_newline)
-
-def _escapeMessage(message, tag, text=0):
-    """Helper: given a decoded message (and possibly its tag), determine
-       whether the message is a text plaintext message (code='TXT'), a
-       binary plaintext message (code 'BIN'), an encrypted message/reply
-       (code='ENC'), or a plaintext possible zlib bomb ('ZB').  If
-       requested, non-TXT messages are base-64 encoded.
-
-       Returns: (code, message, tag (for ENC) or None (for BIN, TXT).
-       Returns None if the message is invalid.
-
-          message -- A (possibly decoded) message
-          tag -- One of: a 20-byte decoding tag [if the message is encrypted
-                            or a reply]
-                         None [if the message is in plaintext]
-                         'err' [if the message was invalid.]
-                         'long' [if the message might be a zlib bomb'].
-          text -- flag: if true, non-TXT messages must be base64-encoded.
-    """
-    if tag == 'err':
-        return None
-    elif tag == 'long':
-        code = "ZB"
-    elif tag is not None:
-        code = "ENC"
-    else:
-        assert tag is None
-        if isPrintingAscii(message, allowISO=1):
-            code = "TXT"
-        else:
-            code = "BIN"
-
-    if text and (code != "TXT") :
-        message = base64.encodestring(message)
-    if text and tag:
-        tag = base64.encodestring(tag).strip()
-
-    return code, message, tag
+    encMsg = packet.getTextEncodedMessage()
+    return "%s%s"%(junk_msg, encMsg.pack())
