@@ -1,9 +1,10 @@
 /* Copyright (c) 2002 Nick Mathewson.  See LICENSE for licensing information */
-/* $Id: crypt.c,v 1.17 2002/12/29 20:30:41 nickm Exp $ */
+/* $Id: crypt.c,v 1.24 2003/06/05 18:41:41 nickm Exp $ */
 #include <Python.h>
 
 #include <time.h>
 
+#ifndef TRUNCATED_OPENSSL_INCLUDES
 #include <openssl/bn.h>
 #include <openssl/rsa.h>
 #include <openssl/aes.h>
@@ -12,6 +13,16 @@
 #include <openssl/err.h>
 #include <openssl/pem.h>
 #include <openssl/evp.h>
+#else
+#include <bn.h>
+#include <rsa.h>
+#include <aes.h>
+#include <sha.h>
+#include <rand.h>
+#include <err.h>
+#include <pem.h>
+#include <evp.h>
+#endif
 #include <_minionlib.h>
 #include <assert.h>
 
@@ -64,10 +75,12 @@ mm_sha1(PyObject *self, PyObject *args, PyObject *kwdict)
                 return NULL;
         }
 
+        Py_BEGIN_ALLOW_THREADS
         SHA1_Init(&ctx);
         SHA1_Update(&ctx,cp,len);
         SHA1_Final(PyString_AS_STRING(output),&ctx);
         memset(&ctx,0,sizeof(ctx));
+        Py_END_ALLOW_THREADS
 
         return output;
 }
@@ -112,6 +125,7 @@ mm_aes_key(PyObject *self, PyObject *args, PyObject *kwdict)
         static char *kwlist[] = { "key", NULL };
         char *key;
         int keylen;
+        int r;
         AES_KEY *aes_key = NULL;
         PyObject *result;
 
@@ -126,7 +140,10 @@ mm_aes_key(PyObject *self, PyObject *args, PyObject *kwdict)
         if (!(aes_key = malloc(sizeof(AES_KEY)))) {
                 PyErr_NoMemory(); goto err;
         }
-        if (AES_set_encrypt_key(key, keylen*8, aes_key)) {
+        Py_BEGIN_ALLOW_THREADS
+        r = AES_set_encrypt_key(key, keylen*8, aes_key);
+        Py_END_ALLOW_THREADS
+        if (r) {
                 mm_SSL_ERR(1);
                 goto err;
         }
@@ -188,8 +205,10 @@ mm_aes_ctr128_crypt(PyObject *self, PyObject *args, PyObject *kwdict)
                 return NULL;
         }
 
+        Py_BEGIN_ALLOW_THREADS
         mm_aes_counter128(input, PyString_AS_STRING(output), inputlen,
                           aes_key, idx);
+        Py_END_ALLOW_THREADS
 
         if (prng) free(input);
         return output;
@@ -224,9 +243,11 @@ mm_strxor(PyObject *self, PyObject *args, PyObject *kwdict)
         }
 
         outp = PyString_AS_STRING(output);
+        Py_BEGIN_ALLOW_THREADS
         while (s1len--) {
                 *(outp++) = *(s1++) ^ *(s2++);
         }
+        Py_END_ALLOW_THREADS
 
         return output;
 }
@@ -249,7 +270,10 @@ mm_openssl_seed(PyObject *self, PyObject *args, PyObject *kwdict)
                                          &seed, &seedlen))
                 return NULL;
 
+        Py_BEGIN_ALLOW_THREADS
         RAND_seed(seed, seedlen);
+        Py_END_ALLOW_THREADS
+
         Py_INCREF(Py_None);
         return Py_None;
 }
@@ -307,6 +331,7 @@ mm_RSA_crypt(PyObject *self, PyObject *args, PyObject *kwdict)
 
         output = PyString_FromStringAndSize(NULL, keylen);
         out = PyString_AS_STRING(output);
+        Py_BEGIN_ALLOW_THREADS
         if (encrypt) {
                 if (pub)
                         i = RSA_public_encrypt(stringlen, string, out, rsa,
@@ -322,6 +347,7 @@ mm_RSA_crypt(PyObject *self, PyObject *args, PyObject *kwdict)
                         i = RSA_private_decrypt(stringlen, string, out, rsa,
                                                 RSA_NO_PADDING);
         }
+        Py_END_ALLOW_THREADS
 
         if (i <= 0) {
                 Py_DECREF(output);
@@ -359,7 +385,10 @@ mm_rsa_generate(PyObject *self, PyObject *args, PyObject *kwdict)
                 return NULL;
         }
 
+        Py_BEGIN_ALLOW_THREADS
         rsa = RSA_generate_key(bits, e, NULL, NULL);
+        Py_END_ALLOW_THREADS
+
         if (rsa == NULL) {
                 mm_SSL_ERR(1);
                 return NULL;
@@ -383,7 +412,7 @@ mm_RSA_encode_key(PyObject *self, PyObject *args, PyObject *kwdict)
 
         int len;
         PyObject *output;
-        unsigned char *out, *outp;
+        unsigned char *out = NULL, *outp;
 
         assert(mm_RSA_Check(self));
         if (!PyArg_ParseTupleAndKeywords(args, kwdict,
@@ -396,19 +425,21 @@ mm_RSA_encode_key(PyObject *self, PyObject *args, PyObject *kwdict)
                 return NULL;
         }
 
+        Py_BEGIN_ALLOW_THREADS
         len = public ? i2d_RSAPublicKey(rsa,NULL) :
                 i2d_RSAPrivateKey(rsa,NULL);
-        if (len < 0) {
-                mm_SSL_ERR(1);
-                return NULL;
+        if (len >= 0) {
+                out = outp = malloc(len+1);
+                if (public)
+                        len = i2d_RSAPublicKey(rsa, &outp);
+                else
+                        len = i2d_RSAPrivateKey(rsa, &outp);
         }
-        out = outp = malloc(len+1);
-        if (public)
-                len = i2d_RSAPublicKey(rsa, &outp);
-        else
-                len = i2d_RSAPrivateKey(rsa, &outp);
+        Py_END_ALLOW_THREADS
+
         if (len < 0) {
-                free(out);
+                if (out)
+                        free(out);
                 mm_SSL_ERR(1);
                 return NULL;
         }
@@ -441,8 +472,10 @@ mm_rsa_decode_key(PyObject *self, PyObject *args, PyObject *kwdict)
                                          &string, &stringlen, &public))
                 return NULL;
 
+        Py_BEGIN_ALLOW_THREADS
         rsa = public ? d2i_RSAPublicKey(NULL, &string, stringlen) :
                 d2i_RSAPrivateKey(NULL, &string, stringlen);
+        Py_END_ALLOW_THREADS
         if (!rsa) {
                 mm_SSL_ERR(1);
                 return NULL;
@@ -468,6 +501,7 @@ mm_RSA_PEM_write_key(PyObject *self, PyObject *args, PyObject *kwdict)
         RSA *rsa = NULL;
         EVP_PKEY *pkey = NULL;
         FILE *file;
+        int ok = 0;
 
         assert(mm_RSA_Check(self));
         if (!PyArg_ParseTupleAndKeywords(args, kwdict, "O!i|s#:PEM_write_key",
@@ -480,6 +514,7 @@ mm_RSA_PEM_write_key(PyObject *self, PyObject *args, PyObject *kwdict)
                 return NULL;
         }
 
+        Py_BEGIN_ALLOW_THREADS
         if (public) {
                 rsa = ((mm_RSA*)self)->rsa;
                 if (!PEM_write_RSAPublicKey(file, rsa))
@@ -507,16 +542,22 @@ mm_RSA_PEM_write_key(PyObject *self, PyObject *args, PyObject *kwdict)
                                 goto error;
                 }
         }
-        Py_INCREF(Py_None);
-        return Py_None;
 
+        ok = 1;
+        goto done;
  error:
+        mm_SSL_ERR(1);
+ done:
         if (rsa && !public)
                 RSA_free(rsa);
         if (pkey)
                 EVP_PKEY_free(pkey);
 
-        mm_SSL_ERR(1);
+        Py_END_ALLOW_THREADS
+        if (ok) {
+                Py_INCREF(Py_None);
+                return Py_None;
+        }
         return NULL;
 }
 
@@ -551,12 +592,15 @@ mm_rsa_PEM_read_key(PyObject *self, PyObject *args, PyObject *kwdict)
         if (!passwordlen)
                 password = "";
 
+        Py_BEGIN_ALLOW_THREADS
         if (public) {
                 rsa = PEM_read_RSAPublicKey(file, NULL, NULL, NULL);
         } else {
                 rsa = PEM_read_RSAPrivateKey(file, NULL,
                                              NULL, password);
         }
+        Py_END_ALLOW_THREADS
+
         if (!rsa) {
                 mm_SSL_ERR(1);
                 return NULL;
@@ -565,10 +609,6 @@ mm_rsa_PEM_read_key(PyObject *self, PyObject *args, PyObject *kwdict)
         return mm_RSA_new(rsa);
 }
 
-
-
-
-
 /**
  * Converts a BIGNUM into a newly allocated PyLongObject.
  **/
@@ -626,6 +666,7 @@ const char mm_RSA_get_public_key__doc__[]=
 PyObject *
 mm_RSA_get_public_key(PyObject *self, PyObject *args, PyObject *kwdict)
 {
+        /* ???? should be threadified? */
         static char *kwlist[] = {  NULL };
 
         RSA *rsa;
@@ -658,6 +699,7 @@ const char mm_RSA_get_exponent__doc__[]=
 PyObject *
 mm_RSA_get_exponent(PyObject *self, PyObject *args, PyObject *kwdict)
 {
+        /* ???? should be threadified? */
         static char *kwlist[] = {  NULL };
 
         RSA *rsa;
@@ -683,6 +725,7 @@ const char mm_rsa_make_public_key__doc__[]=
 PyObject *
 mm_rsa_make_public_key(PyObject *self, PyObject *args, PyObject *kwdict)
 {
+        /* ???? Should be threadified */
         static char *kwlist[] = { "n","e", NULL };
 
         RSA *rsa;
@@ -694,7 +737,6 @@ mm_rsa_make_public_key(PyObject *self, PyObject *args, PyObject *kwdict)
                                          &PyLong_Type, &n, &PyLong_Type, &e))
                 return NULL;
 
-        rsa = RSA_new();
         if (!(rsa = RSA_new())) { PyErr_NoMemory(); return NULL; }
         if (!(rsa->n = pylong2bn(n))) { RSA_free(rsa); return NULL; }
         if (!(rsa->e = pylong2bn(e))) {
@@ -713,6 +755,7 @@ const char mm_RSA_get_modulus_bytes__doc__[]=
 static PyObject *
 mm_RSA_get_modulus_bytes(PyObject *self, PyObject *args, PyObject *kwargs)
 {
+        /* ???? should be threadified? */
         static char *kwlist[] = { NULL };
         RSA *rsa;
 
@@ -745,7 +788,7 @@ static const char mm_RSA_Type__doc__[] =
   "An RSA key.  May be public or private.";
 
 PyTypeObject mm_RSA_Type = {
-        PyObject_HEAD_INIT(&PyType_Type)
+        PyObject_HEAD_INIT(/*&PyType_Type*/ 0)
         0,                                  /*ob_size*/
         "mixminion._minionlib.RSA",         /*tp_name*/
         sizeof(mm_RSA),                     /*tp_basicsize*/
@@ -797,9 +840,12 @@ mm_add_oaep_padding(PyObject *self, PyObject *args, PyObject *kwargs)
                 PyErr_NoMemory(); return NULL;
         }
 
+        Py_BEGIN_ALLOW_THREADS
         r = RSA_padding_add_PKCS1_OAEP(PyString_AS_STRING(output), keylen,
                                        input, inputlen,
                                        param, paramlen);
+        Py_END_ALLOW_THREADS
+
         if (r <= 0) {
                 mm_SSL_ERR(1);
                 Py_DECREF(output);
@@ -842,9 +888,11 @@ mm_check_oaep_padding(PyObject *self, PyObject *args, PyObject *kwargs)
                 PyErr_NoMemory(); return NULL;
         }
 
+        Py_BEGIN_ALLOW_THREADS
         r = RSA_padding_check_PKCS1_OAEP(PyString_AS_STRING(output), r,
                                          input+1, inputlen-1, keylen,
                                          param, paramlen);
+        Py_END_ALLOW_THREADS
         if (r <= 0) {
                 mm_SSL_ERR(1);
                 Py_DECREF(output);
@@ -874,6 +922,7 @@ const char mm_generate_dh_parameters__doc__[] =
 PyObject *
 mm_generate_dh_parameters(PyObject *self, PyObject *args, PyObject *kwargs)
 {
+        /* ???? should be threadified? */
         static char *kwlist[] = { "filename", "verbose", "bits", NULL };
         char *filename;
         int bits=512, verbose=0;
@@ -887,12 +936,12 @@ mm_generate_dh_parameters(PyObject *self, PyObject *args, PyObject *kwargs)
                                          &filename, &verbose, &bits))
                 return NULL;
 
-        if (!(out = BIO_new_file(filename, "w")))
-                goto error;
-        if (!(dh = DH_generate_parameters(bits, 2,
-                                          verbose?gen_dh_callback:NULL,
-                                          NULL)))
-                goto error;
+        out = BIO_new_file(filename, "w");
+        if (out)
+                dh = DH_generate_parameters(bits, 2,
+                                            verbose?gen_dh_callback:NULL,
+                                            NULL);
+        if (out && dh)
         if (!PEM_write_bio_DHparams(out, dh))
                 goto error;
         BIO_free(out);
@@ -910,20 +959,23 @@ mm_generate_dh_parameters(PyObject *self, PyObject *args, PyObject *kwargs)
 }
 
 const char mm_generate_cert__doc__[] =
-  "generate_cert(filename, rsa, cn, start_time, end_time)\n\n"
-  "Generate a self-signed X509 certificate suitable for use by a Mixminion\n"
-  "server.  The certificate will be stored to <filename>, and use the\n"
-  "=private= key <rsa>.  The certificate\'s commonName field will be set to\n"
-  "<cn>.  The key will be valid from <start_time> until <end_time>.\n"
-  "All other fields will be given reasonable defaults.\n";
+  "generate_cert(filename, rsa, rsa_sign, cn, cn_issuer, \n"
+  "              start_time, end_time)\n\n"
+  "Generate a signed X509 certificate suitable for use by a Mixminion\n"
+  "server.  The certificate is stored to <filename>, uses the =private= key\n"
+  "<rsa>, and is signed using the =private= key <rsa_sign>.  The\n"
+  "certificate\'s commonName field is set to <cn>, and the issuer\'s\n"
+  "commonName will be set to <cn_issuer>. The key is valid from <start_time>\n"
+  "until <end_time>.  All other fields are given reasonable defaults.\n";
 
 PyObject *
 mm_generate_cert(PyObject *self, PyObject *args, PyObject *kwargs)
 {
-        static char *kwlist[] = { "filename", "rsa", "cn",
+        static char *kwlist[] = { "filename", "rsa", "rsa_sign",
+                                  "cn", "cn_issuer",
                                   "start_time", "end_time", NULL };
-        char *filename, *cn;
-        PyObject *_rsa;
+        char *filename, *cn, *cn_issuer;
+        PyObject *_rsa, *_rsa_sign;
 
         /*
          * Python wants time to be a double. OpenSSL wants time_t.
@@ -934,20 +986,27 @@ mm_generate_cert(PyObject *self, PyObject *args, PyObject *kwargs)
         double start_time, end_time;
 
         RSA *rsa = NULL;
+        RSA *rsa_sign = NULL;
         EVP_PKEY *pkey = NULL;
+        EVP_PKEY *pkey_sign = NULL;
         BIO *out = NULL;
         X509 *x509 = NULL;
         X509_NAME *name = NULL;
+        X509_NAME *name_issuer = NULL;
         int nid;
         PyObject *retval;
         time_t time;
 
-        if (!PyArg_ParseTupleAndKeywords(args, kwargs, "sO!sdd:generate_cert",
+        if (!PyArg_ParseTupleAndKeywords(args, kwargs,
+                                         "sO!O!ssdd:generate_cert",
                                          kwlist, &filename,
-                                         &mm_RSA_Type, &_rsa, &cn,
+                                         &mm_RSA_Type, &_rsa,
+                                         &mm_RSA_Type, &_rsa_sign,
+                                         &cn, &cn_issuer,
                                          &start_time, &end_time))
                 return NULL;
 
+        Py_BEGIN_ALLOW_THREADS
         if (!(rsa = RSAPrivateKey_dup(((mm_RSA*)_rsa)->rsa)))
                 goto error;
         if (!(pkey = EVP_PKEY_new()))
@@ -956,25 +1015,39 @@ mm_generate_cert(PyObject *self, PyObject *args, PyObject *kwargs)
                 goto error;
         rsa = NULL;
 
+        if (!(rsa_sign = RSAPrivateKey_dup(((mm_RSA*)_rsa_sign)->rsa)))
+                goto error;
+        if (!(pkey_sign = EVP_PKEY_new()))
+                goto error;
+        if (!(EVP_PKEY_assign_RSA(pkey_sign, rsa_sign)))
+                goto error;
+        rsa_sign = NULL;
+
         if (!(x509 = X509_new()))
                 goto error;
         if (!(X509_set_version(x509, 2)))
                 goto error;
         if (!(ASN1_INTEGER_set(X509_get_serialNumber(x509),0L)))
                 goto error;
-        if (!(name = X509_NAME_new()))
-                goto error;
 
-#define SET_PART(part, val)                                     \
-        if ((nid = OBJ_txt2nid(part)) == NID_undef) goto error; \
-        if (!X509_NAME_add_entry_by_NID(name, nid, MBSTRING_ASC,\
+#define SET_PART(n, part, val)                                   \
+        if ((nid = OBJ_txt2nid(part)) == NID_undef) goto error;  \
+        if (!X509_NAME_add_entry_by_NID(n, nid, MBSTRING_ASC,    \
                                         val, -1, -1, 0)) goto error;
 
-        SET_PART("countryName", "US");
-        SET_PART("organizationName", "Mixminion network");
-        SET_PART("commonName", cn);
+        if (!(name = X509_NAME_new()))
+                goto error;
+        SET_PART(name, "organizationName", "Mixminion network");
+        SET_PART(name, "commonName", cn);
 
-        if (!(X509_set_issuer_name(x509, name)))
+        if (!(name_issuer = X509_NAME_new()))
+                goto error;
+        SET_PART(name_issuer, "organizationName", "Mixminion network");
+        SET_PART(name_issuer, "commonName", cn_issuer);
+
+        if (!(X509_set_issuer_name(x509, name_issuer)))
+                goto error;
+        if (!(X509_set_subject_name(x509, name)))
                 goto error;
 
         time = (time_t) start_time;
@@ -985,7 +1058,7 @@ mm_generate_cert(PyObject *self, PyObject *args, PyObject *kwargs)
                 goto error;
         if (!(X509_set_pubkey(x509, pkey)))
                 goto error;
-        if (!(X509_sign(x509, pkey, EVP_md5())))
+        if (!(X509_sign(x509, pkey_sign, EVP_sha1())))
                 goto error;
 
         if (!(out = BIO_new_file(filename, "w")))
@@ -1005,13 +1078,20 @@ error:
                 BIO_free(out);
         if (name)
                 X509_NAME_free(name);
+        if (name_issuer)
+                X509_NAME_free(name_issuer);
         if (x509)
                 X509_free(x509);
         if (rsa)
                 RSA_free(rsa);
+        if (rsa_sign)
+                RSA_free(rsa_sign);
         if (pkey)
                 EVP_PKEY_free(pkey);
+        if (pkey_sign)
+                EVP_PKEY_free(pkey_sign);
 
+        Py_END_ALLOW_THREADS
         return retval;
 }
 
