@@ -1,9 +1,16 @@
 /* Copyright (c) 2002 Nick Mathewson.  See LICENSE for licensing information */
-/* $Id: tls.c,v 1.11 2002/12/16 02:40:11 nickm Exp $ */
+/* $Id: tls.c,v 1.25 2003/06/05 18:41:41 nickm Exp $ */
 #include "_minionlib.h"
 
+#include <time.h>
+
+#ifndef TRUNCATED_OPENSSL_INCLUDES
 #include <openssl/ssl.h>
 #include <openssl/tls1.h>
+#else
+#include <ssl.h>
+#include <tls1.h>
+#endif
 
 char mm_TLSError__doc__[] =
   "mixminion._minionlib.TLSError\n\n"
@@ -28,6 +35,9 @@ PyObject *mm_TLSClosed = NULL;
 /* Convenience macro to set a type error with a given string. */
 #define TYPE_ERR(s) PyErr_SetString(PyExc_TypeError, s)
 
+/* Convenience macro to set a tls error with a given string. */
+#define MM_TLS_ERR(s) PyErr_SetString(mm_TLSError, s)
+
 /* Convenience macro to set an error and quit if a 0-argument function
    was called with arguments.  (We can't just use 'METH_NOARGS', since
    that wasn't available in Python 2.0.) */
@@ -108,17 +118,18 @@ const char mm_TLSContext_new__doc__[] =
 PyObject*
 mm_TLSContext_new(PyObject *self, PyObject *args, PyObject *kwargs)
 {
-        static char *kwlist[] = { "certfile", "pkfile", "dhfile", NULL };
+        static char *kwlist[] = { "certfile", "rsa", "dhfile", NULL };
         char *certfile = NULL, *dhfile=NULL;
         mm_RSA *rsa = NULL;
+        int err = 0;
 
         SSL_METHOD *method;
-        SSL_CTX *ctx;
-        DH *dh;
-        mm_TLSContext *result;
-        BIO *bio;
+        SSL_CTX *ctx = NULL;
+        DH *dh = NULL;
+        BIO *bio = NULL;
         RSA *_rsa = NULL;
         EVP_PKEY *pkey = NULL;
+        mm_TLSContext *result;
 
         if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|sO!s:TLSContext_new",
                                          kwlist,
@@ -127,53 +138,74 @@ mm_TLSContext_new(PyObject *self, PyObject *args, PyObject *kwargs)
                                          &dhfile))
                 return NULL;
 
+        Py_BEGIN_ALLOW_THREADS;
         method = TLSv1_method();
 
-        if (!(ctx = SSL_CTX_new(method))) {
-                mm_SSL_ERR(0); return NULL;
-        }
-        if (!SSL_CTX_set_cipher_list(ctx, TLS1_TXT_DHE_RSA_WITH_AES_128_SHA)){
-                SSL_CTX_free(ctx); mm_SSL_ERR(0); return NULL;
-        }
-        if (certfile &&
-            !SSL_CTX_use_certificate_file(ctx,certfile,SSL_FILETYPE_PEM)) {
-                SSL_CTX_free(ctx); mm_SSL_ERR(0); return NULL;
-        }
-        if (rsa) {
+        if (!(ctx = SSL_CTX_new(method)))
+                err = 1;
+        if (!err && !SSL_CTX_set_cipher_list(ctx,
+                                       TLS1_TXT_DHE_RSA_WITH_AES_128_SHA))
+                err = 1;
+        if (!err && certfile &&
+            !SSL_CTX_use_certificate_chain_file(ctx,certfile))
+                err = 1;
+        if (!err)
+                SSL_CTX_set_session_cache_mode(ctx, SSL_SESS_CACHE_OFF);
+        if (!err && rsa) {
                 if (!(_rsa = RSAPrivateKey_dup(rsa->rsa)) ||
-                    !(pkey = EVP_PKEY_new()) ||
-                    !EVP_PKEY_assign_RSA(pkey, _rsa)) {
-                        if (!pkey && _rsa) RSA_free(_rsa);
-                        if (pkey) EVP_PKEY_free(pkey);
-                        SSL_CTX_free(ctx); mm_SSL_ERR(0); return NULL;
-                }
-                if (!(SSL_CTX_use_PrivateKey(ctx, pkey))) {
+                    !(pkey = EVP_PKEY_new()))
+                        err = 1;
+                if (!err && !EVP_PKEY_assign_RSA(pkey, _rsa))
+                        err = 1;
+                if (!err && !(SSL_CTX_use_PrivateKey(ctx, pkey)))
+                        err = 1;
+                if (pkey)
                         EVP_PKEY_free(pkey);
-                        SSL_CTX_free(ctx); mm_SSL_ERR(0); return NULL;
+                if (!err && certfile) {
+                        if (!SSL_CTX_check_private_key(ctx))
+                                err = 1;
                 }
-                EVP_PKEY_free(pkey);
         }
 
-        if (dhfile) {
-                if ( !(bio = BIO_new_file(dhfile, "r"))) {
-                        SSL_CTX_free(ctx); mm_SSL_ERR(0); return NULL;
-                }
-                dh=PEM_read_bio_DHparams(bio,NULL,NULL,NULL);
-                BIO_free(bio);
-                if (!dh) {
-                        SSL_CTX_free(ctx); mm_SSL_ERR(0); return NULL;
-                }
-                SSL_CTX_set_tmp_dh(ctx,dh);
-                DH_free(dh);
-        }
-        SSL_CTX_set_verify(ctx, SSL_VERIFY_NONE, NULL);
 
-        result = PyObject_New(mm_TLSContext, &mm_TLSContext_Type);
-        if (!result) {
-                SSL_CTX_free(ctx); return NULL;
+        if (!err && dhfile) {
+                if ( !(bio = BIO_new_file(dhfile, "r")))
+                        err = 1;
+                if (!err) {
+                        dh=PEM_read_bio_DHparams(bio,NULL,NULL,NULL);
+                        if (!dh)
+                                err = 1;
+                }
+                if (!err) {
+                        SSL_CTX_set_tmp_dh(ctx,dh);
+                        DH_free(dh);
+                        dh = NULL;
+                }
+                if (bio) {
+                        BIO_free(bio);
+                        bio = NULL;
+                }
+        }
+        if (!err)
+                SSL_CTX_set_verify(ctx, SSL_VERIFY_NONE, NULL);
+        Py_END_ALLOW_THREADS
+
+        if (!err) {
+                result = PyObject_New(mm_TLSContext, &mm_TLSContext_Type);
+                if (!result) {
+                        SSL_CTX_free(ctx); return NULL;
+                }
+                result->ctx = ctx;
+                return (PyObject*)result;
+        } else {
+                if (dh) DH_free(dh);
+                if (bio) BIO_free(bio);
+                if (!pkey && _rsa) RSA_free(_rsa);
+                if (pkey) EVP_PKEY_free(pkey);
+                if (ctx) SSL_CTX_free(ctx);
+                mm_SSL_ERR(0);
+                return NULL;
         }
-        result->ctx = ctx;
-        return (PyObject*)result;
 }
 
 static void
@@ -196,10 +228,10 @@ mm_TLSContext_sock(PyObject *self, PyObject *args, PyObject *kwargs)
         PyObject *sockObj;
         int serverMode = 0;
         int sock;
+        int err = 0;
 
         SSL_CTX *ctx;
-        BIO *bio;
-        SSL *ssl;
+        SSL *ssl = NULL;
         mm_TLSSock *ret;
 
         if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|i:sock",
@@ -214,31 +246,35 @@ mm_TLSContext_sock(PyObject *self, PyObject *args, PyObject *kwargs)
 
         ctx = ((mm_TLSContext*)self)->ctx;
 
-        if (!(ssl = SSL_new(ctx))) {
-                mm_SSL_ERR(0); return NULL;
-        }
+        Py_BEGIN_ALLOW_THREADS
+        if (!(ssl = SSL_new(ctx)))
+                err = 1;
 
-        if (serverMode && !SSL_set_cipher_list(ssl,
+        if (!err && serverMode && !SSL_set_cipher_list(ssl,
                     TLS1_TXT_DHE_RSA_WITH_AES_128_SHA ":"
-                    SSL3_TXT_RSA_DES_192_CBC3_SHA)) {
-                mm_SSL_ERR(0); SSL_free(ssl); return NULL;
+                    SSL3_TXT_RSA_DES_192_CBC3_SHA))
+                err = 1;
+
+        SSL_set_fd(ssl, sock);
+        Py_END_ALLOW_THREADS
+
+        if (!err) {
+                if (!(ret = PyObject_New(mm_TLSSock, &mm_TLSSock_Type))) {
+                        SSL_free(ssl); PyErr_NoMemory(); return NULL;
+                }
+                ret->ssl = ssl;
+                ret->context = self;
+                ret->sock = sock;
+                ret->sockObj = sockObj;
+                Py_INCREF(self);
+                Py_INCREF(sockObj);
+                return (PyObject*)ret;
+        } else {
+                if (ssl) SSL_free(ssl);
+                mm_SSL_ERR(0);
+                return NULL;
         }
 
-        if (!(bio = BIO_new_socket(sock, BIO_NOCLOSE))) {
-                SSL_free(ssl); mm_SSL_ERR(0); return NULL;
-        }
-        SSL_set_bio(ssl,bio,bio);
-
-        if (!(ret = PyObject_New(mm_TLSSock, &mm_TLSSock_Type))) {
-                SSL_free(ssl); PyErr_NoMemory(); SSL_free(ssl); return NULL;
-        }
-        ret->ssl = ssl;
-        ret->context = self;
-        ret->sock = sock;
-        ret->sockObj = sockObj;
-        Py_INCREF(self);
-        Py_INCREF(sockObj);
-        return (PyObject*)ret;
 }
 
 static PyMethodDef mm_TLSContext_methods[] = {
@@ -258,7 +294,7 @@ static const char mm_TLSContext_Type__doc__[] =
    "sockets.  It has a single method, 'sock()', to create new sockets.\n";
 
 PyTypeObject mm_TLSContext_Type = {
-        PyObject_HEAD_INIT(&PyType_Type)
+        PyObject_HEAD_INIT(/*&PyType_Type*/ 0)
         0,                                  /*ob_size*/
         "mixminion._minionlib.TLSContext",  /*tp_name*/
         sizeof(mm_TLSContext),              /*tp_basicsize*/
@@ -329,9 +365,11 @@ mm_TLSSock_connect(PyObject *self, PyObject *args, PyObject *kwargs)
         Py_BEGIN_ALLOW_THREADS
         r = SSL_connect(ssl);
         Py_END_ALLOW_THREADS
-        err = SSL_get_error(ssl,r);
-        if (tls_error(ssl, r, 0))
+        if (r <= 0) {
+                err = SSL_get_error(ssl,r);
+                tls_error(ssl, r, 0);
                 return NULL;
+        }
 
         Py_INCREF(Py_None);
         return Py_None;
@@ -494,6 +532,7 @@ static char mm_TLSSock_get_peer_cert_pk__doc__[] =
 static PyObject*
 mm_TLSSock_get_peer_cert_pk(PyObject *self, PyObject *args, PyObject *kwargs)
 {
+        /* ???? Should be threadified? */
         SSL *ssl;
         X509 *cert;
         EVP_PKEY *pkey;
@@ -508,6 +547,7 @@ mm_TLSSock_get_peer_cert_pk(PyObject *self, PyObject *args, PyObject *kwargs)
                 mm_SSL_ERR(0); return NULL;
         }
         pkey = X509_get_pubkey(cert);
+        X509_free(cert);
         if (!(rsa = EVP_PKEY_get1_RSA(pkey))) {
                 EVP_PKEY_free(pkey); mm_SSL_ERR(0); return NULL;
         }
@@ -521,6 +561,194 @@ mm_TLSSock_get_peer_cert_pk(PyObject *self, PyObject *args, PyObject *kwargs)
         return (PyObject*) result;
 }
 
+
+static char mm_TLSSock_check_cert_alive__doc__[] =
+  "check_cert_alive()\n\n"
+  "Raise a TLSError if the peer\'s certificate on this TLS connection is\n"
+  "not currently valid.  Otherwise return None.\n";
+
+static PyObject*
+mm_TLSSock_check_cert_alive(PyObject *self, PyObject *args, PyObject *kwargs)
+{
+        /* ???? Should be threadified? */
+        time_t now;
+        X509 *cert = NULL;
+        SSL *ssl = NULL;
+
+        assert(mm_TLSSock_Check(self));
+        FAIL_IF_ARGS();
+
+        ssl = ((mm_TLSSock*)self)->ssl;
+        if (!(cert = SSL_get_peer_certificate(ssl))) {
+                mm_SSL_ERR(0); return NULL;
+        }
+
+        /* Check expiration times. */
+        now = time(NULL);
+        if (X509_cmp_time(X509_get_notBefore(cert), &now) > 0) {
+                MM_TLS_ERR("Certificate is not yet valid");
+                goto error;
+        }
+        if (X509_cmp_time(X509_get_notAfter(cert), &now) < 0) {
+                MM_TLS_ERR("Certificate has expired");
+                goto error;
+        }
+        
+        X509_free(cert);
+        
+        Py_INCREF(Py_None);
+        return Py_None;
+ error:
+        X509_free(cert);
+        return NULL;
+}
+
+
+static char mm_TLSSock_verify_cert_and_get_identity_pk__doc__[] =
+  "verify_cert_and_get_identity_pk()\n\n"
+  "Check whether all of the following conditions hold:\n"
+  "    1) The peer of this TLS connection has a 2-certificate chain.\n"
+  "    2) The certificate for the public key used for this connection is\n"
+  "       signed by the other certificate.\n"
+  "If not, raise a TLSError.  Otherwise, return the public key on the\n"
+  "signing (identity) certificate.\n";
+
+static PyObject*
+mm_TLSSock_verify_cert_and_get_identity_pk(
+                        PyObject *self, PyObject *args, PyObject *kwargs)
+{
+        STACK_OF(X509) *chain = NULL;
+        X509 *cert = NULL;
+        X509 *id_cert = NULL;
+        SSL *ssl = NULL;
+        RSA *rsa = NULL;
+        EVP_PKEY *pkey = NULL;
+        mm_RSA *result;
+
+        int i;
+
+        assert(mm_TLSSock_Check(self));
+        FAIL_IF_ARGS();
+
+        ssl = ((mm_TLSSock*)self)->ssl;
+        if (!(chain = SSL_get_peer_cert_chain(ssl))) {
+                mm_SSL_ERR(0); goto error;
+        }
+        if (!(cert = SSL_get_peer_certificate(ssl))) {
+                mm_SSL_ERR(0); goto error;
+        }
+        if (sk_X509_num(chain) != 2) {
+                MM_TLS_ERR("Wrong number of certificates in peer chain.");
+                goto error;
+        }
+        for (i = 0; i < 2; ++i) {
+                id_cert = sk_X509_value(chain, i);
+                assert(id_cert);
+                if (X509_cmp(id_cert, cert) != 0)
+                        break;
+        }
+        if (!id_cert) {
+                MM_TLS_ERR("No distinct identity certificate found.");
+                goto error;
+        }
+        if (!(pkey = X509_get_pubkey(id_cert))) {
+                mm_SSL_ERR(0); goto error;
+        }
+        /* Is the signature correct? */
+        if (X509_verify(cert, pkey) <= 0) {
+                EVP_PKEY_free(pkey); mm_SSL_ERR(0); goto error;
+        }
+        rsa = EVP_PKEY_get1_RSA(pkey);
+        EVP_PKEY_free(pkey);
+        if (!rsa) {
+                mm_SSL_ERR(0); return NULL;
+        }
+        if (!(result = PyObject_New(mm_RSA, &mm_RSA_Type))) {
+                RSA_free(rsa); PyErr_NoMemory(); goto error;
+        }
+        result->rsa = rsa;
+
+        X509_free(cert);
+
+        return (PyObject*) result;
+ error:
+        if (cert)
+                X509_free(cert);
+        return NULL;
+}
+
+static char mm_TLSSock_renegotiate__doc__[] =
+    "tlssock.renegotate()\n\n"
+    "Mark this connection as requiring renegotiation.  No renegotiation is\n"
+    "performed until do_handshake is called.  Note that renegotiation only\n"
+    "works intuitively on the client side.\n";
+
+static PyObject*
+mm_TLSSock_renegotiate(PyObject *self, PyObject *args, PyObject *kwargs)
+{
+        SSL *ssl;
+        int r;
+
+        assert(mm_TLSSock_Check(self));
+        FAIL_IF_ARGS();
+        ssl = ((mm_TLSSock*)self)->ssl;
+
+        Py_BEGIN_ALLOW_THREADS
+        r = SSL_renegotiate(ssl);
+        Py_END_ALLOW_THREADS
+        if (!r) {
+                tls_error(ssl, r, 0);
+                return NULL;
+        }
+        Py_INCREF(Py_None);
+        return Py_None;
+}
+
+static char mm_TLSSock_do_handshake__doc__[] =
+    "tlssock.do_handshake()\n\n"
+    "Perform handshake functions for renegotiating a TLS socket.\n"
+    "Raises TLSWantRead or TLSWantWrite if the underlying nonblocking socket\n"
+    "would block on a read or a write respectively.\n";
+
+static PyObject*
+mm_TLSSock_do_handshake(PyObject *self, PyObject *args, PyObject *kwargs)
+{
+        SSL *ssl;
+        int r;
+
+        assert(mm_TLSSock_Check(self));
+        FAIL_IF_ARGS();
+        ssl = ((mm_TLSSock*)self)->ssl;
+
+        Py_BEGIN_ALLOW_THREADS
+        r = SSL_do_handshake(ssl);
+        Py_END_ALLOW_THREADS
+        if (!r) {
+                tls_error(ssl, r, 0);
+                return NULL;
+        }
+        Py_INCREF(Py_None);
+        return Py_None;
+}
+
+#if 0
+static char mm_TLSSock_get_num_renegotiations__doc__[] =
+"tlssock.get_num_renegotiations()\n\n";
+
+static PyObject*
+mm_TLSSock_get_num_renegotiations(PyObject *self, PyObject *args,
+                                  PyObject *kwargs)
+{
+        SSL *ssl;
+
+        assert(mm_TLSSock_Check(self));
+        FAIL_IF_ARGS();
+        ssl = ((mm_TLSSock*)self)->ssl;
+
+        return PyInt_FromLong(SSL_num_renegotiations(ssl));
+}
+#endif
+
 static PyMethodDef mm_TLSSock_methods[] = {
         METHOD(mm_TLSSock, accept),
         METHOD(mm_TLSSock, connect),
@@ -529,7 +757,14 @@ static PyMethodDef mm_TLSSock_methods[] = {
         METHOD(mm_TLSSock, write),
         METHOD(mm_TLSSock, shutdown),
         METHOD(mm_TLSSock, get_peer_cert_pk),
+        METHOD(mm_TLSSock, check_cert_alive),
+        METHOD(mm_TLSSock, verify_cert_and_get_identity_pk),
         METHOD(mm_TLSSock, fileno),
+        METHOD(mm_TLSSock, do_handshake),
+        METHOD(mm_TLSSock, renegotiate),
+#if 0
+        METHOD(mm_TLSSock, get_num_renegotiations),
+#endif
         { NULL, NULL }
 };
 
@@ -544,7 +779,7 @@ static const char mm_TLSSock_Type__doc__[] =
    "A single TLS connection.";
 
 PyTypeObject mm_TLSSock_Type = {
-        PyObject_HEAD_INIT(&PyType_Type)
+        PyObject_HEAD_INIT(/*&PyType_Type*/ 0)
         0,                                  /*ob_size*/
         "mixminion._minionlib.TLSSock",     /*tp_name*/
         sizeof(mm_TLSSock),                 /*tp_basicsize*/
